/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package tools;

import data.FNNXTreeNode;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Scanner;
import java.util.jar.JarFile;
import plugin.FNNXFunction;
import plugin.FNNXObject;
import scripting.FNNXInterpreter;
import scripting.FNNXLexer;
import scripting.FNNXParser;
import scripting.InterThreadMonitor;

/**
 *
 * @author hsatizab
 */
public class FennixCore {
    
//    public FennixCore(java.io.InputStream in) {
//        console = new Scanner(in);
//        running = false;
//        objects = new Hashtable<String, Object>(50);
//        localObjects = new Hashtable<String, Object>(20);
//        plugins = new Hashtable<String, Class>(75);
//        monitor = new InterThreadMonitor(0);
//        monitor.notRunning();
//
//        toDecode = "";
//    }
    
    public FennixCore() {
        running = true;
        variables = new Hashtable<String, Object>(50);
        localVariables = new Hashtable<String, Object>(20);
        plugins = new Hashtable<String, Class>(100);
        monitor = new InterThreadMonitor(0);
        monitor.notRunning();

        toDecode = "";

    }

    public void runConsole(java.io.InputStream in) {
        Scanner console = new Scanner(in);
        String read;
        
        System.out.println("\nFENNIX launched in console mode.");
        System.out.println("You can enter a command.");
        System.out.println("Enter quit() to exit.");
        System.out.println();

        running = true;
        while (running) {
            System.out.print("> ");
            read = console.nextLine();
            this.decode(read, false);
        }
        
        console.close();
    }

    public void runScriptReader(BufferedReader reader, boolean parallel) {
        FNNXLexer lexer;
        FNNXParser parser = new FNNXParser();
        FNNXTreeNode tokenTree;

        lexer = new FNNXLexer(reader);
        tokenTree = lexer.analyze();

        parser.parse(tokenTree);
//        tokenTree.show();
//        System.out.println("----------------------------------------------------");
        parser.clean(tokenTree);
//        tokenTree.show();
        if (parser.checkErrors(tokenTree)) {
//            tokenTree.show();
            this.decodeTree(tokenTree, parallel);
        }
    }
    
//    public void runScriptFile(String fileName) {
//        String read;
//        
//        try {
//            BufferedReader reader = new BufferedReader(new FileReader(fileName));
//            while (reader.ready()) {
//                read = reader.readLine();
//                this.decode(read, false);
//            }
//            reader.close();
//        }
//        catch (FileNotFoundException fe) {
//            monitor.appendErrorMessage("Error. The file: \"" + fileName + "\" was not found.");
//        }
//        catch (IOException ie) {
//            monitor.appendErrorMessage("Input/Ouptput error.");
//        }
//    }
    
    public void decode(String newLine, boolean parallel) {

        if  (newLine.length() != 0) {

            if (newLine.compareTo("\\") == 0) {
                toDecode = "";
                System.out.println("Escaping...");
                return;
            }

            if (newLine.endsWith("...")) {
                toDecode += newLine.substring(0, newLine.length()-3);
                return;
            }
            else {
                toDecode += newLine;
            }

            if (toDecode.compareTo("quit()") == 0) {
                monitor.abort();
                running = false;
                toDecode = "";
                return;
            }

            if (toDecode.compareTo("pause()") == 0) {
                monitor.pause();
                toDecode = "";
                return;
            }

            if (toDecode.compareTo("resume()") == 0) {
                monitor.resume();
                toDecode = "";
                return;
            }

            if (toDecode.compareTo("stop()") == 0) {
                monitor.notRunning();
                toDecode = "";
                return;
            }
            
            FNNXLexer lexer;
            FNNXParser parser = new FNNXParser();
            FNNXTreeNode tokenTree;

            lexer = new FNNXLexer(new BufferedReader(new StringReader(toDecode)));
            tokenTree = lexer.analyze();

            parser.parse(tokenTree);
            parser.clean(tokenTree);
//            tokenTree.show();
            if (parser.checkErrors(tokenTree)) {
                this.decodeTree(tokenTree, parallel);
            }

            toDecode = "";

        }
    }

    public void quit() {
        monitor.notRunning();
        running = false;
    }

    public boolean isRunning() {
        return running;
    }

    public InterThreadMonitor getMonitor() {
        return monitor;
    }

    public Hashtable<String, Object> getVariables() {
        return variables;
    }

    public Hashtable<String, Class> getPlugins() {
        return plugins;
    }

    public void decodeTree(FNNXTreeNode tree, boolean parallel) {

        FNNXInterpreter interpreter = new FNNXInterpreter(tree, variables, localVariables, plugins, monitor);
        if (parallel) {
            interpreter.start();
        }
        else {
            interpreter.run();
        }

    }
    
    public Hashtable<String, Boolean> loadInitTools() {
        String line = null;
        BufferedReader reader = null;
        File file = null;
        Hashtable<String, Boolean> toReturn = new Hashtable<String, Boolean>(100);
        
        System.out.println("Starting initialization...");
        
        try {
//            reader = new BufferedReader(new FileReader(new File("/home/hsatizab/Desktop/work/010/FENNIX_010/tools.fcg")));
            reader = new BufferedReader(new FileReader(new File("tools.fcg")));
            while (reader.ready()) {
                line = reader.readLine();
                file = new File(line);
                toReturn.putAll(this.loadTool(file));
            }
            reader.close();
        }
        catch (FileNotFoundException fe) {
            System.out.println("Error. File not found.");
        }
        catch (IOException ioe) {
            System.out.println("Error. Corrupted file.");
        }

//        System.load("/home/hsatizab/Desktop/work/009/FNNXTestFunction/build/classes/toolFNNXTestFunction/FNNXTestFunction.class");
//        System.load("/home/hsatizab/Desktop/work/009/FNNXTestFunction/dist/FNNXTestFunction.jar");
                
        System.out.println("Initialization completed...");
        System.out.println();
        return toReturn;

    }

    public Hashtable<String, Boolean> loadTool(File file) {
        Class tool;
        Class[] interfaces;
        int i;
        String className;
        boolean object;
        boolean function;
        Object alreadyInPlugins;
        
        Hashtable<String, Boolean> toReturn = new Hashtable<String, Boolean>(2);

        //String name = file.getName();

        if (file.getName().endsWith(".jar")) {
            try {
                JarFile jar = new JarFile(file);
                Enumeration jare = jar.entries();
                URL[] url = {file.toURI().toURL()};
                MyURLClassLoader loader = new MyURLClassLoader(plugins, url);

                while (jare.hasMoreElements()) {
                    className = jare.nextElement().toString();
                    i = className.lastIndexOf('.');
                    if ((i > 0) && (i < className.length())) {
                        className = className.substring(0, i);
                    }
                    className = className.replace('/', '.');

//                    System.out.println(className);

                    if (className.endsWith(".")) {
                        continue;
                    }
                    
                    try {
                        tool = loader.loadClass(className);
                        interfaces = tool.getInterfaces();
                        object = false;
                        function  = false;
                        for (i = 0; i < interfaces.length; i++) {
                            if (interfaces[i].equals(FNNXObject.class)) {
                                object = true;
                            }
                            if (interfaces[i].equals(FNNXFunction.class)) {
                                function = true;
                            }
                        }

                        alreadyInPlugins = plugins.put(tool.getName(), tool);
                        if (object) {
                            if (alreadyInPlugins == null) {
                                toReturn.put(tool.getSimpleName(), function);
                            }
                            plugins.put(tool.getSimpleName(), tool);
                            System.out.println("FNNX object \"" + tool.getName() + "\" loaded.");
                        }
//                        else {
//                            System.out.println("\tObject \"" + tool.getName() + "\" loaded.");
//                        }

                    }
                    catch (ClassNotFoundException e) {
                    }
                    catch (NoClassDefFoundError ce) {
                        System.out.println("Error. Class \"" + className + "\" undefined.");
                    }
                }

            }
            catch (MalformedURLException me) {
                System.out.println("Error. File URL malformed.");
            }
            catch (IOException ioe) {
                System.out.println("Error. File \"" + file.getName() + "\" not found.");
            }
        }
        else {
            System.out.println("Error. Not a valid .jar file.");
        }

        return toReturn;
    }


//    private Scanner console;
    private boolean running;

    private InterThreadMonitor monitor;

    private Hashtable<String, Object> variables;
    private Hashtable<String, Object> localVariables;
    private Hashtable<String, Class> plugins;

    private String toDecode;

}
