/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXDatasetLoader;

import data.DataBlock;
import data.MyLinkedList;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Hashtable;
import java.util.regex.Pattern;
import javax.swing.JFrame;
import plugin.AbstractFNNXFunction;
import plugin.FNNXFunction;
import plugin.FNNXObject;
import scripting.InterThreadMonitor;
import scripting.Names;
import visualFNNXDatasetLoader.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXDatasetLoader extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXDatasetLoader(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[2];
        parameterNames[0] = "file";
        parameterNames[1] = "format";

        parameterTypes = new Class[2];
        parameterTypes[0] = String.class;
        parameterTypes[1] = String.class;

        isParameterMandatory = new boolean[2];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;

        parameters = new Object[2];
    }

    public FNNXDatasetLoader(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[2];
        parameterNames[0] = "file";
        parameterNames[1] = "format";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        fileName = (String)parameters[i];
                        break;
                    case 1:
                        fileFormat = Names.fileFormats.valueOf((String)parameters[i]);
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        BufferedReader reader;
        blocks = new MyLinkedList();

        switch (fileFormat) {
            case FENNIX_005:
                try {
                    reader = new BufferedReader(new FileReader(fileName));
                    if (!readFile_005(reader)) {
                        monitor.appendErrorMessage("Error. Problems while loading file: \"" + fileName + "\".");
                        return null;
                    }
                }
                catch (FileNotFoundException fe) {
                    monitor.appendErrorMessage("Error. File not found.");
                    return null;
                }
                break;

            case FENNIX_008:
                try {
                    reader = new BufferedReader(new FileReader(fileName));
                    if (!readFile_008(reader)) {
                        monitor.appendErrorMessage("Error. Problems while loading file: \"" + fileName + "\".");
                        return null;
                    }
                }
                catch (FileNotFoundException fe) {
                    monitor.appendErrorMessage("Error. File not found.");
                    return null;
                }
                break;
                
            default:
                monitor.appendErrorMessage("Error. Format not recognized.");
                return null;
        }

        if (blocks.size() > 1) {
            DataBlock[] blockVector = new DataBlock[blocks.size()];
            int i = 0;
            blocks.head();
            while (blocks.hasNode()) {
                blockVector[i] = (DataBlock)blocks.getDoNext();
                i++;
            }
            return blockVector;
        }
        else {
            if (blocks.size() > 0) {
                return (DataBlock)blocks.getFirst();
            }
            else {
                monitor.appendErrorMessage("Error. Impossible to decode file.");
                return null;
            }
        }
    }

    private boolean readFile_008(BufferedReader file) {
        String line = new String();
        int k;
        int p;
        int c;
        int r = 0;
        int nR = 0;
        int nP = 0;
        int nC = 0;
        int[] columns = null;
        Names.readFileCommand code;
        float[] data = null;
        String[] dataString;
        args = null;

        int lineCounter = 1;
//------------------------------------------------------------------------------
        try {
            if (file.ready()) {
                line = file.readLine();
                if (line.compareTo("FENNIX") != 0) {   //file pattern mark
                    monitor.appendErrorMessage("Error. Header does not contain a file pattern mark");
                    return false;
                }
            }
            else {
                monitor.appendErrorMessage("Error. Incomplete file");
                return false;
            }
        }
        catch (IOException ex) {
            monitor.appendErrorMessage("Error opening file");
            return false;
        }
//------------------------------------------------------------------------------
        try {
            while (file.ready()) {
//------------------------------------------------------------------------------
                line = file.readLine().trim();
                lineCounter++;

                if (line.length() == 0) {
                    continue;                   //skip empty lines
                }
//------------------------------------------------------------------------------

                code = decode(line);
                switch (code) {
                    case comment:
                        continue;

                    case newBlock:
                        if (r < nR) {
                            monitor.appendErrorMessage("Error. There were not enough data in the last block. Line " + String.valueOf(lineCounter));
                            return false;
                        }

                        if (args.length < 2) {
                            monitor.appendErrorMessage("Error. Not enough parameters for command \"newBlock\", at line " + lineCounter);
                            return false;
                        }
                        r = 0;
                        try {
                            nR = Integer.valueOf(args[0].trim()).intValue();
                            nP = args.length - 1;
                        }
                        catch (NumberFormatException e) {
                            monitor.appendErrorMessage("Error. Wrong format for first parameter (number of lines) in command \"newBlock\", at line " + lineCounter);
                            return false;
                        }
                        blocks.append(new DataBlock(nP));

                        nC = 0;
                        columns = new int[nP];
                        try {
                            for (p = 0; p < nP; p++) {
                                columns[p] = Integer.valueOf(args[p+1].trim()).intValue();
                                nC += columns[p];
                                ((DataBlock)blocks.getLast()).configDataset(p, nR, columns[p]);
                            }
                        }
                        catch (NumberFormatException e) {
                            monitor.appendErrorMessage("Error. Wrong format for parameter (number of columns) in command: \"newBlock\", at line " + lineCounter);
                            return false;
                        }
                        continue;

                    case unknownCommand:
                        monitor.appendErrorMessage("Error. Unknown command, at line " + lineCounter);
                        return false;

                    case data:
                    default:
                        break;
                }
//------------------------------------------------------------------------------

                if (r >= nR) {
                    monitor.appendErrorMessage("Error. Too many data in block. Line " + String.valueOf(lineCounter));
                    return false;
                }

                dataString = Pattern.compile("[\t,;]").split(line);

                if (dataString.length != nC) {
                    monitor.appendErrorMessage("Error. Wrong number of columns in line " + String.valueOf(lineCounter));
                    return false;
                }

                k = 0;
                for (p = 0; p < nP; p++) {
                    data = ((DataBlock)blocks.getLast()).getDataset(p).getRow(r);
                    for (c = 0; c < columns[p]; c++) {
                        try {
                            data[c] = Float.valueOf(dataString[k]).floatValue();
                        }
                        catch (NumberFormatException e) {
                            data[c] = Float.NaN;
                            System.out.println("Warning. Missing data at line " + String.valueOf(lineCounter));
                        }
                        k++;
                    }
                }
                r++;

            }
            file.close();
        }
        catch (IOException ex) {
            monitor.appendErrorMessage("Error reading file at line " + String.valueOf(lineCounter));
            return false;
        }

//------------------------------------------------------------------------------
        blocks.head();
        while (blocks.hasNode()) {
            ((DataBlock)blocks.getDoNext()).calcStats();
        }

        return true;
//------------------------------------------------------------------------------
    }

    private Names.readFileCommand decode(String l) {
        int p1, p2;
        String a;

        if ( (l.charAt(0) == '/') && (l.charAt(1) == '/') ) {   //comment
            return Names.readFileCommand.comment;
        }

        if ( (l.charAt(0) == '-') && (l.charAt(1) == '>') ) {   //command
            try {
                p1 = l.indexOf('(');
                p2 = l.indexOf(')');
                a = l.substring(p1+1, p2);
//                System.out.println(a);
                args = Pattern.compile("[,]").split(a);
            }
            catch (IndexOutOfBoundsException iobe) {
                return Names.readFileCommand.badFormat;
            }

            if (l.startsWith("->newBlock")) {
                return Names.readFileCommand.newBlock;
            }

            return Names.readFileCommand.unknownCommand;
        }

        return Names.readFileCommand.data;                      //data
    }

    private boolean readFile_005(BufferedReader file) {
        String line = new String();
        int nP;
        int nI;
        int nO;
        DataBlock block;
        
        int k;
        String[] dataString;
        float[] data;
//------------------------------------------------------------------------------
        try {
            if (file.ready()) {
                line = file.readLine();
                if (line.compareTo("FENNIX") != 0) {   //file pattern mark
                    monitor.appendErrorMessage("Error. Header does not contain a file pattern mark");
                    return false;
                }
            }
            else {
                monitor.appendErrorMessage("Error. Incomplete file");
                return false;
            }
        }
        catch (IOException ex) {
            ex.printStackTrace();
            monitor.appendErrorMessage("Error opening file");
            return false;
        }
//------------------------------------------------------------------------------
        if(!this.readHeader("#PAT:", file)) {
            monitor.appendErrorMessage("Error. Header does not contain the number of examples in the file");
            return false;
        }
        nP = temp;
//------------------------------------------------------------------------------
        if(!this.readHeader("#INPUT:", file)) {
            monitor.appendErrorMessage("Error. Header does not contain the number of inputs in the file");
            return false;
        }
        nI = temp;
//------------------------------------------------------------------------------
        if(!this.readHeader("#OUTPUT:", file)) {
            monitor.appendErrorMessage("Error. Header does not contain the number of outputs in the file");
            return false;
        }
        nO = temp;
//------------------------------------------------------------------------------
        block = new DataBlock(2);
        block.configDataset(0, nP, nI);
        block.configDataset(1, nP, nO);
//------------------------------------------------------------------------------

        int lineCounter = 0;
        try {
            while (file.ready()) {
                line = file.readLine().trim();

                if (line.length() == 0) {
                    continue;
                }

                dataString = Pattern.compile("[\t,;]").split(line);

                if (dataString.length > (nI+nO)) {
                    monitor.appendErrorMessage("Error. Example " + String.valueOf(lineCounter+1) + " contains too much data");
                    return false;         //there is too much data
                }
                try {
                    data = block.getDataset(0).getRow(lineCounter);
                    for (k = 0; k < nI; k++) {
                        try {
                            data[k] = Float.valueOf(dataString[k]).floatValue();
                        }
                        catch (NumberFormatException e) {
                            data[k] = Float.NaN;
                        }
                    }
                    data = block.getDataset(1).getRow(lineCounter);
                    for (k = 0; k < nO; k++) {
                        try {
                            data[k] = Float.valueOf(dataString[k+nI]).floatValue();
                        }
                        catch (NumberFormatException e) {
                            data[k] = Float.NaN;
                        }
                    }
                }
                catch (NullPointerException e) {
                    monitor.appendErrorMessage("Error. Example " + String.valueOf(lineCounter+1) + " does not have enough data");
                    return false;         //there is not enough data in the file
                }

                if (lineCounter >= nP) {
                    monitor.appendErrorMessage("Error. Too many observations in file");
                    return false;
                }

                lineCounter++;
            }
        }
        catch (IOException ex) {
            monitor.appendErrorMessage("Error reading example " + String.valueOf(lineCounter+1));
            return false;
        }

//------------------------------------------------------------------------------
        if  (nP != lineCounter) {
            monitor.appendErrorMessage("Error. The number of patterns does not correspond with the actual number of examples");
            return false;
        }

        block.calcStats();
        blocks.append(block);
        return true;
//------------------------------------------------------------------------------
    }

    /**
     * Reads the header labels.
     *
     * @param   mark    the label to be detected.
     * @param   f   the file {@link Scanner}.
     * @return  true, if the specified label is found; false, otherwise.
     * @see #readFile(Scanner)
     */
    private boolean readHeader(String mark, BufferedReader f) {
        String line;
        String[] words;

        try {
            if (f.ready()) {
                line = f.readLine();
                words = line.split(" ");

                if (words.length < 2) {
                    return false;
                }

                if (words[0].trim().compareToIgnoreCase(mark) != 0) {
                    return false;
                }

                try {
                    temp = Integer.valueOf(words[1]).intValue();
                }
                catch (NumberFormatException e) {
                    return false;
                }
            }
            else {
                return false;
            }
        }
        catch (IOException ex) {
            return false;
        }
        return true;
    }

    private String fileName;
    private Names.fileFormats fileFormat;
    private MyLinkedList blocks;

    private String[] args;

    private int temp;
}
