/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXNormalizer;

import plugin.*;
import visualFNNXNormalizer.ConfigurationWindow;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import scripting.InterThreadMonitor;
import scripting.Names;

/**
 *
 * @author hsatizab
 */
public class FNNXNormalizer extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXNormalizer(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[4];
        parameterNames[0] = "dataset";
        parameterNames[1] = "method";
        parameterNames[2] = "p1";
        parameterNames[3] = "p2";

        parameterTypes = new Class[4];
        parameterTypes[0] = DataBlock.class;
        parameterTypes[1] = String.class;
        parameterTypes[2] = Number.class;
        parameterTypes[3] = Number.class;

        isParameterMandatory = new boolean[4];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = true;

        parameters = new Object[4];

    }

    public FNNXNormalizer(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[4];
        parameterNames[0] = "dataset";
        parameterNames[1] = "method";
        parameterNames[2] = "p1";
        parameterNames[3] = "p2";

    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 1:
                        method = NormalizerMethod.Name.valueOf((String)parameters[i]);
                        break;
                    case 2:
                        p1 = ((Number)parameters[i]).floatValue();
                        break;
                    case 3:
                        p2 = ((Number)parameters[i]).floatValue();
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        try {
            switch (method) {
                case MIN_MAX:
                    if (p1 >= p2) {
                        monitor.appendErrorMessage("Error. Parameter \"p2\" must be larger than \"p1\"");
                        return null;
                    }
                    for (int i = 0; i < dataset.getNParts(); i++) {
                        dataset.getDataset(i).calcStats();
                        dataset.getDataset(i).normalizeMaxMin(p1, p2);
                        dataset.getDataset(i).calcStats();
                    }
                    break;

                case AVG_RANGE:
                    if (p2 <= 0) {
                        monitor.appendErrorMessage("Error. Parameter \"p2\" must be greater than 0");
                        return null;
                    }
                    for (int i = 0; i < dataset.getNParts(); i++) {
                        dataset.getDataset(i).calcStats();
                        dataset.getDataset(i).normalizeAvgRange(p1, p2);
                        dataset.getDataset(i).calcStats();
                    }
                    break;
                case PLUS_MULT:
                    for (int i = 0; i < dataset.getNParts(); i++) {
                        dataset.getDataset(i).calcStats();
                        dataset.getDataset(i).normalizePlusMult(p1, p2);
                        dataset.getDataset(i).calcStats();
                    }
                    break;
                case CLIP_MIN_MAX:
                    if (p1 > p2) {
                        monitor.appendErrorMessage("Error. Parameter \"p2\" must be greater than parameter \"p1\"");
                        return null;
                    }
                    for (int i = 0; i < dataset.getNParts(); i++) {
                        dataset.getDataset(i).normalizeClipMinMax(p1, p2);
                        dataset.getDataset(i).calcStats();
                    }
                    break;
                default:
                    return null;
            }
        }
        catch (IllegalArgumentException ia) {
            monitor.appendErrorMessage("Error. Unable to decode method");
            return null;
        }
        return dataset;
    }

    private DataBlock dataset;
    private NormalizerMethod.Name method;
    private float p1;
    private float p2;

}
