/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXSampler;

import visualFNNXSampler.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import java.util.Random;
import javax.swing.JFrame;
import scripting.InterThreadMonitor;
import scripting.Names;

/**
 *
 * @author hsatizab
 */
public class FNNXSampler extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXSampler(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[2];
        parameterNames[0] = "dataset";
        parameterNames[1] = "percent";

        parameterTypes = new Class[2];
        parameterTypes[0] = DataBlock.class;
        parameterTypes[1] = Number.class;

        isParameterMandatory = new boolean[2];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;

        parameters = new Object[2];
    }

    public FNNXSampler(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[2];
        parameterNames[0] = "dataset";
        parameterNames[1] = "percent";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        try {
            dataset = (DataBlock)parameters[0];
            percent = ((Number)parameters[1]).intValue();

            if ((percent  <= 0) || (percent >= 100)) {
                monitor.appendErrorMessage("Error. Sampling percent must be higher than 0%, and smaller than 100%.");
                return false;
            }
        }
        catch(NullPointerException ne) {
            monitor.appendErrorMessage("Internal error. Null pointer.");
            return false;
        }
        catch(IndexOutOfBoundsException ie) {
            monitor.appendErrorMessage("Internal error. Index out of bounds.");
            return false;
        }
        catch(ClassCastException ce) {
            monitor.appendErrorMessage("Internal error. Class cast.");
            return false;
        }

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        Random rand = new Random();
        DataBlock[] blocks = new DataBlock[2];
        blocks[0] = new DataBlock(dataset.getNParts());
        blocks[1] = new DataBlock(dataset.getNParts());

        int total = dataset.getDataset(0).getNRows();
        int sampled = (int)(total*((float)percent/100));
        int i, s, i1, i2;
        boolean[] sampledFlags = new boolean[total];

        if (sampled > total) {
            monitor.appendErrorMessage("Error. Sampling percent must be lesser than 100%.");
            return null;
        }

        for (i = 0; i < dataset.getNParts(); i++) {
            blocks[0].configDataset(i, sampled, dataset.getDataset(i).getNColumns());
            blocks[1].configDataset(i, total-sampled, dataset.getDataset(i).getNColumns());
        }

        for (i = 0; i < sampled; i++) {
            s = rand.nextInt(total);
            while (sampledFlags[s]) {
                s++;
                if (s == total) {
                    s = 0;
                }
            }
            sampledFlags[s] = true;
        }

        for (s = 0; s < dataset.getNParts(); s++) {
            i1 = 0;
            i2 = 0;
            for (i = 0; i < total; i++) {
                if (sampledFlags[i]) {
                    blocks[0].getDataset(s).setRow(i1, dataset.getDataset(s).getRow(i), false);
                    i1++;
                }
                else {
                    blocks[1].getDataset(s).setRow(i2, dataset.getDataset(s).getRow(i), false);
                    i2++;
                }
            }
        }

        blocks[0].copyStats(dataset);
        blocks[1].copyStats(dataset);

        return blocks;

    }


    private DataBlock dataset;
    private int percent;

}
