/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXSplitter;

import visualFNNXSplitter.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import java.util.Random;
import javax.swing.JFrame;
import scripting.InterThreadMonitor;
import scripting.Names;

/**
 *
 * @author hsatizab
 */
public class FNNXSplitter extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXSplitter(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[2];
        parameterNames[0] = "dataset";
        parameterNames[1] = "parts";

        parameterTypes = new Class[2];
        parameterTypes[0] = DataBlock.class;
        parameterTypes[1] = Number.class;

        isParameterMandatory = new boolean[2];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;

        parameters = new Object[2];
    }

    public FNNXSplitter(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[2];
        parameterNames[0] = "dataset";
        parameterNames[1] = "parts";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        try {
            dataset = (DataBlock)parameters[0];
            parts = ((Number)parameters[1]).intValue();

            if ((parts < 2) || (parts > dataset.getNRows())){
                monitor.appendErrorMessage("Error. Parts must be a positive number, smaller than the numbers of rows of the dataset.");
                return false;
            }
        }
        catch(NullPointerException ne) {
            monitor.appendErrorMessage("Internal error. Null pointer.");
            return false;
        }
        catch(IndexOutOfBoundsException ie) {
            monitor.appendErrorMessage("Internal error. Index out of bounds.");
            return false;
        }
        catch(ClassCastException ce) {
            monitor.appendErrorMessage("Internal error. Class cast.");
            return false;
        }

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        Random rand = new Random();
        DataBlock[] blocks = new DataBlock[parts];
        int total = dataset.getDataset(0).getNRows();
        int residual = total%parts;
        int[] sizes = new int[parts];
        int[] indexes = new int[parts];
        int i, j, s;
        boolean[] sampledFlags = new boolean[total];

        for (i = 0; i < parts; i++) {
            sizes[i] = (int)Math.floor(total/parts);
            indexes[i] = 0;
        }
        i = 0;
        while (residual > 0) {
            sizes[i]++;
            residual--;
            i++;
        }

        for (i = 0; i < parts; i++) {
            blocks[i] = new DataBlock(dataset.getNParts());
            for (j = 0; j < dataset.getNParts(); j++) {
                blocks[i].configDataset(j, sizes[i], dataset.getDataset(j).getNColumns());
            }

            while (indexes[i] < sizes[i]) {
                s = rand.nextInt(total);
                while (sampledFlags[s]) {
                    s++;
                    if (s == total) {
                        s = 0;
                    }
                }
                sampledFlags[s] = true;

                for (j = 0; j < dataset.getNParts(); j++) {
                    blocks[i].getDataset(j).setRow(indexes[i], dataset.getDataset(j).getRow(s), false);
                }
                indexes[i]++;
            }
            
            blocks[i].copyStats(dataset);
        }

        return blocks;

    }

    private DataBlock dataset;
    private int parts;

}
