/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package data;

import java.io.Serializable;
import java.util.Arrays;

/**
 *
 * @author hsatizab
 */
public class DataBlock implements Cloneable, Serializable {

    public DataBlock(int np) {
        nParts = np;
        parts = new DataMatrix[nParts];
    }
    
    public DataBlock(DataMatrix d) {
        nParts = 1;
        parts = new DataMatrix[nParts];
        parts[0] = d;
    }

    public DataBlock clone() throws CloneNotSupportedException {
        return (DataBlock)super.clone();
    }
    
    public int getNParts() {
        return nParts;
    }
    
    public int getNRows() {
        return parts[0].getNRows();
    }

    public void configDataset(int n, int nr, int nc) {
        if ( (n < 0) || (n > nParts) )
            return;
        
        parts[n] = new DataMatrix(nr, nc);
    }
    
    public DataMatrix getDataset(int n) {
        if ( (n < 0) || (n > nParts) )
            return null;

        return parts[n];
    }
    
    public void calcStats() {
        for (int i = 0; i < nParts; i++) {
            parts[i].calcStats();
        }
    }

    public void copyStats(DataBlock source) {
        for (int i = 0; (i < nParts) && (i < source.nParts); i++) {
            getDataset(i).copyStats(source.getDataset(i));
        }
    }

//    public DataFrame sample(int percent, boolean copy) {
//        Random rand = new Random();
//        DataBlock block1 = new DataBlock(nParts);
//        DataBlock block2 = new DataBlock(nParts);
//
//        int total = getDataset(0).getNRows();
//        int sampled = (int)(total*((float)percent/100));
//        int i, s, i1, i2;
//        boolean[] sampledFlags = new boolean[total];
//
//        if (sampled > total) return null;
//
//        for (i = 0; i < nParts; i++) {
//            block1.configDataset(i, sampled, getDataset(i).getNColumns());
//            block2.configDataset(i, total-sampled, getDataset(i).getNColumns());
//        }
//
//        for (i = 0; i < sampled; i++) {
//            s = rand.nextInt(total);
//            while (sampledFlags[s]) {
//                s++;
//                if (s == total) {
//                    s = 0;
//                }
//            }
//            sampledFlags[s] = true;
//        }
//
//        for (s = 0; s < nParts; s++) {
//            i1 = 0;
//            i2 = 0;
//            for (i = 0; i < total; i++) {
//                if (sampledFlags[i]) {
//                    block1.getDataset(s).setRow(i1, this.getDataset(s).getRow(i), copy);
//                    i1++;
//                }
//                else {
//                    block2.getDataset(s).setRow(i2, this.getDataset(s).getRow(i), copy);
//                    i2++;
//                }
//            }
//        }
//
//        block1.copyStats(this);
//        block2.copyStats(this);
//        DataFrame frame = new DataFrame();
//        frame.appendDataBlock(block1);
//        frame.appendDataBlock(block2);
//
//        return frame;
//    }

    public String toString() {
//        String toReturn = "DataBlock\n";
//
//        for (int i = 0; i < nParts; i++) {
//            toReturn += parts[i].toString();
//        }

        String toReturn = "";
        for (int i = 0; i < nParts; i++) {
            toReturn += "\nDataBlock: [" + parts[i].getNRows() + "x" + parts[i].getNColumns() + "]";
//            if (i != (nParts-1)) {
//                toReturn += ", ";
//            }
            toReturn += "\nMax = " + Arrays.toString(parts[i].getMax());
            toReturn += "\nMean = " + Arrays.toString(parts[i].getMean());
            toReturn += "\nMin = " + Arrays.toString(parts[i].getMin());
            toReturn += "\nRange = " + Arrays.toString(parts[i].getRange());
            toReturn += "\nMSStot = " + Arrays.toString(parts[i].getMSStot()) + "\n";
        }

        return toReturn;
    }
    
    private int nParts;
    private DataMatrix[] parts;
}
