/*
 * MyLinkedList.java
 *
 * Created on 18. juin 2006, 18:57
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package data;

import java.io.Serializable;


/**
 *
 * @author Hector
 */
public class MyLinkedList implements Serializable, Cloneable {
    
    /**
     * Creates a new instance of MyLinkedList.
     *
     */
    public MyLinkedList() {
        nNodes = 0;
        first = null;
        last = null;
        current = null;
        temp = null;
    }

    /**
     * Returns the actual state of the actual list window. Used to detect the end of the list.
     *
     * @return  true, if there is a {@link Node} in the list window, false if this is the end of the list.
     * @see #getDoNext()
     * @see #getDoPrevious()
     * @see #next()
     * @see #previous
     */
    public MyLinkedList clone() {
//        System.out.println("Clonning MyLinkedList");
        MyLinkedList tempc = null;
        try {
            tempc = (MyLinkedList) super.clone();
            tempc.nNodes = 0;
            tempc.current = null;
            tempc.first = null;
            tempc.last = null;
        }
        catch (CloneNotSupportedException ex) {
            ex.printStackTrace();
        }

        return tempc;
    }
    
    /**
     * Returns the state of the list window. Used to detect the end of the list.
     *
     * @return  true, if there is a {@link Node} in the list window, false if this is the end of the list.
     * @see #getDoNext()
     * @see #getDoPrevious()
     * @see #next()
     * @see #previous
     */
    public final boolean hasNode() {
        if (current == null) {
            return false;
        }
        else {
            return true;
        }
    }
    
    /**
     * Returns the state of the previous element of the list window. Used to detect the first element of the list.
     *
     * @return  true, if there is a {@link Node} in the previous element of the list window, false if this is the first element of the list.
     * @see #hasNode()
     * @see #nextHasNode()
     * @see #getDoNext()
     * @see #getDoPrevious()
     * @see #next()
     * @see #previous
     */
    public final boolean previousHasNode() {
        if (current == null) {
            return false;
        }

        if (current.previous == null) {
            return false;
        }
        else {
            return true;
        }
    }

    /**
     * Returns the state of the next element of the list window. Used to detect the last element of the list.
     *
     * @return  true, if there is a {@link Node} in the next element of the list window, false if this is the last element of the list.
     * @see #hasNode()
     * @see #previousHasNode()
     * @see #getDoNext()
     * @see #getDoPrevious()
     * @see #next()
     * @see #previous
     */
    public final boolean nextHasNode() {
        if (current == null) {
            return false;
        }

        if (current.next == null) {
            return false;
        }
        else {
            return true;
        }
    }

    /**
     * Points the list window to the first element. Used to begin iterations over the list.
     *
     * @see #tail()
     */
    public final void head() {
        current = first;
    }

    /**
     * Points the list window to the last element. Used to begin iterations over the list.
     *
     * @see #head()
     */
    public final void tail() {
        current = last;
    }

    /**
     * Adds an object to the end of the list. 
     *
     * @param   o   the {@link Object} to be append to the list.
     * @see #insert(Object)
     */
    public void append(Object o) {
        Node node = new Node(this.last, null, o);
        if (nNodes == 0) {
            first = node;
            current = node;
        }
        else {
            last.next = node;
        }
        last = node;
        nNodes++;
    }

    /**
     * Adds an object before the current position of the list window.
     *
     * @param   o   the {@link Object} to be added to the list.
     * @see #add(Object)
     */
    public void insert(Object o) {
        if (nNodes == 0) {
            Node node = new Node(null, null, o);
            last = node;
            first = node;
            current = node;
        }
        else {
            Node node = new Node(current.previous, current, o);
            if (current != first) {
                current.previous.next = node;
            }
            else {
                first = node;
            }
            current.previous = node;
        }
        nNodes++;
    }

    /**
     * Return the object pointed by the list window and moves the window to the next element.
     *
     * @return  the data {@link Object} contained by the {@link Node} pointed by the window.
     * @see #getDoPrevious()
     * @see #next()
     * @see #previous
     */
    public final Object getDoNext() {
        temp = current;
        current = current.next;
        return temp.data;
    }
    
    /**
     * Return the object pointed by the list window and moves the window to the previous element. 
     *
     * @return  the data {@link Object} contained by the {@link Node} pointed by the window.
     * @see #getDoNext()
     * @see #next()
     * @see #previous
     */
    public final Object getDoPrevious() {
        temp = current;
        current = current.previous;
        return temp.data;
    }
    
    /**
     * Return the {@link Object} contained by the first element of the list.
     *
     * @return  the data {@link Object} contained by the first {@link Node} of the list.
     * @see #getLast()
     * @see #getCurrent()
     */
    public final Object getFirst() {
        return first.data;
    }
    
    /**
     * Return the {@link Object} contained by the last element of the list.
     *
     * @return  the data {@link Object} contained by the last {@link Node} of the list.
     * @see #getFirst()
     * @see #getCurrent()
     */
    public final Object getLast() {
        return last.data;
    }
    
    /**
     * Return the {@link Object} contained by the window of the list.
     *
     * @return  the data {@link Object} contained by the window {@link Node} of the list.
     * @see #getFirst()
     * @see #getLast()
     */
    public final Object getCurrent() {
        return current.data;
    }
    
    /**
     * Moves the window to the next element.
     *
     * @see #getDoNext()
     * @see #getDoPrevious()
     * @see #previous
     */
    public final void next() {
        current = current.next;
    }
    
    /**
     * Moves the window to the previous element.
     *
     * @see #getDoNext()
     * @see #getDoPrevious()
     * @see #next()
     */
    public final void previous() {
        current = current.previous;
    }

    /**
     * Return the number of {@link Node}s contained by the list.
     *
     * @return  an integer value indicating the number of nodes in the list.
     */
    public final int size() {
        return nNodes;
    }

    /**
     * Deletes the current {@link Node} in the list and goes to the next element.
     *
     * @see #deleteCurrentDoPrevious()
     * @see #deleteFirst()
     * @see #deleteLast()
     */
    public void deleteCurrentDoNext() {
        if (nNodes == 0) {
            return;
        }
        
        if (nNodes == 1) {
            first = null;
            last = null;
            current = null;
        }
        else {
            if (current == first) {
                current.next.previous = null;
                first = current.next;
            }
            else {
                if (current == last) {
                    current.previous.next = null;
                    last = current.previous;
                }
                else {
                    current.previous.next = current.next;
                    current.next.previous = current.previous;
                }
            }
            current = current.next;
        }

        nNodes--;
    }
    
    /**
     * Deletes the current {@link Node} in the list and goes to the previous element.
     *
     * @see #deleteCurrentDoNext()
     * @see #deleteFirst()
     * @see #deleteLast()
     */
    public void deleteCurrentDoPrevious() {
        if (nNodes == 0) {
            return;
        }

        if (nNodes == 1) {
            first = null;
            last = null;
            current = null;
        }
        else {
            if (current == first) {
                current.next.previous = null;
                first = current.next;
            }
            else {
                if (current == last) {
                    current.previous.next = null;
                    last = current.previous;
                }
                else {
                    current.previous.next = current.next;
                    current.next.previous = current.previous;
                }
            }
            current = current.previous;
        }

        nNodes--;
    }

    /**
     * Deletes the first {@link Node} in the list.
     *
     * @see #deleteLast()
     */
    public void deleteFirst() {
        if (nNodes == 0) {
            return;
        }
        
        if (nNodes == 1) {
            first = null;
            last = null;
            current = null;
        }
        else {
			if (current == first) {
				current = null;
			}
            first.next.previous = null;
            first = first.next;
        }

        nNodes--;
    }
    
    /**
     * Deletes the last {@link Node} in the list.
     *
     * @see #deleteFirst()
     */
    public void deleteLast() {
        if (nNodes == 0) {
            return;
        }
        
        if (nNodes == 1) {
            first = null;
            last = null;
            current = null;
        }
        else {
			if (current == last) {
				current = null;
			}
            last.previous.next = null;
            last = last.previous;
        }

        nNodes--;
    }
    
    /**
     * Deletes every {@link Node} in the list.
     *
     * @see #deleteLast()
     * @see #deleteFirst()
     */
    public void clear() {
        first = null;
        last = null;
        current = null;
        nNodes = 0;
    }
    
    public boolean isLast() {
        if (current.next == null) {
            return true;
        }
        else {
            return false;
        }
    }
    
    public Object getAt(int index) {
        Object object = null;
        this.head();
        for (int i = 0; this.hasNode(); i++) {
            if (index == i) {
                object = this.getCurrent();
                break;
            }
            else {
                this.next();
            }
        }
        return object;
    }

    public void debug() {
        System.out.println("Current: " + current.data.toString());
    }

    protected Node first;
    protected Node last;
    protected Node current;
    protected Node temp;
    protected int nNodes;
}
