/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package model;

import data.DataBlock;
import data.DataMatrix;

/**
 *
 * @author hsatizab
 */
public abstract class AbstractFNNXSupervisedModel extends AbstractFNNXUnsupervisedModel implements FNNXSupervisedModel {

    public boolean setDataset(DataBlock data) {
        if (data.getNParts() < 2) {
            System.out.println("Error. The dataset must contain two parts at least.");
            return false;
        }

        inputDataset = data.getDataset(0);
        outputDataset = data.getDataset(1);

        if ( (inputDataset == null) || (outputDataset == null) ) {
            inputDataset = null;
            outputDataset = null;
            System.out.println("Error. Empty dataset.");
            return false;
        }

        if ( (inputDataset.getNRows() < 1) || (outputDataset.getNRows() < 1) ) {
            inputDataset = null;
            outputDataset = null;
            System.out.println("Error. Empty dataset.");
            return false;
        }

        if ( (inputDataset.getNColumns() != nInputs) || (outputDataset.getNColumns() != nOutputs) ) {
            inputDataset = null;
            outputDataset = null;
            System.out.println("Error. Dimension mismatch.");
            return false;
        }

        currentInput = new float[nInputs];
        currentTarget = new float[nOutputs];
        output = new float[nOutputs];

        return true;
    }

    public boolean hasNextExampleSequential() {
        if (obsCounter < inputDataset.getNRows()) {
            this.setCurrentInput(inputDataset.getRow(obsCounter));
            this.setCurrentTarget(outputDataset.getRow(obsCounter));
            return true;
        }
        else {
            return false;
        }
    }

    public void nextExampleRandom() {
        obsCounter = gen.nextInt(inputDataset.getNRows());

        this.setCurrentInput(inputDataset.getRow(obsCounter));
        this.setCurrentTarget(outputDataset.getRow(obsCounter));
    }

    public void setCurrentTarget(float[] output) {
        System.arraycopy(output, 0, currentTarget, 0, nOutputs);
    }

    public int getAmountOfOutputs() {
        return nOutputs;
    }

    public float[] getCurrentTarget() {
        return currentTarget;
    }

    public DataMatrix getOutputDataset() {
        return outputDataset;
    }

    public void takeExampleFrom(int i, DataBlock data) {
        setCurrentInput(data.getDataset(0).getRow(i));
        setCurrentTarget(data.getDataset(1).getRow(i));
    }

    protected int nOutputs;
    protected float[] output;
    protected DataMatrix outputDataset;
    protected float[] currentTarget;

}
