/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package model;

import data.DataBlock;
import data.DataMatrix;

/**
 *
 * @author hsatizab
 */
public abstract class AbstractFNNXUnsupervisedModel extends AbstractFNNXModel implements FNNXUnsupervisedModel {

    public boolean setDataset(DataBlock data) {
        if (data.getNParts() < 1) {
            System.out.println("Error. The dataset must contain one part at least.");
            return false;
        }

        inputDataset = data.getDataset(0);

        if (inputDataset == null) {
            inputDataset = null;
            System.out.println("Error. Empty dataset.");
            return false;
        }

        if (inputDataset.getNRows() < 1) {
            inputDataset = null;
            System.out.println("Error. Empty dataset.");
            return false;
        }

        if (inputDataset.getNColumns() != nInputs) {
            inputDataset = null;
            System.out.println("Error. Dimension mismatch.");
            return false;
        }

        currentInput = new float[nInputs];

        return true;
    }

    public boolean setDataset(DataMatrix data) {

        inputDataset = data;

        if (inputDataset == null) {
            inputDataset = null;
            System.err.println("Error. Empty dataset.");
            return false;
        }

        if (inputDataset.getNRows() < 1) {
            inputDataset = null;
            System.err.println("Error. Empty dataset.");
            return false;
        }

        if (inputDataset.getNColumns() != nInputs) {
            inputDataset = null;
            System.err.println("Error. Dimension mismatch.");
            return false;
        }

        currentInput = new float[nInputs];

        return true;
    }

    public boolean hasNextExampleSequential() {
        if (obsCounter < inputDataset.getNRows()) {
            this.setCurrentInput(inputDataset.getRow(obsCounter));
            return true;
        }
        else {
            return false;
        }
    }

    public void nextExampleRandom() {
        obsCounter = gen.nextInt(inputDataset.getNRows());

        this.setCurrentInput(inputDataset.getRow(obsCounter));
    }

    public void setCurrentInput(float[] input) {
        System.arraycopy(input, 0, currentInput, 0, nInputs);
    }

    public float[] getCurrentInput() {
        return currentInput;
    }

    public int getAmountOfInputs() {
        return nInputs;
    }

    public void setInput(int i, float value) {
        currentInput[i] = value;
    }

    public DataMatrix getInputDataset() {
        return inputDataset;
    }

    public void takeExampleFrom(int i, DataBlock data) {
        setCurrentInput(data.getDataset(0).getRow(i));
    }

    protected int nInputs;
    protected DataMatrix inputDataset;
    protected float[] currentInput;


}
