/*
 * AbstractNeuron.java
 *
 * Created on 24 de septiembre de 2007, 13:52
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package model.distributedRepresentation;

import data.MyFloat;
import data.MyLinkedList;
import java.io.Serializable;
import java.util.Random;

/**
 *
 * @author Hector
 */
public abstract class AbstractPerceptronBase implements Serializable, Cloneable {
    
    /**
     * Creates a new instance of AbstractNeuron
     */
    public AbstractPerceptronBase() {
        backwardLinks = new MyLinkedList();
        forwardLinks = new MyLinkedList();
        use = true;

//        activation = 0;
//        neta = 0;
//        output = 0;
//        delta = 0;
    }
    
    /**
     * Returns a copy of the object.
     */
    public AbstractPerceptronBase clone() {
        AbstractPerceptronBase temp = null;
        try {
            temp = (AbstractPerceptronBase) super.clone();
            temp.backwardLinks = (MyLinkedList) this.backwardLinks.clone();
            temp.forwardLinks = (MyLinkedList) this.forwardLinks.clone();
        }
        catch (CloneNotSupportedException ex) {
            ex.printStackTrace();
        }
        return temp;
    }

    /**
     * Sets the value of the {@link #use} flag.
     *
     * @param   u   the new value of the flag
     */
    public void setUse(boolean u) {
        use = u;
    }
    
    /**
     * Returns the state of the {@link #use} flag.
     *
     * @return  true, if the unit is used, false otherwise
     */
    public boolean getUse() {
        return use;
    }
    
    /**
     * Returns the index of the unit.
     *
     * @return  the index of the unit
     */
    public int getNumber() {
        return number;
    }

    /**
     * Sets the value of the variable {@link #number}.
     *
     * @param   n   the new value
     */
    public void setNumber(int n) {
        number = n;
    }

    /**
     * Initializes the unit.
     *
     * @param   gen the random number generator to use in the initialization
     */
    public void initNeuron(Random gen) {
        Connector temp;
        float r = (float)(1 / Math.sqrt(backwardLinks.size()));
        
        backwardLinks.head();
        while (backwardLinks.hasNode()) {
            temp = (Connector) backwardLinks.getDoNext();
            temp.setWeight((float)(2 * (gen.nextFloat() - 0.5) * r));     // uniform distribution [-r, r]
        }
        reset();
    }

    public float getForwardWeightSum() {
        float sum = 0;
        forwardLinks.head();

        while (forwardLinks.hasNode()) {
            sum += (float)Math.abs(((Connector)forwardLinks.getDoNext()).getWeightValue());
        }

        return sum/forwardLinks.size();
    }

    /**
     * Copies the connections from a given unit.
     *
     * @param   toCopy  the unit from where the values are copied
     */
    public void copyConnections(AbstractPerceptronBase toCopy) {
        backwardLinks.head();
        toCopy.backwardLinks.head();
        while (backwardLinks.hasNode() && toCopy.backwardLinks.hasNode()) {
            ((Connector)backwardLinks.getDoNext()).setWeight(((Connector)toCopy.backwardLinks.getDoNext()).getWeightValue());
        }
    }

    /**
     * Calculates the neta of the unit.
     */
    public void calcNeta() {
        Connector temp;
        float input;
        this.neta = 0;

        if (use) {
            backwardLinks.head();
            while (backwardLinks.hasNode()) {                       //calculate neta
                temp = (Connector) backwardLinks.getDoNext();
                input = temp.getRefNeuron().output;
                neta += temp.getWeightValue() * input;
                temp.setInput(input);

//                System.err.print(this.number);
//                System.err.print(" -> ");
//                System.err.print(temp.getRefNeuron().number);
//                System.err.print(" : ");
//                System.err.println(input);

            }
        }
        else {
            neta = 0;
        }
    }

    /**
     * Calculates the output of the unit.
     */
    public void calcOutput() {
        activation = activationFunction.apply(this.neta);       //calculate activation
        output = outputFunction.apply(this.activation);         //calculate output
        d_output = activationFunction.firstDeriv(this.output);  //calculate output first derivative
    }

    public float getActivation() {
        return activation;
    }
    public float getNeta() {
        return neta;
    }

    /**
     * Returns the output of the unit.
     *
     * @return  the output value of the unit
     */
    public float getOutput() {
        return output;
    }

    /**
     * Moves the window of the list containing the forward connections to the first element.
     */
    public void firstForwardConnection() {
        try {
            forwardLinks.head();
        }
        catch (NullPointerException ex) {
        }
    }

    /**
     * Moves the window of the list containing the backward connections to the first element.
     */
    public void firstBackwardConnection() {
        try {
            backwardLinks.head();
        }
        catch (NullPointerException ex) {
        }
    }

    /**
     * Checks if there is a connector in the position of the window.
     */
    public boolean hasForwardConnection() {
        try {
            return forwardLinks.hasNode();
        }
        catch (NullPointerException ex) {
            return false;
        }
    }

    /**
     * Checks if there is a connector in the position of the window.
     */
    public boolean hasBackwardConnection() {
        try {
            return backwardLinks.hasNode();
        }
        catch (NullPointerException ex) {
            return false;
        }
    }

    /**
     * Returns the connector pointed by the window and moves the window to the next connector.
     *
     * @return  the connector currently in the window.
     */
    public Connector getForwardConnectorDoNext() {
        try {
            return (Connector)forwardLinks.getDoNext();
        }
        catch (NullPointerException ex) {
            return null;
        }
    }

    /**
     * Returns the connector pointed by the window and moves the window to the next connector.
     *
     * @return  the connector currently in the window.
     */
    public Connector getBackwardConnectorDoNext() {
        try {
            return (Connector)backwardLinks.getDoNext();
        }
        catch (NullPointerException ex) {
            return null;
        }
    }

    /**
     * Checks if this unit is already connected to the given unit (forwards).
     *
     * @param an the neuron to check
     */
    public boolean isAlreadyConnectedTo(AbstractPerceptronBase an) {
        this.firstForwardConnection();
        while (this.hasForwardConnection()) {
            if (this.getForwardConnectorDoNext().getRefNeuron() == an) {
                return true;
            }
        }
        return false;
    }

    /**
     * Checks if this unit is already connected from the given unit (backwards).
     *
     * @param an the neuron to check
     */
    public boolean isAlreadyConnectedFrom(AbstractPerceptronBase an) {
        this.firstBackwardConnection();
        while (this.hasBackwardConnection()) {
            if (this.getBackwardConnectorDoNext().getRefNeuron() == an) {
                return true;
            }
        }
        return false;
    }

    /**
     * Clear the list of backward and forward connections.
     */
    public void deleteConnections() {
        backwardLinks = new MyLinkedList();
        forwardLinks = new MyLinkedList();
    }

    public boolean receiveConnections() {
        return receiveConnections;
    }


    /**
     * Connects this unit with the given unit.
     *
     * @param   an  the unit to connect with
     * @param   w   the new value of the connection
     */
    public void connectForward(AbstractPerceptronBase an, float w) {
        if (!an.receiveConnections) {
            return;
        }

        if (this.isAlreadyConnectedTo(an)) {
            return;
        }

        MyFloat temp1 = new MyFloat(w);
        MyFloat temp2 = new MyFloat(0);

        this.forwardLinks.append(new Connector(an, temp1, temp2));
        an.backwardLinks.append(new Connector(this, temp1, temp2));
    }

    /**
     * Connects this unit with the given unit.
     *
     * @param   an  the unit to connect with
     */
    public void connectForward(AbstractPerceptronBase an) {
        if (!an.receiveConnections) {
            return;
        }

        if (this.isAlreadyConnectedTo(an)) {
            return;
        }

        MyFloat temp1 = new MyFloat(1);
        MyFloat temp2 = new MyFloat(0);

        this.forwardLinks.append(new Connector(an, temp1, temp2));
        an.backwardLinks.append(new Connector(this, temp1, temp2));
    }
    
    /**
     * Connects this unit with the given units.
     *
     * @param   ld  the layer containing the units to connect with
     */
    public void connectForward(AbstractPerceptronBaseLayer ld) {
        ld.head();
        while (ld.hasNode()) {
            this.connectForward((AbstractPerceptronBase)ld.getDoNext());
        }
    }

    public void appendToBackwardLinks(Connector c) {
        backwardLinks.append(c);
    }

    public void appendToForwardLinks(Connector c) {
        forwardLinks.append(c);
    }

    /**
     * Connects this unit with the given unit.
     *
     * @param   an  the unit to connect with
     * @param   w   the new value of the connection
     */
    public void connectBackward(AbstractPerceptronBase an, float w) {
        if (!this.receiveConnections) {
            return;
        }

        if (this.isAlreadyConnectedFrom(an)) {
            return;
        }

        MyFloat temp1 = new MyFloat(w);
        MyFloat temp2 = new MyFloat(0);

        this.backwardLinks.append(new Connector(an, temp1, temp2));
        an.forwardLinks.append(new Connector(this, temp1, temp2));
    }
    
    /**
     * Connects this unit with the given unit.
     *
     * @param   an  the unit to connect with
     */
    public void connectBackward(AbstractPerceptronBase an) {
        if (!this.receiveConnections) {
            return;
        }

        if (this.isAlreadyConnectedFrom(an)) {
            return;
        }

        MyFloat temp1 = new MyFloat(1);
        MyFloat temp2 = new MyFloat(0);

        this.backwardLinks.append(new Connector(an, temp1, temp2));
        an.forwardLinks.append(new Connector(this, temp1, temp2));
    }

    /**
     * Connects this unit with the given units.
     *
     * @param   ld  the layer containing the units to connect with
     */
    public void connectBackward(AbstractPerceptronBaseLayer ld) {
        ld.head();
        while (ld.hasNode()) {
            this.connectBackward((AbstractPerceptronBase)ld.getDoNext());
        }
    }

    /**
     * Copies the information saved in a connector in order to connect forward this unit with a given one.
     *
     * @param   an  the unit to connect with
     * @param   tc  the connector from where the weight values are copied
     */
    public void connectForwardClone(AbstractPerceptronBase an, Connector tc) {
        if (!an.receiveConnections()) {
            return;
        }

        MyFloat temp1 = new MyFloat(tc.getWeightValue());
        MyFloat temp2 = new MyFloat(tc.getLastWeightIncrement());

        this.forwardLinks.append(new Connector(an, temp1, temp2));
        an.backwardLinks.append(new Connector(this, temp1, temp2));
    }

    /**
     * Copies the information saved in a unit in order to connect forward this unit with a given layer.
     *
     * @param   ld  the layer to connect with
     * @param   tc  the unit from where the weight values are copied
     */
    public void connectForwardClone(AbstractPerceptronBaseLayer ld, AbstractPerceptronBase tc) {
        ld.head();
        tc.forwardLinks.head();

        while (ld.hasNode() && tc.forwardLinks.hasNode()) {
            this.connectForwardClone((AbstractPerceptronBase)ld.getDoNext(), (Connector)tc.forwardLinks.getDoNext());   //It connect this to ld.unit using information in tc
        }
    }

    /**
     * Copies the information saved in a connector in order to connect backward this unit with a given one.
     *
     * @param   an  the unit to connect with
     * @param   cInfo   the connector from where the weight values are copied
     */
    public void connectBackwardClone(AbstractPerceptronBase an, Connector cInfo) {
        if (!this.receiveConnections) {
            return;
        }

        MyFloat temp1 = new MyFloat(cInfo.getWeightValue());
        MyFloat temp2 = new MyFloat(cInfo.getLastWeightIncrement());

        this.backwardLinks.append(new Connector(an, temp1, temp2));
        an.forwardLinks.append(new Connector(this, temp1, temp2));
    }

    /**
     * Copies the information saved in a unit in order to connect backward this unit with a given layer.
     *
     * @param   lTo the layer to connect with
     * @param   uInfo   the unit from where the weight values are copied
     */
    public void connectBackwardClone(AbstractPerceptronBaseLayer lTo, AbstractPerceptronBase uInfo) {
        lTo.head();
        uInfo.backwardLinks.head();

        while (lTo.hasNode() && uInfo.backwardLinks.hasNode()) {
            this.connectBackwardClone((AbstractPerceptronBase)lTo.getDoNext(), (Connector)uInfo.backwardLinks.getDoNext());   //It connect this to lTo.units using information in uInfo
        }
    }

    /**
     * Calculates the parameter delta of the unit.
     */
    public void calcDelta() {
        Connector temp;
        delta = 0;

        forwardLinks.head();
        while (forwardLinks.hasNode()) {
            temp = (Connector) forwardLinks.getDoNext();
            delta += temp.getWeightValue() * temp.getRefNeuron().delta;
        }

        delta *= this.d_output;
    }

    public void addToWeightIncrements() {
        Connector temp;
        backwardLinks.head();
        while (backwardLinks.hasNode()) {
            temp = (Connector) backwardLinks.getDoNext();
//            temp.addToWeightIncrement(delta * temp.getRefNeuron().getOutput());
            temp.addToWeightIncrement(delta * temp.getInput());
        }
    }

    /**
     * Applies the delta rule in order to update the connection weights.
     *
     * @param   eta the eta learning parameter
     * @param   mu  the momentum term
     * @param   wd  the regularization term (weight decay)
     */
    public void updateWeightsDeltaRuleOnline(float eta, float mu, float wd) {     //eta must be negative for minimization
        float etadelta = eta * delta;
        float etawd = eta * wd;
        float currentW;
        Connector temp;

        backwardLinks.head();
        while (backwardLinks.hasNode()) {
            temp = (Connector) backwardLinks.getDoNext();
            //temp.subtractToWeight((etadelta * temp.getRefNeuron().output) + (mu * temp.getLastWeightIncrement()));
            if (temp.getRefNeuron().use) {
//                if (temp.getRefNeuron().isBias() || (Double.compare(wd, 0.0) == 0)) {
                if ((!temp.getRefNeuron().receiveConnections()) || (Float.compare(wd, 0) == 0)) {
                    temp.addToWeight((etadelta * temp.getRefNeuron().output) + (mu * temp.getLastWeightIncrement()));
                }
                else {
                    //temp.addToWeight((etadelta * temp.getRefNeuron().output) + (mu * temp.getLastWeightIncrement()) - (wd * temp.getWeightValue()));
                    currentW = temp.getWeightValue();
                    temp.addToWeight((float)((etadelta * temp.getRefNeuron().output) + (mu * temp.getLastWeightIncrement()) + ( (0.0025 * etawd * currentW) / Math.pow(0.0025 + (currentW * currentW), 2) ) ));
                }

            }
            else {
                temp.setWeight(0);
            }
        }
    }

    public void updateWeightsDeltaRuleBatch(float eta, float mu, float wd) {     //eta must be negative for minimization
        Connector temp;
        float currentW;

        backwardLinks.head();
        while (backwardLinks.hasNode()) {
            temp = (Connector)backwardLinks.getDoNext();
            currentW = temp.getWeightValue();
            temp.multiplyWeightIncrementBy(eta);

            if ((!temp.getRefNeuron().receiveConnections) || (Float.compare(wd, 0) == 0)) {
                temp.addToWeightIncrement(mu * temp.getLastWeightIncrement());
            }
            else {
                temp.addToWeightIncrement((float)(mu * temp.getLastWeightIncrement() + ( (0.0025 * eta * wd * currentW) / Math.pow(0.0025 + (currentW * currentW), 2) ) ));
            }
            temp.updateWeight();
        }
    }

    /**
     * Applies quickprop in order to update the connection weights.
     *
     * @param   eta the eta learning parameter
     * @param   mu  the maximum growth factor
     */
    public void updateWeightsQuickprop(float eta, float mu) {
        float currentPartialDerivative;
        float lastPartialDerivative;
        float deltaW;
        float lastDeltaW;
        Connector temp;
        float factor;

        backwardLinks.head();
        while (backwardLinks.hasNode()) {
            temp = (Connector) backwardLinks.getDoNext();
            if (temp.getRefNeuron().use) {
                currentPartialDerivative = delta * temp.getRefNeuron().output;
                lastPartialDerivative = temp.getLastPartialDerivative();
                lastDeltaW = temp.getLastWeightIncrement();

                deltaW = 0;

                factor = currentPartialDerivative / lastPartialDerivative - currentPartialDerivative;
                if (factor < 0) {
                    if (Math.signum(currentPartialDerivative) == Math.signum(lastPartialDerivative)) {
                        if (Math.abs(currentPartialDerivative) > Math.abs(lastPartialDerivative)) {     //uphill conditions
                            factor *= -1;
                        }
                    }
                }

                if (factor > mu) {
                    deltaW += mu * lastDeltaW;
                }
                else if (factor < -mu) {
                    deltaW -= mu * lastDeltaW;
                }
                else {
                    deltaW += factor *lastDeltaW;
                }
                deltaW -= eta * currentPartialDerivative;

                temp.addToWeight(deltaW);
                temp.setLastPartialDerivative(currentPartialDerivative);
            }
            else {
                temp.resetWeight();
            }
        }
    }

//    /**
//     * Applies quickprop in order to update the connection weights.
//     *
//     * @param   eta the eta learning parameter
//     * @param   mu  the maximum growth factor
//     */
//    public void updateWeightsQuickprop(double eta, double mu) {
//        double currentPartialDerivative;
//        double lastPartialDerivative;
//        double deltaW;
//        double lastDeltaW;
//        Connector temp;
//        double factor;
//
//        backwardLinks.head();
//        while (backwardLinks.hasNode()) {
//            temp = (Connector) backwardLinks.getDoNext();
//            if (temp.getRefNeuron().use) {
//                currentPartialDerivative = delta * temp.getRefNeuron().output;
//                lastPartialDerivative = temp.getLastPartialDerivative();
//                lastDeltaW = temp.getLastWeightIncrement();
//
//                deltaW = 0;                                     //this is a modification to ensure downhill
//                factor = currentPartialDerivative / Math.abs(lastPartialDerivative - currentPartialDerivative);
//                if (factor > mu) {
//                    deltaW -= mu * Math.abs(lastDeltaW);
//                }
//                else if (factor < -mu) {
//                    deltaW += mu * Math.abs(lastDeltaW);
//                }
//                else {
//                    deltaW -= factor * Math.abs(lastDeltaW);
//                }
//                deltaW -= eta * currentPartialDerivative;
//
//                temp.addToWeight(deltaW);
//                temp.setLastPartialDerivative(currentPartialDerivative);
//            }
//            else {
//                temp.resetWeight();
//            }
//        }
//    }

//    /**
//     * Applies quickprop in order to update the connection weights.
//     *
//     * @param   eta the eta learning parameter
//     * @param   mu  the maximum growth factor
//     */
//    public void updateWeightsQuickprop(double eta, double mu) {
//        double currentPartialDerivative;
//        double lastPartialDerivative;
//        double deltaW;
//        double lastDeltaW;
//        Connector temp;
//        double shrinkFactor = mu / (1 + mu);
//
//        backwardLinks.head();
//        while (backwardLinks.hasNode()) {
//            temp = (Connector) backwardLinks.getDoNext();
//            if (temp.getRefNeuron().use) {
//                currentPartialDerivative = delta * temp.getRefNeuron().output;
//                lastPartialDerivative = temp.getLastPartialDerivative();
//                lastDeltaW = temp.getLastWeightIncrement();
//
//                deltaW = 0;
//                if(lastDeltaW < 0.0) {
//                    if(currentPartialDerivative > 0.0) {
//                        deltaW -= eta * currentPartialDerivative;
//                    }
//                    if(currentPartialDerivative >= (shrinkFactor*lastPartialDerivative)) {
//                        deltaW += mu * lastDeltaW;
//                    }
//                    else {
//                        deltaW += lastDeltaW * currentPartialDerivative / (lastPartialDerivative - currentPartialDerivative);
//                    }
//                }
//                else {
//                    if(lastDeltaW > 0.0) {
//                        if(currentPartialDerivative < 0.0) {
//                            deltaW -= eta * currentPartialDerivative;
//                        }
//                        if(currentPartialDerivative <= (shrinkFactor*lastPartialDerivative)) {
//                            deltaW += mu * lastDeltaW;
//                        }
//                        else {
//                            deltaW += lastDeltaW * currentPartialDerivative / (lastPartialDerivative - currentPartialDerivative);
//                        }
//                    }
//                    else {
//                        deltaW -= eta * currentPartialDerivative;
//                    }
//                }
//
//                temp.addToWeight(deltaW);
//                temp.setLastPartialDerivative(currentPartialDerivative);
//            }
//            else {
//                temp.resetWeight();
//            }
//        }
//    }

//epsilon = eta;
//s = currentPartialDerivative = delta * temp.getRefNeuron().output;
//p = lastPartialDerivative
//w = temp.getWeightValue();
//deltas = deltaW;
//next_step = deltaW(+1);
//d = deltaW(-1);
//shrink_factor = mu/(1+mu);



//void QUICKPROP_UPDATE(int i, float weights[], float deltas[], float slopes[],
//		      float prevs[], float epsilon, float decay, float mu,
//		      float shrink_factor)
//{
//  float w,d,s,p, next_step;
//  /********/
//
//  w = weights[i];
//  d = deltas[i];
//  s = slopes[i] +  decay * w;
//  p = prevs[i];
//  next_step = 0.0;
//
//  /* The step must always be in direction opposite to the slope. */
//
//  if(d < 0.0){
//    /* If last step was negative...  */
//    if(s > 0.0)
//      /*  Add in linear term if current slope is still positive.*/
//      next_step -= epsilon * s;
//    /*If current slope is close to or larger than prev slope...  */
//    if(s >= (shrink_factor*p))
//      next_step += mu * d;	/* Take maximum size negative step. */
//    else
//      next_step += d * s / (p - s); /* Else, use quadratic estimate. */
//  }
//  else if(d > 0.0){
//    /* If last step was positive...  */
//    if(s < 0.0)
//      /*  Add in linear term if current slope is still negative.*/
//      next_step -= epsilon * s;
//    /* If current slope is close to or more neg than prev slope... */
//    if(s <= (shrink_factor*p))
//      next_step += mu * d;	/* Take maximum size negative step. */
//    else
//      next_step += d * s / (p - s); /* Else, use quadratic estimate. */
//  }
//  else
//    /* Last step was zero, so use only linear term. */
//    next_step -= epsilon * s;
//
//  /* update global data arrays */
//  deltas[i] = next_step;
//  weights[i] = w + next_step;
//  prevs[i] = s;
//  slopes[i] = 0.0;
//}

    public void disconnectForwardFrom(AbstractPerceptronBase to) {
        forwardLinks.head();
        while (forwardLinks.hasNode()) {
            if (((Connector)forwardLinks.getCurrent()).getRefNeuron() == to) {
                forwardLinks.deleteCurrentDoNext();
                break;
            }
            else {
                forwardLinks.next();
            }
        }
    }

    public void reset() {
        if (activationFunction != null) {
            output = activationFunction.apply(0);
        }
        else {
            output = 0;
        }
    }

    protected boolean use;
    protected int number;
    protected AbstractActivationFunction activationFunction;
    protected AbstractOutputFunction outputFunction;

    protected boolean receiveConnections;
    protected MyLinkedList backwardLinks;
    protected MyLinkedList forwardLinks;

    protected float output;
    protected float d_output;           //output first derivative
    protected float activation;
    protected float neta;

    protected float delta;


}
