/*
 * AbstractLayer.java
 *
 * Created on 1. mai 2006, 16:34
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package model.distributedRepresentation;

import data.MyLinkedList;
import java.io.Serializable;
import java.util.Random;

/**
 *
 * @author Hector
 */
public abstract class AbstractPerceptronBaseLayer extends MyLinkedList implements Serializable, Cloneable {
    
    /**
     * Creates a new instance of AbstractLayer.
     */
    public AbstractPerceptronBaseLayer() {
    }

    /**
     * Returns a copy of the object.
     */
    public AbstractPerceptronBaseLayer clone() {
        AbstractPerceptronBaseLayer tempc = null;
        tempc = (AbstractPerceptronBaseLayer) super.clone();
        this.head();
        while (this.hasUnit()) {
            tempc.append(this.getUnitDoNext().clone());
        }
        return tempc;
    }
    
    
    /**
     * Returns the number of units in the layer.
     *
     * @return the mount of units in the layer
     */
    public int getAmountOfUnits() {
        return this.size();
    }

    /**
     * Points the list window of the layer to the first element.
     */
    public void firstUnit() {
        this.head();
    }
    
    /**
     * Checks the state of the list window.
     *
     * @return  true, if the window contains an unit, false if the window is pointing to null
     */
    public boolean hasUnit() {
        return this.hasNode();
    }
    
    /**
     * Returns a list with the identifiers of all the units in the layer.
     *
     * @return  a {@link String} object containing the all the identifiers of the objects in the list.
     */
    public String getUnitList() {
        String temps = new String();
        
        this.head();
        while (this.hasNode()) {
            temps += String.valueOf(this.getUnitDoNext().getNumber());
            temps += "\t";
        }
        return temps;
    }

    /**
     * Sets the flag {@link AbstractNeuron#use} of every unit in the layer.
     *
     */
     public void useAllUnits() {
        this.head();
        while(this.hasNode()) {
            this.getUnitDoNext().setUse(true);
        }
    }
    
     /**
      * Adds new units to the layer.
      *
      * @param  h   an {@link AbstractLayer} object containing the units to add to the current layer
      */
    public void addUnits(AbstractPerceptronBaseLayer h) {
        h.firstUnit();
        for (int i = 0; i < h.getAmountOfUnits(); i++) {
            this.append(h.getUnitDoNext());
        }
    }
    
    /**
     * Returns the current unit in the list and moves forward the window.
     *
     * @return  the {@link AbstractNeuron} object in the window of the list
     */
    public AbstractPerceptronBase getUnitDoNext() {
        return (AbstractPerceptronBase)this.getDoNext();
    }

    public void setNumber(int n) {
        layerNumber = n;
    }
    
    public int getNumber() {
        return layerNumber;
    }

    public float getForwardWeightSum() {
        float sum = 0;
        this.head();

        while (this.hasNode()) {
            sum += this.getUnitDoNext().getForwardWeightSum();
        }
        return sum/this.size();
    }

    /**
     * Connects this layer with a given layer.
     *
     * @param   ld  the layer containing the units to connect with the units in this layer
     */
    public void connectForward(AbstractPerceptronBaseLayer ld) {
        this.head();
        while (this.hasNode()) {
            this.getUnitDoNext().connectForward(ld);
        }
    }


    /**
     * Copies the forward connection weights of a given layer in order to build new connections between two layers.
     *
     * @param   ld  the layer to connect with
     * @param   tc  the layer from where the information is extracted
     */
    public void connectForwardClone(AbstractPerceptronBaseLayer ld, AbstractPerceptronBaseLayer tc) {
        this.head();
        tc.head();

        while (this.hasNode() && tc.hasNode()) {
            this.getUnitDoNext().connectForwardClone(ld, tc.getUnitDoNext());      //It connects unit to ld.units creating new connections with the information in tc.units.unit
        }
    }

    /**
     * Copies the backward connection weights of a given layer in order to build new connections between two layers.
     *
     * @param   lTo  the layer to connect with
     * @param   lInfo  the layer from where the information is extracted
     */
    public void connectBackwardClone(AbstractPerceptronBaseLayer lTo, AbstractPerceptronBaseLayer lInfo) {
        this.head();
        lInfo.head();

        while (this.hasNode() && lInfo.hasNode()) {
            this.getUnitDoNext().connectBackwardClone(lTo, lInfo.getUnitDoNext());      //It connects unit to lTo.units creating new connections with the information in lInfo.units.unit
        }
    }

    /**
     *  Copies the value of the connection weights of a given layer.
     *
     * @param   toCopy  layer containing the information to copy
     */
    public void copyConnections (AbstractPerceptronBaseLayer toCopy) {
        this.head();
        toCopy.head();
        while (this.hasNode() && toCopy.hasNode()) {
            this.getUnitDoNext().copyConnections(toCopy.getUnitDoNext());
        }
    }

    /**
     * Initializes every unit in the layer.
     *
     * @param   gen the random generator object to initialize each unit
     */
    public void initLayer(Random gen) {
        this.head();
        while (this.hasNode()) {
            this.getUnitDoNext().initNeuron(gen);
        }
    }

    /**
     * Calculates the output of every unit in the layer
     */
    public void calcOutput() {
        this.head();
        while (this.hasNode()) {
            this.getUnitDoNext().calcNeta();
        }
        this.head();
        while (this.hasNode()) {
            this.getUnitDoNext().calcOutput();
        }
    }

    /**
     * Calculates the parameter delta of every unit in the layer.
     */
    public void calcDelta() {
        this.head();
        while (this.hasNode()) {
            this.getUnitDoNext().calcDelta();
        }
    }

    /**
     * Update the weights of every unit in the layer using the delta rule.
     */
    public void updateWeightsDeltaRuleOnline(float eta, float mu, float wd) {
        this.head();
        while (this.hasNode()) {
            this.getUnitDoNext().updateWeightsDeltaRuleOnline(eta, mu, wd);
        }
     }

    /**
     *
     */
    public void addToWeightIncrements() {
        this.head();
        while (this.hasNode()) {
            this.getUnitDoNext().addToWeightIncrements();
        }
     }

    /**
     * Update the weights of every unit in the layer using the last batch calculations.
     */
    public void updateWeightsDeltaRuleBatch(float eta, float mu, float wd) {
        this.head();
        while (this.hasNode()) {
            this.getUnitDoNext().updateWeightsDeltaRuleBatch(eta, mu, wd);
        }
    }

    /**
     * Update the weights of every unit in the layer using the delta rule.
     */
    public void updateWeightsQuickprop(float eta, float mu) {
        this.head();
        while (this.hasNode()) {
            this.getUnitDoNext().updateWeightsQuickprop(eta, mu);
        }
     }

    /**
     * Connects a given unit in this layer to a specific unit in a given layer.
     *
     * @param   fromUnit  the position of the unit in this layer
     * @param   toLayer  the target layer
     * @param   toUnit  the position of the unit in the target layer
     */
    public void connectForward(int fromUnit, AbstractPerceptronBaseLayer toLayer, int toUnit) {
        ((AbstractPerceptronBase)this.getAt(fromUnit)).connectForward((AbstractPerceptronBase)toLayer.getAt(toUnit));
    }

    public void reset() {
        firstUnit();
        while (hasUnit()) {
            getUnitDoNext().reset();
        }
    }

    protected int layerNumber;
}
