/*
 * Connection.java
 *
 * Created on 1. mai 2006, 12:36
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package model.distributedRepresentation;

//import data.MyDouble;
import data.MyFloat;
import java.io.Serializable;

/**
 *
 * @author Hector
 */
public class Connector implements Serializable {
    
    /**
     * Creates a new instance of Connection.
     *
     */
    public Connector() {
        weight = null;
        lastWeightIncrement = null;
        refNeuron = null;
        lastPartialDerivative = new MyFloat(0);
        weightIncrement = 0;
        input = 0;
    }
    
    /**
     * Creates a new instance of Connection.
     *
     * @param   r   the neuron connected
     * @param   w   the value of the connection
     * @param   lw  the last value of the connection
     */
    public Connector(AbstractPerceptronBase r, MyFloat w, MyFloat lw) {
        weight = w;
        lastWeightIncrement = lw;
        refNeuron = r;
        lastPartialDerivative = new MyFloat(0);
        weightIncrement = 0;
        input = 0;
    }

    /**
     * Sets the value of the variable {@link #weight}.
     *
     * @param   d   the new value
     */
    public void setWeight(float d) {
        weight.setValue(d);
    }
    
    /**
     * Sets the value of variables {@link #weight}, {@link #weightIncrement}, {@link #lastWeightIncrement} to 0.
     *
     */
    public void resetWeight() {
        weight.setValue(0);
        weightIncrement = 0;
        lastWeightIncrement.setValue(0);
    }

    /**
     * Returns the object containing the value of the connection.
     *
     * @return  a {@link MyDouble} object containing the value of the connection
     */
    public MyFloat getWeight() {
        return weight;
    }

    /**
     * Returns  the value of the connection.
     *
     * @return   the value of the connection
     */
    public float getWeightValue() {
        return weight.getValue();
    }
    
    /**
     * Return the connected neuron.
     *
     * @return  the connected unit
     */
    public AbstractPerceptronBase getRefNeuron() {
        return refNeuron;
    }

    /**
     * Adds a given value to the connection value and updates the last value.
     *
     * @param   d   the value to add
     */
    public void addToWeight(float d) {
        weight.setValue(weight.getValue() + d);
        lastWeightIncrement.setValue(d);
    }

    /**
     * Returns  the last value of the connection.
     *
     * @return   the last value of the connection
     */
    public float getLastWeightIncrement() {
        return lastWeightIncrement.getValue();
    }
    
    /**
     * Returns  the last value of the partial derivative.
     *
     * @return   the last value of the partial derivative
     */
    public float getLastPartialDerivative() {
        return lastPartialDerivative.getValue();
    }

    /**
     * Sets  the value of the last partial derivative.
     *
     */
    public void setLastPartialDerivative(float lp) {
        lastPartialDerivative.setValue(lp);
    }

    public void addToWeightIncrement(float wi) {
//        float temp = weightIncrement;
        weightIncrement += wi;
//        if ((Float.compare(wi, Float.NaN) == 0) || (Float.compare(weightIncrement, Float.NaN) == 0)) {
//            System.err.print(temp);
//            System.err.print(" + ");
//            System.err.print(wi);
//            System.err.print(" = ");
//            System.err.println(weightIncrement);
//        }
    }

    public void multiplyWeightIncrementBy(float wif) {
        weightIncrement *= wif;
    }

    public void updateWeight() {
        weight.setValue(weight.getValue() + weightIncrement);
        lastWeightIncrement.setValue(weightIncrement);
        weightIncrement = 0;
    }

    public void setInput(float i) {
        input = i;
    }

    public float getInput() {
        return input;
    }

    protected MyFloat weight;
    protected MyFloat lastWeightIncrement;
    protected float weightIncrement;
    protected MyFloat lastPartialDerivative;
    protected AbstractPerceptronBase refNeuron;
    protected float input;                        //input value in the current iteration
}
