package model.localRepresentation;

/*
 * AbstractCell.java
 *
 * Created on 24 de septiembre de 2007, 11:32
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

//import dataModules.MyDouble;
import java.io.Serializable;
import java.util.Random;

/**
 *
 * @author Hector
 */
public abstract class AbstractFNNXCell implements Serializable {
    
    /**
     * Initializes the unit.
     *
     * @param   g   the random number generator to be used in the initialization
     */
    public void initNeuron(Random g) {
        for (int i = 0; i < position.length; i++) {
            position[i] = (2*g.nextFloat())-1;
        }
    }
    
    /**
     * Initializes the unit.
     *
     * @param   p   a point randomly chosen from the set of examples
     */
    public void initNeuron(float[] p) {
        for (int i = 0; i < position.length; i++) {
            position[i] = p[i];
        }
    }

    /**
     * Returns the position of the unit.
     *
     * @return  an array containing the coordinates of the unit
     */
    public float[] getPosition() {
        return position;
    }
    
    /**
     * Returns the content of the variable {@link #dist}.
     *
     * @return  an array containing the distance for each dimension
     */
    public float[] getDistanceVector() {
        return dist;
    }

    public void move(float e) {
        for (int i = 0; i < position.length; i++) {
            position[i] += e*dist[i];
        }
    }
    
    public float getMaxDistance() {
        float max = Float.MIN_VALUE;
        float temp;
        for (int i = 0; i < dist.length; i++) {
            temp = Math.abs(dist[i]);
            if (temp > max)  {
                max = temp;
            }
        }
        return max;
    }
    
    public float calcDistanceTo(float[] in) {
//        if (position.length != in.length) {
//            return Float.NaN;
//        }

        int missingCounter = 0;
        distMag = 0;
        for (int i = 0; i < position.length; i++) {
            if (Float.isNaN(in[i]) || Float.isNaN(this.position[i])) {
                dist[i] = 0;
                missingCounter++;
            }
            else {
                dist[i] = in[i] - this.position[i];
                distMag += dist[i]*dist[i];
            }
        }
        return (position.length / (position.length - missingCounter)) * distMag;
    }

    public float calcDistanceTo(AbstractFNNXCell cell) {
//        if (cell.getPosition().length != this.position.length) {
//            return Float.NaN;
//        }

        int missingCounter = 0;
        distMag = 0;
        for (int i = 0; i < this.position.length; i++) {
            if (Float.isNaN(cell.position[i]) || Float.isNaN(this.position[i])) {
                dist[i] = 0;
                missingCounter++;
            }
            else {
                dist[i] = this.position[i] - cell.position[i];
                distMag += dist[i]*dist[i];
            }
        }
        return (position.length / (position.length - missingCounter)) * distMag;

    }
    
    public String toString() {
        String temp = "";
        int i;
        
        for (i = 0; i < position.length; i++) {
            temp += String.valueOf(position[i]);
            if ((i+1) < position.length) {
                temp += "\t";
            }
            else {
                temp += "\n";
            }
        }
        return temp;
    }

    public void setNumber(int n) {
        number = n;
    }

    public int getNumber() {
        return number;
    }

    public void decreaseError(float d) {
        error = d * error;
    }

    public void setError(float e) {
        error = e;
    }
    
    public float getError() {
        return error;
    }

    protected int number;
    protected float[] position;
    protected float[] dist;
    protected float distMag;
    protected float error;

}
