/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package plugin;

import java.io.Serializable;
import java.lang.reflect.Array;
import java.util.Arrays;
import java.util.Hashtable;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;

/**
 *
 * @author hector
 */
public abstract class AbstractFNNXObject implements Serializable, FNNXObject {

    public String toString() {
        return name;
    }

    protected void setName() {
        name = this.getClass().getSimpleName();
    }

    public boolean configure(Hashtable <String, Object> vars, Hashtable <String, Object> localVariables, InterThreadMonitor m) {
        int i;
        Object temp;
        monitor = m;
        variables = vars;
        complete = true;

        for (i = 0; i < parameterNames.length; i++) {
            if (localVariables.containsKey(parameterNames[i])) {
                temp = (Object)localVariables.get(parameterNames[i]);
            }
            else {
                if (variables.containsKey(parameterNames[i])) {
                    temp = (Object)variables.get(parameterNames[i]);
                }
                else {
                    if (parameters[i] == null) {
                        if (isParameterMandatory[i]) {
                            complete = false;
                        }
                    }
                    continue;
                }
            }
            if (!this.configure(i, temp, monitor)) {
                return false;
            }

        }

        if (this instanceof FNNXFunction) {
            if (complete) {
                return this.prepare(monitor);
            }
            else {
                return true;
            }
        }
        else {
            return this.prepare(monitor);
        }
        
    }

    public boolean configure(int varIndex, Object value, InterThreadMonitor m) {
        int j;
        Object[] tempArray;
        monitor = m;
        configured = false;

        if (varIndex < 0) {
            System.out.println("Warning. Trying to configure an inexistent vartiable");
            return true;
        }

        if ((varIndex < 0) || (varIndex > parameterNames.length)) {
            monitor.appendErrorMessage("Error. Index out of range. Impossible to configure object.");
            return false;
        }

        try {
            if (parameterTypes[varIndex].isArray()) {
                if (value.getClass().isArray()) {
                    tempArray = (Object[])value;
                }
                else {
                    tempArray = new Object[1];
                    tempArray[0] = value;
                }
                parameters[varIndex] = parameterTypes[varIndex].cast(Array.newInstance(parameterTypes[varIndex].getComponentType(), tempArray.length));
                for (j = 0; j < tempArray.length; j++) {
                    Array.set(parameters[varIndex], j, parameterTypes[varIndex].getComponentType().cast(tempArray[j]));
                }
            }
            else {
                parameters[varIndex] = parameterTypes[varIndex].cast(value);
            }
        }
        catch (ClassCastException ce) {
            monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[varIndex] + "\".");
            return false;
        }

        return true;
    }

    public String[] getParameterNames() {
        return parameterNames;
    }

    public boolean configure() {
        configurationWindow.setVisible(true);

        if (!configurationWindow.OKPressed()) {
            return false;
        }

        return true;
    }
    
    public String[] getParameterText() {
        return configurationWindow.getFields();
    }

    public void setParameterText(String[] p) {
        configurationWindow.setFields(p);
    }

    public int getParameterIndex(String name) {
        int index;
        for (index = (parameterNames.length-1); index >= 0; index--) {
            if (parameterNames[index].compareTo(name) == 0) {
                break;
            }
        }

//        if (index == -1) {
//            monitor.appendErrorMessage("Error. Variable \""+ name + "\" not found.");
//        }

        return index;
    }

    public String show() {
        if (parameterNames == null) {
            return this.toString();
        }
        else{
            String temp = name + "{\n";
            for (int i = 0; i < parameterNames.length; i++) {
                if (parameters[i] != null) {
                    temp += "\t" + parameterNames[i] + " = ";
                    if (parameters[i] instanceof Object[]) {
                        temp += Arrays.toString((Object[])parameters[i]);
                    }
                    else {
                        temp += parameters[i].toString();
                    }
                    temp += "\n";
                }
            }
            temp += "}\n";
            return temp;
        }
    }

    public boolean prepare(InterThreadMonitor m) {
        monitor.appendErrorMessage("Error. Fuction \"prepare\" must be implemmented in user plug-ins.");
        return false;
    }

    public boolean testCompleteness() {
        int i;
        for (i = 0; i < parameterNames.length; i++) {
            if (isParameterMandatory[i]) {
                if (parameters[i] == null) {
                    monitor.appendErrorMessage("Error. Parameter \"" + parameterNames[i] + "\" not found when configuring " + this.name);
                    return false;
                }
            }
        }
        return true;
    }

    public boolean testMin(int pn, int value, int min) {
        if (value < min) {
            monitor.appendErrorMessage("Error. Parameter \"" + parameterNames[pn] + "\" must be larger than " + String.valueOf(min));
            return false;
        }
        else {
            return true;
        }
    }

    public boolean testMin(int pn, float value, float min) {
        if (value < min) {
            monitor.appendErrorMessage("Error. Parameter \"" + parameterNames[pn] + "\" must be larger than " + String.valueOf(min));
            return false;
        }
        else {
            return true;
        }
    }

    public boolean testMax(int pn, int value, int max) {
        if (value > max) {
            monitor.appendErrorMessage("Error. Parameter \"" + parameterNames[pn] + "\" must be smaller than " + String.valueOf(max));
            return false;
        }
        else {
            return true;
        }
    }

    public boolean testMax(int pn, float value, float max) {
        if (value > max) {
            monitor.appendErrorMessage("Error. Parameter \"" + parameterNames[pn] + "\" must be smaller than " + String.valueOf(max));
            return false;
        }
        else {
            return true;
        }
    }

    protected String name;
    protected String parameterName;
    transient protected AbstractFNNXConfigurationWindow configurationWindow;

    transient protected Hashtable <String, Object> variables;

    protected String[] parameterNames;
    protected Class[] parameterTypes;
    protected boolean[] isParameterMandatory;
    protected Object[] parameters;

    transient protected InterThreadMonitor monitor;
    protected InterThreadBuffer results;

    private boolean complete;
    protected boolean configured;

}
