/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package plugin;

import data.DataBlock;
import java.io.Serializable;
import java.util.Arrays;
import model.AbstractFNNXModel;

/**
 *
 * @author hsatizab
 */
public abstract class AbstractFNNXTrainingAlgorithm extends AbstractFNNXFunction implements Serializable, FNNXTrainingAlgorithm {

    public void singleStep(int it) {
        evaluateModel(it);
        updateModel(it);
    }

    public boolean testAndLog(int it, int expectedTestL, int expectedValL) {
        float[] resultTest;
        float[] resultVal;
        boolean stop = false;

        if (hasTester) {
            resultTest = (float[])tester.eval(monitor);

            if (resultTest == null) {
                if (expectedTestL != 0) {
                    monitor.appendErrorMessage("Error. " + tester.toString() + " failed.");
                    stop = true;
                    return stop;
                }
            }
            else {
                System.arraycopy(resultTest, 0, res, 0, resultTest.length);
                if (resultTest[0] < stopIfTesterBelow) {
                    stop = true;                                            //stops the algorithm
                }
            }
        }

        if (hasValidator) {
            resultVal = (float[])validator.eval(monitor);
            if (resultVal == null) {
                if (expectedValL != 0) {
                    monitor.appendErrorMessage("Error. " + validator.toString() + "failed.");
                    stop = true;
                    return stop;
                }
            }
            else {
                System.arraycopy(resultVal, 0, res, expectedTestL, resultVal.length);
            }
                                    //this is necesary because the validator attachs the validation dataset to the network
            if (!network.setDataset(dataset)) {
                monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
                    stop = true;
                    return stop;
            }
        }

        results.newData(res, it+1);
        //----------------------------------------------
        if (hasPlotter && (hasTester || hasValidator)) {
            plotter.getMonitor().newResults(results);
        }
        if (hasSaver && (hasTester || hasValidator)) {
            saver.getMonitor().newResults(results);
        }
        //----------------------------------------------

        return stop;
    }

    public void log(int it) {
        res[0] = (float)network.getAccError()/logCounter;

        results.newData(res, it);
        //----------------------------------------------
        if (hasPlotter) {
            plotter.getMonitor().newResults(results);
        }
        if (hasSaver) {
            saver.getMonitor().newResults(results);
        }
        //----------------------------------------------
        network.resetAccError();
    }

    protected AbstractFNNXModel network;
    protected DataBlock dataset;
    protected int epochs;
    protected AbstractFNNXFunction tester;
    protected boolean hasTester;
    protected AbstractFNNXFunction validator;
    protected boolean hasValidator;
    protected float stopIfTesterBelow;
    protected int logCounter;

    protected int datasetLength;
    protected int limit;


}
