/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package scripting;

import data.DataBlock;
import data.DataMatrix;
import data.FNNXTreeNode;
import data.MyLinkedList;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.lang.reflect.Array;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.Hashtable;
import plugin.AbstractFNNXParallelObserver;
import plugin.FNNXFunction;
import plugin.FNNXObject;

/**
 *
 * @author hsatizab
 */
public class FNNXInterpreter extends Thread {

    public FNNXInterpreter(FNNXTreeNode t, Hashtable<String, Object> o, Hashtable<String, Object> lo, Hashtable<String, Class> p, InterThreadMonitor m) {

        tree = t;
        objects = o;
        localObjects = lo;
        plugins = p;
        monitor = m;

        local = false;
        variableName = null;

    }

    public void run() {
        monitor.running();
        decodeNode(tree);

        if (monitor.isRunning()) {
            if (monitor.hasErrorMessage()) {
                System.out.println(monitor.getErrorMessage());
            }
//            else {
//                System.out.println("Script executed succesfully.");
//            }
        }
        else {
            System.out.println("Script halted.");
        }
        monitor.resetErrorMessage();
        monitor.notRunning();
    }

    public Object decodeNode(FNNXTreeNode tree) {
        FNNXTreeNode currentNode;
        FNNXToken currentToken;
        String name;
        Object object;
        Object tempObject;
        Object[] objectArray;
        Number[] numberArray;
//        String variableName = null;
        Class plugin;
        float op1 = 0, op2 = 0;
        boolean bop1 = false, bop2 = false;
        int i, mi, ma;

        if (!monitor.isRunning()) {
            return null;
        }

        currentToken = (FNNXToken)tree.getObject();
        switch (currentToken.getToken()) {
//            case LoBl:
//                local = true;
            case main:
//            case GlBl:
            case Bl:
                tree.head();
                while (tree.hasNode()) {
                    currentNode = (FNNXTreeNode)tree.getDoNext();
                    object = this.decodeNode(currentNode);
                    if (monitor.hasErrorMessage()) {
                        return null;
                    }
                }
                return null;

            case As:
                currentNode = (FNNXTreeNode)tree.getFirst();
                name = ((FNNXToken)currentNode.getObject()).getText();
                if (name == null) {
                    monitor.appendErrorMessage("Error. Impossible to find a variable name.");
                    return null;
                }
                currentNode = (FNNXTreeNode)tree.getLast();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Impossible to assign a value to variable \"" + name + "\".");
                    return null;
                }

                if (local) {
                    localObjects.put(name, object);
                }
                else {
                    tempObject = objects.put(name, object);
                    
                    if (tempObject != null) {
                        if (!objects.containsValue(tempObject)) {
                            if (tempObject instanceof AbstractFNNXParallelObserver) {
                                ((AbstractFNNXParallelObserver)tempObject).getMonitor().abort();
                                ((AbstractFNNXParallelObserver)tempObject).free();
                            }
                        }
                    }
                }
                return null;

//            case CrObLoBl:
//            case CaFuLoBl:
            case CrObBl:
            case CaFuBl:
                local = true;
                currentNode = (FNNXTreeNode)tree.getLast();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage()) {
                    monitor.appendErrorMessage("Error. Malformed block.");
                    return null;
                }
                currentNode = (FNNXTreeNode)tree.getFirst();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage()) {
                    monitor.appendErrorMessage("Error. Impossible to create object.");
                    return null;
                }
                return object;

            case CrOb:
                local = false;
                currentNode = (FNNXTreeNode)tree.getFirst();
                name = ((FNNXToken)currentNode.getObject()).getText();
                if (name == null) {
                    monitor.appendErrorMessage("Error. Impossible to find an object name.");
                    return null;
                }
                if (plugins.containsKey(name)) {
                    plugin = plugins.get(name);

                    try {
                        object = plugin.getConstructor(new Class[] {Hashtable.class, InterThreadMonitor.class}).newInstance(new Object[] {objects, monitor});
                        if (((FNNXObject)object).configure(objects, localObjects, monitor) == false) {
                            monitor.appendErrorMessage("Error. Impossible to configure plugin.");
                            return null;
                        }
                        localObjects.clear();
                        return object;
                    }
                    catch (NoSuchMethodException ne) {
                        monitor.appendErrorMessage("Error. Impossible to access plugin \"" + name +"\". (method)");
                        return null;
                    }
                    catch (IllegalAccessException ie) {
                        monitor.appendErrorMessage("Error. Impossible to access plugin \"" + name +"\". (access)");
                        return null;
                    }
                    catch (InvocationTargetException ie) {
                        monitor.appendErrorMessage("Error. Impossible to access plugin \"" + name +"\". (target)");
                        return null;
                    }
                    catch (InstantiationException ie) {
                        monitor.appendErrorMessage("Error. Impossible to access plugin \"" + name +"\". (instantiation)");
                        return null;
                    }
                }
                else {
                    monitor.appendErrorMessage("Error. Impossible to find plugin \"" + name + "\".");
                    return null;
                }

            case CaFu:
                local = false;
                currentNode = (FNNXTreeNode)tree.getFirst();
                name = ((FNNXToken)currentNode.getObject()).getText();
                if (name == null) {
                    monitor.appendErrorMessage("Error. Impossible to find an object name.");
                    return null;
                }
                if (objects.containsKey(name)) {
                    object = objects.get(name);
                    try {
                        if (((FNNXObject)object).configure(objects, localObjects, monitor) == false) {
                            monitor.appendErrorMessage("Error. Impossible to configure plugin \"" + name +"\".");
                            return null;
                        }
                        localObjects.clear();
                        object = ((FNNXFunction)object).eval(monitor);
//                        return ((FNNXFunction)object).eval(monitor);
                    }
                    catch (ClassCastException ce) {
                        monitor.appendErrorMessage("Error. Variable \"" + name +"\" is not a valid plugin.");
                        return null;
                    }
                }
                else{
                    if (plugins.containsKey(name)) {
                        plugin = plugins.get(name);

                        try {
                            object = plugin.getConstructor(new Class[] {Hashtable.class, InterThreadMonitor.class}).newInstance(new Object[] {objects, monitor});
                            if (((FNNXObject)object).configure(objects, localObjects, monitor) == false) {
                                monitor.appendErrorMessage("Error. Impossible to configure plugin \"" + name +"\".");
                                return null;
                            }
                            localObjects.clear();
                            object = ((FNNXFunction)object).eval(monitor);
//                            return ((FNNXFunction)object).eval(monitor);
                        }
                        catch (NoSuchMethodException ne) {
                            monitor.appendErrorMessage("Error. Impossible to access plugin \"" + name +"\". (method)");
                            return null;
                        }
                        catch (IllegalAccessException ie) {
                            monitor.appendErrorMessage("Error. Impossible to access plugin \"" + name +"\". (access)");
                            return null;
                        }
                        catch (InvocationTargetException ie) {
                            monitor.appendErrorMessage("Error. Impossible to access plugin \"" + name +"\". (target)");
                            return null;
                        }
                        catch (InstantiationException ie) {
                            monitor.appendErrorMessage("Error. Impossible to access plugin \"" + name +"\". (instantiation)");
                            return null;
                        }
                    }
                    else {
                        monitor.appendErrorMessage("Error. Impossible to find plugin \"" + name + "\".");
                        return null;
                    }
                }
                if (object != null) {
                    if (object.getClass().isArray()) {
                        if (!(object instanceof Object[])) {
                            objectArray = new Object[Array.getLength(object)];
                            for (i = 0; i < objectArray.length; i++) {
                                objectArray[i] = Array.get(object, i);
                            }
                            object = objectArray;
                        }
                    }
                }
                return object;

            case Ve:
//            case Ve1:
                MyLinkedList tempVec = new MyLinkedList();
                tree.head();
                while (tree.hasNode()) {
                    currentNode = (FNNXTreeNode)tree.getDoNext();
                    object = this.decodeNode(currentNode);
                    if (monitor.hasErrorMessage() || (object == null)) {
                        monitor.appendErrorMessage("Error. Impossible to decode vector element.");
                        return null;
                    }
                    tempVec.append(object);
                }
                Object[] vector = new Object[tempVec.size()];
                tempVec.head();
                for (i = 0; tempVec.hasNode(); i++) {
                    vector[i] = tempVec.getDoNext();
                }
                return vector;

            case In:
                int index = 0;
                currentNode = (FNNXTreeNode)tree.getFirst();

                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Object not found.");
                    return null;
                }
                if (object instanceof Object[]) {
                    objectArray = (Object[])object;
                }
                else {
                    monitor.appendErrorMessage("Error. Not a valid array.");
                    return null;
                }

                currentNode = (FNNXTreeNode)tree.getLast();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Missing index.");
                    return null;
                }
                if (object instanceof Number) {
                    index = ((Number)object).intValue();
                }
                else {
                    monitor.appendErrorMessage("Error. Wrong index type.");
                    return null;
                }

                if ((index >= objectArray.length) || (index < 0)) {
                    monitor.appendErrorMessage("Error. Index out of bounds.");
                    return null;
                }
                return objectArray[index];

            case Ex:
            case LoEx:
                currentNode = (FNNXTreeNode)tree.getFirst();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Impossible to decode expression.");
                    return null;
                }
//                switch (((FNNXToken)currentNode.getObject()).getToken()) {
//                    case name:
//                        variableName = (String)object;
//                        if (objects.containsKey(variableName)) {
//                            object = objects.get(variableName);
//                        }
//                        else {
//                            monitor.appendErrorMessage("Error. Impossible to find variable \"" + variableName + "\".");
//                            return null;
//                        }
//                        break;
//                }
                return object;

            case Ex3:
            case LoEx3:
                currentNode = (FNNXTreeNode)tree.getFirst();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Impossible to decode first operand.");
                    return null;
                }

                currentNode = (FNNXTreeNode)tree.getLast();
                tempObject = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (tempObject == null)) {
                    monitor.appendErrorMessage("Error. Impossible to decode second operand.");
                    return null;
                }

                tree.tail();
                tree.previous();
                currentNode = (FNNXTreeNode)tree.getCurrent();

                switch (((FNNXToken)currentNode.getObject()).getToken()) {
                    case add:
                    case sub:
                    case mult:
                    case div:
                    case mod:
                    case eq:
                    case diff:
                    case less:
                    case lessEq:
                    case more:
                    case moreEq:
                        try {
                            op1 = ((Number)object).floatValue();
                        }
                        catch (ClassCastException ce) {
                            monitor.appendErrorMessage("Error. First operand is not a valid number.");
                            return null;
                        }
                        try {
                            op2 = ((Number)tempObject).floatValue();
                        }
                        catch (ClassCastException ce) {
                            monitor.appendErrorMessage("Error. Second operand is not a valid number.");
                            return null;
                        }
                        break;

                    case and:
                    case or:
                    case xor:
                        try {
                            bop1 = ((Boolean)object).booleanValue();
                        }
                        catch (ClassCastException ce) {
                            monitor.appendErrorMessage("Error. First operand is not a valid boolean value.");
                            return null;
                        }
                        try {
                            bop2 = ((Boolean)tempObject).booleanValue();
                        }
                        catch (ClassCastException ce) {
                            monitor.appendErrorMessage("Error. Second operand is not a valid boolean value.");
                            return null;
                        }
                        break;

                    default:
                        monitor.appendErrorMessage("Error. Invalid operation.");
                        return null;
                }

                switch (((FNNXToken)currentNode.getObject()).getToken()) {
                    case add:
                        return new Double(op1 + op2);
                    case sub:
                        return new Double(op1 - op2);
                    case mult:
                        return new Double(op1 * op2);
                    case div:
                        if (op2 == 0) {
                            monitor.appendErrorMessage("Error. Division by zero in line " + String.valueOf(((FNNXToken)currentNode.getObject()).getLineNumber()) + ".");
                            return null;
                        }
                        return new Double(op1 / op2);
                    case mod:
                        if (op2 == 0) {
                            monitor.appendErrorMessage("Error. Division by zero in line " + String.valueOf(((FNNXToken)currentNode.getObject()).getLineNumber()) + ".");
                            return null;
                        }
                        return new Double(op1 % op2);

                    case eq:
                        return op1 == op2;
                    case diff:
                        return op1 != op2;
                    case less:
                        return op1 < op2;
                    case lessEq:
                        return op1 <= op2;
                    case more:
                        return op1 > op2;
                    case moreEq:
                        return op1 >= op2;
                    case and:
                        return bop1 && bop2;
                    case or:
                        return bop1 || bop2;
                    case xor:
                        return bop1 ^ bop2;

                    default:
                        monitor.appendErrorMessage("Error. Invalid operation.");
                        return null;
                }

            case Ex2Pre:
                currentNode = (FNNXTreeNode)tree.getLast();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Impossible to decode operand.");
                    return null;
                }

                try {
                    op1 = ((Number)object).floatValue();
                }
                catch (ClassCastException ce) {
                    monitor.appendErrorMessage("Error. Operand is not a valid number.");
                    return null;
                }

                currentNode = (FNNXTreeNode)tree.getFirst();
                switch (((FNNXToken)currentNode.getObject()).getToken()) {
                    case neg:
                        return new Double((-1)*op1);
                    case inc:
                        return new Double(op1 + 1);
                    case dec:
                        return new Double(op1 - 1);
                    default:
                        monitor.appendErrorMessage("Error. Invalid operation.");
                        return null;
                }

            case Ex2Post:
                variableName = null;
                currentNode = (FNNXTreeNode)tree.getFirst();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Impossible to decode operand.");
                    return null;
                }

                try {
                    op1 = ((Number)object).floatValue();
                }
                catch (ClassCastException ce) {
                    monitor.appendErrorMessage("Error. Operand is not a valid number.");
                    return null;
                }

                currentNode = (FNNXTreeNode)tree.getLast();
                switch (((FNNXToken)currentNode.getObject()).getToken()) {
                    case inc:
                        object = new Double(op1 + 1);
                        if (variableName != null) {
                            objects.put(variableName, object);
                        }
                        return object;

                    case dec:
                        object = new Double(op1 - 1);
                        if (variableName != null) {
                            objects.put(variableName, object);
                        }
                        return object;

                    default:
                        monitor.appendErrorMessage("Error. Invalid operation.");
                        return null;
                }

            case WhLo:
                currentNode = (FNNXTreeNode)tree.getFirst();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Impossible to decode condition.");
                    return null;
                }
                while (((Boolean)object).booleanValue() == true) {
                    currentNode = (FNNXTreeNode)tree.getLast();
                    object = this.decodeNode(currentNode);
                    if (monitor.hasErrorMessage()) {
                        monitor.appendErrorMessage("Error. Impossible to continue loop execution.");
                        return null;
                    }

                    currentNode = (FNNXTreeNode)tree.getFirst();
                    object = this.decodeNode(currentNode);
                    if (monitor.hasErrorMessage() || (object == null)) {
                        monitor.appendErrorMessage("Error. Impossible to decode condition.");
                        return null;
                    }
                }
                return null;

            case IfBl:
                currentNode = (FNNXTreeNode)tree.getFirst();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Impossible to decode condition.");
                    return null;
                }
                if (((Boolean)object).booleanValue() == true) {
                    currentNode = (FNNXTreeNode)tree.getLast();
                    object = this.decodeNode(currentNode);
                    if (monitor.hasErrorMessage()) {
                        monitor.appendErrorMessage("Error. Impossible to continue loop execution.");
                        return null;
                    }
                }
                return null;

            case Mi:
            case Ma:
            case WhMi:
            case WhMa:
                currentNode = (FNNXTreeNode)tree.getFirst();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Impossible to decode object. Impossible to calculate min-max.");
                    return null;
                }
                if (!(object instanceof Object[])) {
                    monitor.appendErrorMessage("Error. The object is not an array. Impossible to calculate min-max.");
                    return null;
                }

                objectArray = (Object[])object;
                try {
                    numberArray = new Number[objectArray.length];
                    for (i = 0; i < numberArray.length; i++) {
                        numberArray[i] = (Float)objectArray[i];
                    }
                    
                }
                catch (ClassCastException ce) {
                    monitor.appendErrorMessage("Error. The object is not an array of numbers. Impossible to calculate min-max.");
                    return null;
                }

                op1 = numberArray[0].floatValue();
                mi = 0;
                op2 = numberArray[0].floatValue();
                ma = 0;
                for (i = 1; i < numberArray.length; i++) {
                    if ( (numberArray[i].floatValue() < op1) || (Float.isNaN(op1)) ) {
                        op1 = numberArray[i].floatValue();
                        mi = i;
                    }
                    if ( (numberArray[i].floatValue() > op2) || (Float.isNaN(op2)) ) {
                        op2 = numberArray[i].floatValue();
                        ma = i;
                    }
                }
                switch (currentToken.getToken()) {
                    case Mi:
                        return new Float(op1);
                    case Ma:
                        return new Float(op2);
                    case WhMi:
                        return new Float(mi);
                    case WhMa:
                        return new Float(ma);
                }

            case Le:
                currentNode = (FNNXTreeNode)tree.getFirst();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Impossible to decode object. Impossible to calculate length.");
                    return null;
                }
                if (object instanceof Object[]) {
                    objectArray = (Object[])object;
                    return new Float(objectArray.length);
                }
                else if (object instanceof DataBlock) {
                    return new Float(((DataBlock)object).getNParts());
                }
                else if (object instanceof DataMatrix) {
                    return new Float(((DataMatrix)object).getNRows());
                }
                else {
                    monitor.appendErrorMessage("Error. The object is not an array. Impossible to calculate length.");
                    return null;
                }

            case Sh:
                currentNode = (FNNXTreeNode)tree.getFirst();
                object = this.decodeNode(currentNode);
                if (monitor.hasErrorMessage() || (object == null)) {
                    monitor.appendErrorMessage("Error. Impossible to decode object. Impossible to show.");
                    return null;
                }

                if (object instanceof Object[]) {
                    System.out.println(Arrays.toString((Object[])object));
                }
                else {
                    if (object instanceof FNNXObject) {
                        System.out.println(((FNNXObject)object).show());
                    }
                    else {
                        System.out.println(object.toString());
                    }
                }
                return null;

            case So:
                currentNode = (FNNXTreeNode)tree.getFirst();
                try {
                    name = (String)this.decodeNode(currentNode);
                }
                catch (ClassCastException cce) {
                    monitor.appendErrorMessage("Error. Expecting a name for the script file.");
                    name = null;
                }
                if (monitor.hasErrorMessage() || (name == null)) {
                    monitor.appendErrorMessage("Error. Impossible to decode object. Impossible to launch script.");
                    return null;
                }
                
                FNNXLexer tempLexer = null;
                FNNXTreeNode tempTree= null;
                // read script
                try {
                    BufferedReader reader = new BufferedReader(new FileReader(name));
                    tempLexer = new FNNXLexer(reader);
                    tempTree = tempLexer.analyze();
                    reader.close();
                }
                catch (FileNotFoundException fe) {
                    monitor.appendErrorMessage("Error. The file: \"" + name + "\" was not found.");
                    return null;
                }
                catch (IOException ie) {
                    monitor.appendErrorMessage("Input/Ouptput error.");
                    return null;
                }
                // parse and decode
                FNNXParser tempParser = new FNNXParser();
                tempParser.parse(tempTree);
                tempParser.clean(tempTree);
                if (tempParser.checkErrors(tempTree)) {
                    this.decodeNode(tempTree);
                }
                else {
                    monitor.appendErrorMessage("Error. There were some errors while parsing the script.");
                    return null;
                }
                
                System.out.println("Script \"" + name + "\" executed.");
                return null;
            
            case Cat:
                String temp = "";
                tree.head();
                while (tree.hasNode()) {
                    currentNode = (FNNXTreeNode)tree.getDoNext();
                    object = this.decodeNode(currentNode);
                    if (monitor.hasErrorMessage() || (object == null)) {
                        monitor.appendErrorMessage("Error. Impossible to decode vector element.");
                        return null;
                    }

                    if (object instanceof Number) {
                        temp += object.toString().replace('.', '_');
                    }
                    else {
                        temp += object.toString();
                    }
                    
                }
                return temp;

            case name:
//                System.out.println(currentToken.toString());
                variableName = currentToken.getText();
                if (objects.containsKey(variableName)) {
                    object = objects.get(variableName);
                }
                else {
                    monitor.appendErrorMessage("Error. Impossible to find variable \"" + variableName + "\".");
                    return null;
                }
                return object;

            case number:
//                System.out.println(currentToken.toString());
                return Float.valueOf(currentToken.getText());

            case bool:
                return Boolean.valueOf(currentToken.getText());

            case quot:
//                System.out.println(currentToken.toString());
                return currentToken.getText().substring(1, currentToken.getText().length()-1);

            default:
                return null;
        }
    }

    private boolean local;
    private String variableName;

    private FNNXTreeNode tree;
    private Hashtable<String, Object> objects;
    private Hashtable<String, Object> localObjects;
    private Hashtable<String, Class> plugins;

    private InterThreadMonitor monitor;

}
