/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package scripting;

import java.io.Serializable;

/**
 *
 * @author hector
 */
public class InterThreadMonitor implements Serializable {
    
    /**
     * Creates a new instance of InterThreadMonitor. This object is used in order to synchronize threads.
     *
     */
    public InterThreadMonitor(int n) {
        isRunning = false;

        resultsIn = new InterThreadBuffer(n);
        resultsOut = new InterThreadBuffer(n);

        resultsBufferFull = false;

        paused = false;

        message = "";
        messageBufferFull = false;

        errorMessage = "";
        comment = "";

    }

    /**
     * Creates a new instance of InterThreadMonitor. This object is used in order to synchronize threads.
     *
     */
    public InterThreadMonitor() {
        isRunning = false;
        
        resultsIn = new InterThreadBuffer(0);
        resultsOut = new InterThreadBuffer(0);

        resultsBufferFull = false;
        
        paused = false;

        message = "";
        messageBufferFull = false;
        
        errorMessage = "";
        comment = "";

    }

//    public synchronized void changeConfiguration(int n) {
//        if (isRunning) {
//            while (isRunning && resultsBufferFull) {
//                System.err.println("\tMonitor - waiting to configure ...");
//                try {
//                    wait();
//                }
//                catch (InterruptedException ex) {
//                }
//            }
//            System.err.println("\tMonitor - configuring ...");
//
//            resultsIn = new InterThreadBuffer(n);
//            resultsOut = new InterThreadBuffer(n);
//
//            resultsBufferFull = false;
//
//            System.err.println("\tMonitor - configured ...");
//            notifyAll();
//        }
//    }

    /**
     * Returns the state of the flag {@link #isRunning}.
     *
     * @return  true, if the thread is running, flase otherwise
     */
//    public synchronized boolean isRunning() {
    public boolean isRunning() {
        return isRunning;
    }

    /**
     * Sets the flag {@link #isRunning} and saves the system hour and date.
     */
//    public synchronized void running() {
    public void running() {
        isRunning = true;
    }
    
    /**
     * Resets the flag {@link #isRunning}.
     */
    public synchronized void notRunning() {
        if (!resultsBufferFull && isRunning) {
            isRunning = false;
        }

        notify();
    }

    /**
     * Resets the flag {@link #isRunning}.
     */
    public synchronized void abort() {
        isRunning = false;
        resultsBufferFull = false;

        notify();
    }

    /**
     * Takes the new results and copies them to the plotter buffer.
     *
     * @param   r   the new results to take
     */
    public synchronized void newResults(InterThreadBuffer r) {
        if (isRunning) {
            while (isRunning && resultsBufferFull) {
//                System.err.println("\tMonitor - " + r.getCommand().toString() + " waiting...");
                try {
                    wait();
                }
                catch (InterruptedException ex) {
                }
            }

//            System.err.println("\tMonitor - " + r.getCommand().toString() + " blocking...");

            if (r.getCommand() == Names.interThreadCommands.init) {
                resultsIn = new InterThreadBuffer(r.getParameter());
                resultsOut = new InterThreadBuffer(r.getParameter());
            }
            resultsIn.copyFrom(r);

            while (paused && isRunning) {
                try {
                    wait();
                }
                catch (InterruptedException ex) {
                }
            }

            resultsBufferFull = true;
//            System.err.println("\tMonitor - " + r.getCommand().toString() + " releasing...");
            
            notify();
        }
    }

    /**
     * Returns the last results saved in the plotter buffer.
     *
     * @return  the last saved results
     */
    public synchronized InterThreadBuffer getResults() {
        while (!resultsBufferFull && isRunning) {
//            System.err.print("\t.");
            try {
                wait();
            }
            catch (InterruptedException ex) {
            }
        }
        
        if (isRunning) {
            resultsOut.copyFrom(resultsIn);
        }
        else {
            resultsOut.setEmpty();
        }

//        System.err.println("\tMonitor " + resultsIn.getCommand().toString() + "." + resultsOut.getCommand().toString() + " read ...");

        resultsBufferFull = false;
        notify();
        return resultsOut;
    }
    
    public synchronized void pause() {
        paused = true;
    }
    
    public synchronized void resume() {
        paused = false;
        notify();
    }
    
//    public synchronized boolean isPaused() {
    public boolean isPaused() {
        return paused;
    }

    public String toString() {
        return ("Monitor");
    }

    public void appendErrorMessage(String error) {
        errorMessage = errorMessage.concat(error).concat("\n");
    }

    public void appendComment(String com) {
        comment = comment.concat(com);
    }

    public synchronized void newMessage(String msg) {
        if (isRunning) {
            while (isRunning && messageBufferFull) {
                try {
                    wait();
                }
                catch (InterruptedException ex) {
                }
            }

            message = new String(msg);

            while (paused && isRunning) {
                try {
                    wait();
                }
                catch (InterruptedException ex) {
                }
            }

            messageBufferFull = true;
            notify();
        }
    }

    public synchronized String getMessage() {
        while (!messageBufferFull && isRunning) {
            try {
                wait();
            }
            catch (InterruptedException ex) {
            }
        }

        String toReturn = null;
        if (isRunning) {
            toReturn = new String(message);
        }
        messageBufferFull = false;
        notify();
        return toReturn;
    }

    public synchronized boolean hasMessage() {
        return messageBufferFull;
    }

    public synchronized String getErrorMessage() {
        String toReturn = new String(errorMessage);
        errorMessage = "";
        return toReturn;
    }

    public synchronized String getComment() {
        String toReturn = new String(comment);
        comment = "";
        return toReturn;
    }

    public synchronized boolean hasErrorMessage() {
        if (errorMessage.length() > 0) {
            return true;
        }
        else {
            return false;
        }
    }

    public synchronized boolean hasComment() {
        if (comment.length() > 0) {
            return true;
        }
        else {
            return false;
        }
    }

    public void resetErrorMessage() {
        errorMessage = "";
    }

    public void resetComment() {
        comment = "";
    }

    private InterThreadBuffer resultsIn;
    private InterThreadBuffer resultsOut;

    private boolean isRunning;
    
    private boolean resultsBufferFull;
    
    private boolean paused;

    private String message;
    private boolean messageBufferFull;

    private String errorMessage;
    private String comment;
}
