/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXCrossValidator;

import visualFNNXCrossValidator.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import java.util.Arrays;
import java.util.Hashtable;
import javax.swing.JFrame;
import networks.AbstractFNNXFeedForwardNetwork;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;

/**
 *
 * @author hsatizab
 */
public class FNNXCrossValidator extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXCrossValidator(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[7];
        parameterNames[0] = "network";
        parameterNames[1] = "datasets";
        parameterNames[2] = "trainer";
        parameterNames[3] = "validator";
        parameterNames[4] = "saver";
        parameterNames[5] = "tester";
        parameterNames[6] = "delay";

        parameterTypes = new Class[7];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = DataBlock[].class;
        parameterTypes[2] = AbstractFNNXFunction.class;
        parameterTypes[3] = AbstractFNNXFunction.class;
        parameterTypes[4] = AbstractFNNXParallelObserver.class;
        parameterTypes[5] = AbstractFNNXFunction.class;
        parameterTypes[6] = Number.class;

        isParameterMandatory = new boolean[7];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = true;
        isParameterMandatory[4] = false;
        isParameterMandatory[5] = false;
        isParameterMandatory[6] = false;

        parameters = new Object[7];

        //deafault values
        delay = 0;
    }

    public FNNXCrossValidator(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[7];
        parameterNames[0] = "network";
        parameterNames[1] = "datasets";
        parameterNames[2] = "trainer";
        parameterNames[3] = "validator";
        parameterNames[4] = "saver";
        parameterNames[5] = "tester";
        parameterNames[6] = "delay";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXFeedForwardNetwork)parameters[i];
                        break;
                    case 1:
                        datasets = (DataBlock[])parameters[i];
                        break;
                    case 2:
                        trainer = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 3:
                        validator = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 4:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 5:
                        tester = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 6:
                        delay = ((Number)parameters[i]).intValue();
                        if (!testMin(i, delay, 0)) {
                            return false;
                        }
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        if (datasets.length < 2) {
            monitor.appendErrorMessage("Error. Too few datasets to perform cross-validation.");
            return false;
        }

        //------------------------------------------------------------------
        if (!validator.configure(validator.getParameterIndex("dataset"), datasets[0], monitor)) {
            monitor.appendErrorMessage("Error. Impossible to configure tester.");
            return false;
        }
        if (!validator.configure(validator.getParameterIndex("network"), network, monitor)) {
            monitor.appendErrorMessage("Error. Impossible to configure tester.");
            return false;
        }
        validator.prepare(monitor);
        //------------------------------------------------------------------
        if (validator.getNumberOfReturningTraces() < 1) {
            monitor.appendErrorMessage("Error. The testing algorithm must return something.");
            return false;
        }

        if (tester != null) {
            hasTester = true;
        }
        else {
            hasTester = false;
        }

        res = new float[validator.getNumberOfReturningTraces()];
        results = new InterThreadBuffer(res.length);

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }
        this.initSaver();

        int i, j, k, l, index;
        int trainSize;
        float[] avg = new float[res.length];
        DataBlock training = new DataBlock(datasets[0].getNParts());
        boolean flag;

        Arrays.fill(avg, 0);

        for (i = 0; i < datasets.length; i++) {

            trainSize = 0;
            for (j = 0; j < datasets.length; j++) {
                if (i != j) {
                    trainSize += datasets[j].getNRows();
                }
            }
            for (j = 0; j < datasets[i].getNParts(); j++) {
                training.configDataset(j, trainSize, datasets[i].getDataset(j).getNColumns());
                for (k = 0, index = 0; k < datasets.length; k++) {
                    if (i != k) {
                        for (l = 0; l < datasets[k].getNRows(); l++, index++) {
                            training.getDataset(j).setRow(index, datasets[k].getDataset(j).getRow(l), false);
                        }
                    }
                }
            }
            training.copyStats(datasets[i]);

            network.initModel();

            flag = true;
            flag &= validator.configure(validator.getParameterIndex("dataset"), datasets[i], monitor);
            flag &= validator.configure(validator.getParameterIndex("network"), network, monitor);
            flag &= validator.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure validator.");
                return null;
            }

            flag = true;
            flag &= trainer.configure(trainer.getParameterIndex("dataset"), training, monitor);
//            flag &= trainer.configure(trainer.getParameterIndex("validator"), validator, monitor);
            flag &= trainer.configure(trainer.getParameterIndex("network"), network, monitor);
            flag &= trainer.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure trainer.");
                return null;
            }

            if (hasTester) {
                flag = true;
                flag &= tester.configure(tester.getParameterIndex("dataset"), training, monitor);
                flag &= tester.configure(tester.getParameterIndex("network"), network, monitor);
                flag &= tester.prepare(monitor);
                if (!flag) {
                    monitor.appendErrorMessage("Error. Impossible to configure tester.");
                    return null;
                }
//                tester.eval(monitor);
//                if (monitor.hasErrorMessage()) {
//                    return null;
//                }
//                if (delay > 0) {
//                    try {
//                        Thread.sleep(delay);
//                    }
//                    catch (InterruptedException ie) {
//                    }
//                }
            }

            trainer.eval(monitor);
            if (monitor.hasErrorMessage()) {
                return null;
            }

            if (hasTester) {
                tester.eval(monitor);
                if (monitor.hasErrorMessage()) {
                    return null;
                }
                if (delay > 0) {
                    try {
                        Thread.sleep(delay);
                    }
                    catch (InterruptedException ie) {
                    }
                }
            }

            try {
                res = (float[])validator.eval(monitor);
            }
            catch (ClassCastException ce) {
                monitor.appendErrorMessage("Error. The tester does not return an appropriate object.");
                return null;
            }
            if (res == null) {
                monitor.appendErrorMessage("Error. The tester did not return.");
                return null;
            }

            if (hasSaver) {
                results.newData(res, i);
                saver.getMonitor().newResults(results);
            }
            for (j = 0; j < avg.length; j++) {
                avg[j] += res[j];
            }
        }

        this.endSaver();

        for (j = 0; j < avg.length; j++) {
            avg[j] /= datasets.length;
        }
        return avg;
    }

    private AbstractFNNXFeedForwardNetwork network;
    private AbstractFNNXFunction trainer;
    private AbstractFNNXFunction validator;
    private DataBlock[] datasets;
    private AbstractFNNXFunction tester;
    private boolean hasTester;
    private int delay;

}
