/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXExportGNGNetwork;

import cellGNG.AbstractGNGCell;
import cellGNG.Edge;
import visualFNNXExportGNGNetwork.ConfigurationWindow;
import plugin.*;
import java.util.Hashtable;
import javax.swing.JFrame;
import networkGNG.AbstractFNNXGNGNetwork;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;


/**
 *
 * @author hsatizab
 */
public class FNNXExportGNGNetwork extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXExportGNGNetwork(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[2];
        parameterNames[0] = "network";
        parameterNames[1] = "saver";

        parameterTypes = new Class[2];
        parameterTypes[0] = AbstractFNNXGNGNetwork.class;
        parameterTypes[1] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[2];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;

        parameters = new Object[2];
    }

    public FNNXExportGNGNetwork(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[2];
        parameterNames[0] = "network";
        parameterNames[1] = "saver";

    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        try {
            network = (AbstractFNNXGNGNetwork)parameters[0];
            saver = (AbstractFNNXParallelObserver)parameters[1];

            if (network == null) {
                monitor.appendErrorMessage("Error. No network to export.");
                return false;
            }
            if (saver == null) {
                monitor.appendErrorMessage("Error. No saver.");
                return false;
            }

            res = new float[network.getAmountOfInputs() + 1];
            results = new InterThreadBuffer(res.length);

        }
        catch(NullPointerException ne) {
            monitor.appendErrorMessage("Internal error. Null pointer.");
            return false;
        }
        catch(IndexOutOfBoundsException ie) {
            monitor.appendErrorMessage("Internal error. Index out of bounds.");
            return false;
        }
        catch(ClassCastException ce) {
            monitor.appendErrorMessage("Internal error. Class cast.");
            return false;
        }

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        String message = "";
        int c = 0;
        AbstractGNGCell temp;
        float[] pos;
        Edge edge;

        this.initSaver();
        
        for (c = 1; c <= network.getAmountOfInputs(); c++) {
            message += "\tI_" + String.valueOf(c);
        }
        message += "\tweight\tNeighbours\tIndexes\n";

        saver.getMonitor().newMessage(message);
        results.setCommand(Names.interThreadCommands.writeMessage, 0);
        saver.getMonitor().newResults(results);

        network.inventory();
        network.firstCell();
        while (network.hasCell()) {
            temp = (AbstractGNGCell) network.getCellDoNext();
            pos = temp.getPosition();

            System.arraycopy(pos, 0, res, 0, pos.length);
            res[res.length-1] = temp.getWeight();

            results.newData(res, temp.getNumber());

//            while (saver.getMonitor().hasErrorMessage()) {
//                try {
//                    Thread.sleep(10);
//                }
//                catch (InterruptedException ex) {
//                }
//            }

            message = String.valueOf(temp.getAmountOfEdges()) + " :";          //writing number of neighbours
            temp.firstEdge();
            while (temp.hasEdge()) {
                edge = temp.getEdgeDoNext();
                message += "\t" + String.valueOf(edge.getUnitConnectedTo(temp).getNumber());    //writing index of neighbours
            }

            saver.getMonitor().newMessage(message);
//            saver.getMonitor().appendComment(message);
            saver.getMonitor().newResults(results);

        }

        this.endSaver();
        
        return null;

    }

    private AbstractFNNXGNGNetwork network;

}
