/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXGNGAlgorithm;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import networkGNG.AbstractFNNXGNGNetwork;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import visualFNNXGNGAlgorithm.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXGNGAlgorithm extends AbstractFNNXTrainingAlgorithm implements FNNXObject, FNNXFunction, FNNXTrainingAlgorithm {

    public FNNXGNGAlgorithm(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[19];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "eBest";
        parameterNames[3] = "eNeighors";
        parameterNames[4] = "aMax";
        parameterNames[5] = "lambda";
        parameterNames[6] = "alfa";
        parameterNames[7] = "d";
        parameterNames[8] = "qError";
        parameterNames[9] = "hold";
        parameterNames[10] = "superposition";
        parameterNames[11] = "weightIncrement";
        parameterNames[12] = "weightDecrement";
        parameterNames[13] = "size";
        parameterNames[14] = "loops";
        parameterNames[15] = "stopIfTesterBelow";
        parameterNames[16] = "logCounter";
        parameterNames[17] = "plotter";
        parameterNames[18] = "saver";

        parameterTypes = new Class[19];
        parameterTypes[0] = AbstractFNNXGNGNetwork.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = Number.class;
        parameterTypes[3] = Number.class;
        parameterTypes[4] = Number.class;
        parameterTypes[5] = Number.class;
        parameterTypes[6] = Number.class;
        parameterTypes[7] = Number.class;
        parameterTypes[8] = Number.class;
        parameterTypes[9] = Number.class;
        parameterTypes[10] = Number.class;
        parameterTypes[11] = Number.class;
        parameterTypes[12] = Number.class;
        parameterTypes[13] = Number.class;
        parameterTypes[14] = Number.class;
        parameterTypes[15] = Number.class;
        parameterTypes[16] = Number.class;
        parameterTypes[17] = AbstractFNNXParallelObserver.class;
        parameterTypes[18] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[19];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = true;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = true;
        isParameterMandatory[6] = true;
        isParameterMandatory[7] = true;
        isParameterMandatory[8] = true;
        isParameterMandatory[9] = true;
        isParameterMandatory[10] = true;
        isParameterMandatory[11] = true;
        isParameterMandatory[12] = true;
        isParameterMandatory[13] = true;
        isParameterMandatory[14] = true;
        isParameterMandatory[15] = false;
        isParameterMandatory[16] = false;
        isParameterMandatory[17] = false;
        isParameterMandatory[18] = false;

        parameters = new Object[19];

        //default values
        logCounter = 100;
        stopIfTesterBelow = 0;
    }

    public FNNXGNGAlgorithm(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[19];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "eBest";
        parameterNames[3] = "eNeighors";
        parameterNames[4] = "aMax";
        parameterNames[5] = "lambda";
        parameterNames[6] = "alfa";
        parameterNames[7] = "d";
        parameterNames[8] = "qError";
        parameterNames[9] = "hold";
        parameterNames[10] = "superposition";
        parameterNames[11] = "weightIncrement";
        parameterNames[12] = "weightDecrement";
        parameterNames[13] = "size";
        parameterNames[14] = "loops";
        parameterNames[15] = "stopIfTesterBelow";
        parameterNames[16] = "logCounter";
        parameterNames[17] = "plotter";
        parameterNames[18] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXGNGNetwork)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        network.setDataset(dataset);
                        break;
                    case 2:
                        eBest = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, eBest, 0)) {
                            return false;
                        }
                        break;
                    case 3:
                        eNeighbors = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, eNeighbors, 0)) {
                            return false;
                        }
                        break;
                    case 4:
                        aMax = ((Number)parameters[i]).intValue();
                        if (!testMin(i, aMax, 0)) {
                            return false;
                        }
                        break;
                    case 5:
                        lambda = ((Number)parameters[i]).intValue();
                        if (!testMin(i, lambda, 0)) {
                            return false;
                        }
                        break;
                    case 6:
                        alfa = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, alfa, 0)) {
                            return false;
                        }
                        break;
                    case 7:
                        d = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, d, 0)) {
                            return false;
                        }
                        if (!testMax(i, d, 1)) {
                            return false;
                        }
                        break;
                    case 8:
                        qError = (float)Math.pow(((Number)parameters[i]).doubleValue(), 2.0);
                        if (!testMin(i, qError, 0)) {
                            return false;
                        }
                        break;
                    case 9:
                        hold = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, hold, 0)) {
                            return false;
                        }
                        if (!testMax(i, hold, 1)) {
                            return false;
                        }
                        break;
                    case 10:
                        superposition = ((Number)parameters[i]).intValue();
                        if (!testMin(i, superposition, 0)) {
                            return false;
                        }
                        if (!testMax(i, superposition, 100)) {
                            return false;
                        }
                        superposition = 100/superposition;
                        break;
                    case 11:
                        weightIncrement = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, weightIncrement, 0)) {
                            return false;
                        }
                        break;
                    case 12:
                        weightDecrement = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, weightDecrement, 0)) {
                            return false;
                        }
                        break;
                    case 13:
                        size = ((Number)parameters[i]).intValue();
                        if (!testMin(i, size, 0)) {
                            return false;
                        }
                        break;
                    case 14:
                        epochs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, epochs, 0)) {
                            return false;
                        }
                        break;
                    case 15:
                        stopIfTesterBelow = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, stopIfTesterBelow, 0)) {
                            return false;
                        }
                        break;
                    case 16:
                        logCounter = ((Number)parameters[i]).intValue();
                        if (!testMin(i, logCounter, 0)) {
                            return false;
                        }
                        break;
                    case 17:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 18:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        res = new float[1];
        results = new InterThreadBuffer(1);

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }
    
    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int loop;
        int iteration;
        int counter = 1;
       
        this.initSaver();

        this.initPlotter(dataset.getNRows()*epochs, 1);

        for (loop = 0; (loop < epochs) && (monitor.isRunning()); loop++) {
            while (monitor.isPaused());

            network.nextExampleRandom();
            
//            long time = System.currentTimeMillis();

            for (iteration = 0; (iteration < dataset.getNRows()) && monitor.isRunning(); network.nextExampleRandom(), iteration++, counter++) {
                while (monitor.isPaused());

                singleStep(counter);

                if ((counter % logCounter) == 0) {
                    if ((network.getAccError()/logCounter) < stopIfTesterBelow) {
                        iteration = dataset.getNRows();
                        loop = epochs;
                    }
                    this.log(counter);
                }
            }

//            time = System.currentTimeMillis() - time;
//            System.err.println(time);

            ((AbstractFNNXGNGNetwork)network).forget();                                                   //delete unused units

        }

        this.endPlotter();
        this.endSaver();

        return null;
    }

    public void evaluateModel(int it) {
        ((AbstractFNNXGNGNetwork)network).calcNearest(d);
    }

    public void updateModel(int it) {
//        ((AbstractFNNXGNGNetwork)network).trainStep(it, eBest, eNeighbors, aMax, lambda, alfa, d, qError, hold, superposition, size);
        ((AbstractFNNXGNGNetwork)network).trainStep(it, eBest, eNeighbors, aMax, lambda, alfa, qError, hold, superposition, weightIncrement, weightDecrement, size);

//        network.near_1.incrementAge();                                          //3. Increment the age of all edges emanating from s1.
//
//        network.addToError(network.dNear_1);                                    //cumulated error -> plot
//
//        //---------------------------------------
//        if (network.dNear_1 > qError) {
//            network.near_1.update(network.dNear_1, eBest, eNeighbors, network.near_2);
//                                                                                //4. Add the squared distance between the input signal and the nearest unit in input space to a local counter variable.
//                                                                                //5. Move s1 and its direct topological neighbors towards I.
//                                                                                //6. If s1 and s2 are connected by an edge, set the age of this edge to zero. If such an edge does not exist, create it.
//        }
//        else {
//        //---------------------------------------
//
//            network.near_1.update(network.dNear_1*hold, eBest*hold, eNeighbors*hold, network.near_2);
//                                                                                //4. Add the squared distance between the input signal and the nearest unit in input space to a local counter variable.
//                                                                                //5. Move s1 and its direct topological neighbors towards I.
//                                                                                //6. If s1 and s2 are connected by an edge, set the age of this edge to zero. If such an edge does not exist, create it.
//        }
//
//        network.near_1.deleteOlderConnections(aMax);                            //7. Remove edges with an age larger than a_max.
//        //neuralGas.forget(aMax);                                               //  If this results in points having no emanating edges, remove them as well.
//
//        if ((it%lambda == 0) && (network.getNumberOfCells() < size)){
//            network.insertUnit(alfa, superposition, qError);                    //8. Insert new unit
//        }
//
////        network.decreaseErrors(d);                                              //9. Decrease all error variables by multiplying them with a constant d.


    }

    private float eBest;
    private float eNeighbors;
    private int aMax;
    private int lambda;
    private float alfa;
    private float d;
    private float qError;
    private float hold;
    private float superposition;
    private float weightIncrement;
    private float weightDecrement;
    private int size;

}
