/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXGNGLLMNetwork;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import layerGNGLLM.FNNXGNGLLMLayerNetwork;
import networkGNGLLM.AbstractFNNXGNGLLM;
import scripting.InterThreadMonitor;
import visualFNNXGNGLLMNetwork.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXGNGLLMNetwork extends AbstractFNNXGNGLLM implements FNNXObject {

    public FNNXGNGLLMNetwork(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;

        parameterNames = new String[3];
        parameterNames[0] = "inputs";
        parameterNames[1] = "outputs";
        parameterNames[2] = "dataset";

        parameterTypes = new Class[3];
        parameterTypes[0] = Number.class;
        parameterTypes[1] = Number.class;
        parameterTypes[2] = DataBlock.class;

        isParameterMandatory = new boolean[3];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;

        parameters = new Object[3];
    }

    public FNNXGNGLLMNetwork(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[3];
        parameterNames[0] = "inputs";
        parameterNames[1] = "outputs";
        parameterNames[2] = "dataset";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        nInputs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, nInputs, 0)) {
                            return false;
                        }
                        break;

                    case 1:
                        nOutputs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, nOutputs, 0)) {
                            return false;
                        }
                        
                        competitionLayer = new FNNXGNGLLMLayerNetwork(nInputs, nOutputs);
                        if (monitor.hasErrorMessage()) {
                            monitor.appendErrorMessage("Error. Problems creating competition layer for \"" + name + "\".");
                            return false;
                        }
                        break;

                    case 2:
                        if (!setDataset((DataBlock)parameters[i])) {
                            monitor.appendErrorMessage("Error. Problems connecting \"" + this.name + " \" to dataset.");
                            return false;
                        }
                        break;
                }
            }
            catch(NullPointerException ne) {
                monitor.appendErrorMessage("Error. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name);
                return false;
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        this.initModel();

        configured = true;
        return true;
    }

}
