/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package cellGNGLLM;

import cellGNG.AbstractGNGCell;
import data.MyLinkedList;
import java.util.Random;

/**
 *
 * @author hsatizab
 */
public class GNGLLMCell extends AbstractGNGCell {

    /** Creates a new instance of GNGLLMCell */
    public GNGLLMCell(int nI, int nO, Random r) {
        edges = new MyLinkedList();

        position = new float[nI];
        dist = new float[nI];

        theta = new float[nO];
        A = new float[nO][nI];
        output = new float[nO];
        cellError = new float[nO];

        int i, j;
        for (j = 0; j < nO; j++) {
            theta[j] = r.nextFloat() - (float)0.5;
            for (i = 0; i < nI; i++) {
                A[j][i] = r.nextFloat() - (float)0.5;
            }
        }
    }

    /** Creates a new instance of GNGLLMCell */
    public GNGLLMCell(GNGLLMCell a, GNGLLMCell b) {
        int nI = a.position.length;
        int nO = a.output.length;
//        if (nI != b.getPosition().length) {
//            throw new NullPointerException("Cells with different dimensions");
//        }

        edges = new MyLinkedList();

        position = new float[nI];
        dist = new float[nI];

        for (int i = 0; i < nI; i++) {
            position[i] = (a.getPosition()[i] + b.getPosition()[i]) / 2;
        }

        theta = new float[nO];
        A = new float[nO][nI];
        output = new float[nO];
        cellError = new float[nO];

        int i, j;
        for (j = 0; j < nO; j++) {
            theta[j] = (a.theta[j] + b.theta[j])/2;
            for (i = 0; i < nI; i++) {
                A[j][i] = (a.A[j][i] + b.A[j][i])/2;
            }
        }
    }

    public int getAmountOfOutputs() {
        return output.length;
    }

    public float[] calcOutput(float[] input) {
        int i, j;

        this.calcDistanceTo(input);
//        for (i = 0; i < input.length; i++) {
//            temp[i] = input[i] - position[i];
//        }

        for (j = 0; j < output.length; j++) {
            output[j] = 0;
            for (i = 0; i < input.length; i++) {
                output[j] += A[j][i] * dist[i];
            }
            output[j] += theta[j];
        }

        return output;
    }

    public float calcLocalError(float[] target) {
        int j;
        localError = 0;

        for (j = 0; j < target.length; j++) {
            cellError[j] = target[j] - output[j];
            localError += (cellError[j] * cellError[j]);
        }
        return localError;
    }

    public float[] calcLocalErrorByOutput(float[] target) {
        int j;
        localError = 0;

        for (j = 0; j < target.length; j++) {
            cellError[j] = target[j] - output[j];
            localError += (cellError[j] * cellError[j]);
        }
        return cellError;
    }

    public void updateSupervised(float eta, float[] input, float[] target) {
        int i, j;

//        calcOutput(input);
//        calcLocalError(target);

        for (j = 0; j < target.length; j++) {
            theta[j] += eta * cellError[j];

            for (i = 0; i < input.length; i++) {
                A[j][i] += eta * cellError[j] * dist[i];
            }
        }
    }

    private float[] theta;
    private float[][] A;
    private float[] output;
    private float[] cellError;

}
