/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXGNGMapAlgorithm;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import networkGNGMap.AbstractFNNXGNGMap;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import visualFNNXGNGMapAlgorithm.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXGNGMapAlgorithm extends AbstractFNNXTrainingAlgorithm implements FNNXObject, FNNXFunction, FNNXTrainingAlgorithm {

    public FNNXGNGMapAlgorithm(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[16];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "tester";
        parameterNames[3] = "validator";
        parameterNames[4] = "inputTrainer";
        parameterNames[5] = "outputTrainer";
        parameterNames[6] = "linkIncrement";
        parameterNames[7] = "maxLinkWeight";
        parameterNames[8] = "decayDivergence";
        parameterNames[9] = "decayConvergence";
        parameterNames[10] = "decay";
        parameterNames[11] = "epochs";
        parameterNames[12] = "stopIfTesterBelow";
        parameterNames[13] = "logCounter";
        parameterNames[14] = "plotter";
        parameterNames[15] = "saver";

        parameterTypes = new Class[16];
        parameterTypes[0] = AbstractFNNXGNGMap.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXFunction.class;
        parameterTypes[3] = AbstractFNNXFunction.class;
        parameterTypes[4] = AbstractFNNXTrainingAlgorithm.class;
        parameterTypes[5] = AbstractFNNXTrainingAlgorithm.class;
        parameterTypes[6] = Number.class;
        parameterTypes[7] = Number.class;
        parameterTypes[8] = Number.class;
        parameterTypes[9] = Number.class;
        parameterTypes[10] = Number.class;
        parameterTypes[11] = Number.class;
        parameterTypes[12] = Number.class;
        parameterTypes[13] = Number.class;
        parameterTypes[14] = AbstractFNNXParallelObserver.class;
        parameterTypes[15] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[16];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;
        isParameterMandatory[3] = false;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = true;
        isParameterMandatory[6] = true;
        isParameterMandatory[7] = true;
        isParameterMandatory[8] = true;
        isParameterMandatory[9] = true;
        isParameterMandatory[10] = true;
        isParameterMandatory[11] = true;
        isParameterMandatory[12] = false;
        isParameterMandatory[13] = false;
        isParameterMandatory[14] = false;
        isParameterMandatory[15] = false;

        parameters = new Object[16];

        //default values
        stopIfTesterBelow = 0;
        logCounter = 100;

    }

    public FNNXGNGMapAlgorithm(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[16];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "tester";
        parameterNames[3] = "validator";
        parameterNames[4] = "inputTrainer";
        parameterNames[5] = "outputTrainer";
        parameterNames[6] = "linkIncrement";
        parameterNames[7] = "maxLinkWeight";
        parameterNames[8] = "decayDivergence";
        parameterNames[9] = "decayConvergence";
        parameterNames[10] = "decay";
        parameterNames[11] = "epochs";
        parameterNames[12] = "stopIfTesterBelow";
        parameterNames[13] = "logCounter";
        parameterNames[14] = "plotter";
        parameterNames[15] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        boolean flag;
        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXGNGMap)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        tester = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 3:
                        validator = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 4:
                        inputTrainer = (AbstractFNNXTrainingAlgorithm)parameters[i];
                        break;
                    case 5:
                        outputTrainer = (AbstractFNNXTrainingAlgorithm)parameters[i];
                        break;
                    case 6:
                        linkIncrement = ((Number)parameters[i]).intValue();
                        if (!testMin(i, linkIncrement, 0)) {
                            return false;
                        }
                        break;
                    case 7:
                        maxLinkWeight = ((Number)parameters[i]).intValue();
                        if (!testMin(i, maxLinkWeight, 0)) {
                            return false;
                        }
                        break;
                    case 8:
                        decayDivergence = ((Number)parameters[i]).intValue();
                        if (!testMin(i, decayDivergence, 0)) {
                            return false;
                        }
                        break;
                    case 9:
                        decayConvergence = ((Number)parameters[i]).intValue();
                        if (!testMin(i, decayConvergence, 0)) {
                            return false;
                        }
                        break;
                    case 10:
                        decay = ((Number)parameters[i]).intValue();
                        if (!testMin(i, decay, 0)) {
                            return false;
                        }
                        break;
                    case 11:
                        epochs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, epochs, 0)) {
                            return false;
                        }
                        break;
                    case 12:
                        stopIfTesterBelow = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, stopIfTesterBelow, 0)) {
                            return false;
                        }
                        break;
                    case 13:
                        logCounter = ((Number)parameters[i]).intValue();
                        if (!testMin(i, logCounter, 0)) {
                            return false;
                        }
                        break;
                    case 14:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 15:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Impossible to set dataset in " + this.name);
            return false;
        }
        
        flag = true;
        flag &= inputTrainer.configure(inputTrainer.getParameterIndex("dataset"), new DataBlock(dataset.getDataset(0)), monitor);
        flag &= inputTrainer.configure(inputTrainer.getParameterIndex("network"), ((AbstractFNNXGNGMap)network).getInputLayer(), monitor);
        flag &= inputTrainer.configure(inputTrainer.getParameterIndex("loops"), epochs, monitor);
        flag &= inputTrainer.prepare(monitor);
        if (!flag) {
            monitor.appendErrorMessage("Error. Impossible to configure inputTrainer in " + this.name);
            return false;
        }

        flag = true;
        flag &= outputTrainer.configure(outputTrainer.getParameterIndex("dataset"), new DataBlock(dataset.getDataset(1)), monitor);
        flag &= outputTrainer.configure(outputTrainer.getParameterIndex("network"), ((AbstractFNNXGNGMap)network).getOutputLayer(), monitor);
        flag &= outputTrainer.configure(outputTrainer.getParameterIndex("loops"), epochs, monitor);
        flag &= outputTrainer.prepare(monitor);
        if (!flag) {
            monitor.appendErrorMessage("Error. Impossible to configure outputTrainer in" + this.name);
            return false;
        }

        int resBufferSize = 0;
        if (tester != null) {
            flag = true;
            flag &= tester.configure(tester.getParameterIndex("dataset"), dataset, monitor);
            flag &= tester.configure(tester.getParameterIndex("network"), network, monitor);
            flag &= tester.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasTester = true;
            resBufferSize += tester.getNumberOfReturningTraces();
        }
        else {
            hasTester = false;
        }

        if (validator != null) {
            flag = true;
            flag &= validator.configure(validator.getParameterIndex("network"), network, monitor);
            flag &= validator.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasValidator = true;
            resBufferSize += validator.getNumberOfReturningTraces();
        }
        else {
            hasValidator = false;
        }

        res = new float[resBufferSize];
        results = new InterThreadBuffer(resBufferSize);

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }
    
    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int it;
        int posCopy;
        int expectedTestLength = 0;
        int expectedValidationLength = 0;
        int datasetLength;
        int limit;
        float[] resultTest;
        float[] resultVal;
        boolean stop = false;

        datasetLength = dataset.getNRows();
        limit = epochs * datasetLength;

        this.initPlotter(limit, 1);
        inputTrainer.initPlotter(limit, 1);
        outputTrainer.initPlotter(limit, 1);

        this.initSaver();
        inputTrainer.initSaver();
        outputTrainer.initSaver();

        posCopy = 0;
        if (hasTester) {
            posCopy = tester.getNumberOfReturningTraces();
            expectedTestLength = tester.getNumberOfReturningTraces();
        }
        if (hasValidator) {
            expectedValidationLength = validator.getNumberOfReturningTraces();
        }

        for (it = 0; (it < limit) && (monitor.isRunning()); it++) {
            while (monitor.isPaused());

            //------------------------------------------------------------------
            network.nextExampleRandom();
//            for (iteration = 0; (iteration < dataset.getNRows()) && monitor.isRunning(); network.nextExampleRandom(), iteration++) {
//                while (monitor.isPaused());
//
//                this.singleStep(accIt + iteration + 1);
//            }
            this.singleStep(it + 1);

            if ((it % datasetLength) == 0) {
                ((AbstractFNNXGNGMap)network).forget();
            }
            //------------------------------------------------------------------
            if ((it % logCounter) == 0) {
                if (hasTester) {
                    resultTest = (float[])tester.eval(monitor);
                    if (resultTest == null) {
                        if (expectedTestLength != 0) {
                            monitor.appendErrorMessage("Error. " + tester.toString() + " failed.");
                            return null;
                        }
                    }
                    else {
                        System.arraycopy(resultTest, 0, res, 0, resultTest.length);
                        if (resultTest[0] < stopIfTesterBelow) {
                            stop = true;                                            //stops the algorithm
                        }
                    }
                }

                if (hasValidator) {
                    resultVal = (float[])validator.eval(monitor);
                    if (resultVal == null) {
                        if (expectedValidationLength != 0) {
                            monitor.appendErrorMessage("Error. " + validator.toString() + "failed.");
                            return null;
                        }
                    }
                    else {
                        System.arraycopy(resultVal, 0, res, posCopy, resultVal.length);
                    }
                                                                 //this is necesary because the validator attachs the validation dataset to the network
                    if (!network.setDataset(dataset)) {
                        monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
                        return null;
                    }
                }

                results.newData(res, it+1);
                //----------------------------------------------
                if (hasPlotter && (hasTester || hasValidator)) {
                    plotter.getMonitor().newResults(results);
                }
                if (hasSaver && (hasTester || hasValidator)) {
                    saver.getMonitor().newResults(results);
                }
                //----------------------------------------------

                if (stop) {
                    it = limit;                                                    //stops the algorithm
                }
            }

        }

        this.endPlotter();
        this.endSaver();

        inputTrainer.endPlotter();
        outputTrainer.endPlotter();

        inputTrainer.endSaver();
        outputTrainer.endSaver();

        return null;
    }

    public void singleStep(int it) {
        inputTrainer.singleStep(it);
        outputTrainer.singleStep(it);

        ((AbstractFNNXGNGMap)network).updateLinks(linkIncrement, maxLinkWeight, decayDivergence, decayConvergence);
        ((AbstractFNNXGNGMap)network).getInputLayer().decreaseLinks(decay);

        ((AbstractFNNXGNGMap)network).deleteUnusedLinks();
    }

    public void evaluateModel(int it) {
        inputTrainer.evaluateModel(it);
        outputTrainer.evaluateModel(it);

        ((AbstractFNNXGNGMap)network).calcOutput();
        ((AbstractFNNXGNGMap)network).calcModelError();
    }

    public void updateModel(int it) {
        inputTrainer.updateModel(it);
        outputTrainer.updateModel(it);

        ((AbstractFNNXGNGMap)network).updateLinks(linkIncrement, maxLinkWeight, decayDivergence, decayConvergence);
        ((AbstractFNNXGNGMap)network).getInputLayer().decreaseLinks(decay);
        ((AbstractFNNXGNGMap)network).deleteUnusedLinks();
    }

    private AbstractFNNXTrainingAlgorithm inputTrainer;
    private AbstractFNNXTrainingAlgorithm outputTrainer;
    private int linkIncrement;
    private int maxLinkWeight;
    private int decayDivergence;
    private int decayConvergence;
    private int decay;
}
