/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package networkGNGMap;

import cellGNGMap.GNGMapCell;
import cellGNGMap.GNGMapLink;
import data.DataBlock;
import java.util.Random;
import layerGNGMap.FNNXGNGMapLayerNetwork;
import model.AbstractFNNXSupervisedModel;
import model.FNNXModel;

/**
 *
 * @author hsatizab
 */
public abstract class AbstractFNNXGNGMap extends AbstractFNNXSupervisedModel implements FNNXModel {

    public int inventory() {
        return inputLayer.inventory() + outputLayer.inventory();
    }

    public void initModel() {
        gen = new Random();
        inputLayer.initModel();
        outputLayer.initModel();
    }

    public void nextExampleRandom() {
        super.nextExampleRandom();
        inputLayer.setCurrentInputNoCopy(currentInput);
        outputLayer.setCurrentInputNoCopy(currentTarget);

//        inputLayer.nextExampleRandom();
//        outputLayer.nextExampleRandom();
    }

    public boolean hasNextExampleSequential() {
        if(super.hasNextExampleSequential()) {
            inputLayer.setCurrentInputNoCopy(currentInput);
            outputLayer.setCurrentInputNoCopy(currentTarget);
            return true;
        }
        else {
            return false;
        }
    }

    public FNNXGNGMapLayerNetwork getInputLayer() {
        return inputLayer;
    }

    public FNNXGNGMapLayerNetwork getOutputLayer() {
        return outputLayer;
    }

    public boolean setDataset(DataBlock data) {
        if (!super.setDataset(data)) {
            return false;
        }

        if (!inputLayer.setDataset(inputDataset)) {
            return false;
        }
        if (!outputLayer.setDataset(outputDataset)) {
            return false;
        }

        return true;
    }

    public void forget() {
        inputLayer.forget();
        outputLayer.forget();
    }

    public void updateLinks(int inc, int max, int dD, int dC) {
        GNGMapCell firstIn = (GNGMapCell)inputLayer.getNearest();
        GNGMapCell firstOut = (GNGMapCell)outputLayer.getNearest();

        firstIn.linkTo(firstOut, inc, max);

        firstIn.decayDifferentBy(firstOut, dD);
        firstOut.decayDifferentBy(firstIn, dC);
    }

    public void deleteUnusedLinks() {
        inputLayer.deleteUnusedLinks();
        outputLayer.deleteUnusedLinks();
    }

    public float[] calcOutput() {
        GNGMapCell firstIn;
        GNGMapLink tempLink;
        float temp;

        inputLayer.calcNearest(1);
        firstIn = (GNGMapCell)inputLayer.getNearest();

        for (int o = 0; o < nOutputs; o++) {
            output[o] = 0;
            temp = 0;

            firstIn.firstLink();
            while (firstIn.hasLink()) {
                tempLink = firstIn.getLinkDoNext();
                temp += tempLink.getWeight();
                output[o] += tempLink.getWeight() * tempLink.getTo().getPosition()[o];
            }

            if (firstIn.getAmountOfLinks() > 0) {
                output[o] /= temp;
            }
            else {
                output[o] = outputDataset.getMean()[o];
            }
        }

        return output;
    }

    public float calcModelError() {
        int o;

        error = 0;
        for (o = 0; o < this.nOutputs; o++) {
            error += (output[o] - currentTarget[o]) * (output[o] - currentTarget[o]);
        }

        return error;
    }

    public float[] calcModelErrorByOutput() {
        float[] e = new float[this.nOutputs];

        error = 0;
        for (int o = 0; o < this.nOutputs; o++) {
            e[o] = output[o] - currentTarget[o];
            error += e[o] * e[o];
        }

        return e;
    }

    public void reset() {
    }

    protected FNNXGNGMapLayerNetwork inputLayer;
    protected FNNXGNGMapLayerNetwork outputLayer;
}
