/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXGNGRBFAlgorithm;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import networkGNGRBF.AbstractFNNXGNGRBF;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import visualFNNXGNGRBFAlgorithm.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXGNGRBFAlgorithm extends AbstractFNNXTrainingAlgorithm implements FNNXObject, FNNXFunction, FNNXTrainingAlgorithm {

    public FNNXGNGRBFAlgorithm(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[11];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "tester";
        parameterNames[3] = "validator";
        parameterNames[4] = "networkTrainer";
        parameterNames[5] = "eta";
        parameterNames[6] = "epochs";
        parameterNames[7] = "stopIfTesterBelow";
        parameterNames[8] = "logCounter";
        parameterNames[9] = "plotter";
        parameterNames[10] = "saver";

        parameterTypes = new Class[11];
        parameterTypes[0] = AbstractFNNXGNGRBF.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXFunction.class;
        parameterTypes[3] = AbstractFNNXFunction.class;
        parameterTypes[4] = AbstractFNNXTrainingAlgorithm.class;
        parameterTypes[5] = Number.class;
        parameterTypes[6] = Number.class;
        parameterTypes[7] = Number.class;
        parameterTypes[8] = Number.class;
        parameterTypes[9] = AbstractFNNXParallelObserver.class;
        parameterTypes[10] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[11];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;
        isParameterMandatory[3] = false;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = true;
        isParameterMandatory[6] = true;
        isParameterMandatory[7] = false;
        isParameterMandatory[8] = false;
        isParameterMandatory[9] = false;
        isParameterMandatory[10] = false;

        parameters = new Object[11];

        //default values
        stopIfTesterBelow = 0;
        logCounter = 100;

    }

    public FNNXGNGRBFAlgorithm(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[11];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "tester";
        parameterNames[3] = "validator";
        parameterNames[4] = "networkTrainer";
        parameterNames[5] = "eta";
        parameterNames[6] = "epochs";
        parameterNames[7] = "stopIfTesterBelow";
        parameterNames[8] = "logCounter";
        parameterNames[9] = "plotter";
        parameterNames[10] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        boolean flag;
        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXGNGRBF)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        tester = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 3:
                        validator = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 4:
                        networkTrainer = (AbstractFNNXTrainingAlgorithm)parameters[i];
                        break;
                    case 5:
                        eta = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, eta, 0)) {
                            return false;
                        }
                        break;
                    case 6:
                        epochs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, epochs, 0)) {
                            return false;
                        }
                        break;
                    case 7:
                        stopIfTesterBelow = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, stopIfTesterBelow, 0)) {
                            return false;
                        }
                        break;
                    case 8:
                        logCounter = ((Number)parameters[i]).intValue();
                        if (!testMin(i, logCounter, 0)) {
                            return false;
                        }
                        break;
                    case 9:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 10:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Impossible to set dataset in " + this.name);
            return false;
        }
        
        flag = true;
        flag &= networkTrainer.configure(networkTrainer.getParameterIndex("dataset"), new DataBlock(dataset.getDataset(0)), monitor);
        flag &= networkTrainer.configure(networkTrainer.getParameterIndex("network"), ((AbstractFNNXGNGRBF)network).getCompetitionLayer(), monitor);
        flag &= networkTrainer.configure(networkTrainer.getParameterIndex("loops"), epochs, monitor);
        flag &= networkTrainer.prepare(monitor);
        if (!flag) {
            monitor.appendErrorMessage("Error. Impossible to configure inputTrainer in " + this.name);
            return false;
        }

        int resBufferSize = 0;
        if (tester != null) {
            flag = true;
            flag &= tester.configure(tester.getParameterIndex("dataset"), dataset, monitor);
            flag &= tester.configure(tester.getParameterIndex("network"), network, monitor);
            flag &= tester.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasTester = true;
            resBufferSize += tester.getNumberOfReturningTraces();
        }
        else {
            hasTester = false;
        }

        if (validator != null) {
            flag = true;
            flag &= validator.configure(validator.getParameterIndex("network"), network, monitor);
            flag &= validator.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasValidator = true;
            resBufferSize += validator.getNumberOfReturningTraces();
        }
        else {
            hasValidator = false;
        }

        datasetLength = dataset.getNRows();
        limit = epochs * datasetLength;

        res = new float[resBufferSize];
        results = new InterThreadBuffer(resBufferSize);

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }
    
    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int it;
        int posCopy;
        int expectedTestLength = 0;
        int expectedValidationLength = 0;
        float[] resultTest;
        float[] resultVal;
        boolean stop = false;
       
        this.initPlotter(limit, 1);
        networkTrainer.initPlotter(limit, 1);
        this.initSaver();
        networkTrainer.initSaver();

        posCopy = 0;
        if (hasTester) {
            posCopy = tester.getNumberOfReturningTraces();
            expectedTestLength = tester.getNumberOfReturningTraces();
        }
        if (hasValidator) {
            expectedValidationLength = validator.getNumberOfReturningTraces();
        }

        for (it = 0; (it < limit) && (monitor.isRunning()); it++) {
            while (monitor.isPaused());

            //------------------------------------------------------------------
            network.nextExampleRandom();

            this.singleStep(it + 1);

            if ((it % datasetLength) == 0) {
                ((AbstractFNNXGNGRBF)network).forget();
            }
            //------------------------------------------------------------------
            if ((it % logCounter) == 0) {
                if (this.testAndLog(it, expectedTestLength, expectedValidationLength)) {
                    it = limit;                                                    //stops the algorithm
                }
            }
        }

        this.endPlotter();
        this.endSaver();

        networkTrainer.endPlotter();
        networkTrainer.endSaver();

        return null;
    }

    public void evaluateModel(int it) {
        networkTrainer.evaluateModel(it);
        ((AbstractFNNXGNGRBF)network).calcOutput();
        ((AbstractFNNXGNGRBF)network).calcModelError();
    }

    public void updateModel(int it) {
        networkTrainer.updateModel(it);
        ((AbstractFNNXGNGRBF)network).updateSupervised(eta);
    }

    private AbstractFNNXTrainingAlgorithm networkTrainer;
    private float eta;

}
