/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package cellGNGRBF;

import cellGNG.AbstractGNGCell;
import cellGNG.Edge;
import data.MyLinkedList;
import java.util.Random;

/**
 *
 * @author hsatizab
 */
public class GNGRBFCell extends AbstractGNGCell {

    /** Creates a new instance of GNGLLMCell */
    public GNGRBFCell(int nI, int nO, Random r) {
        edges = new MyLinkedList();

        position = new float[nI];
        dist = new float[nI];

        output = new float[nO];
        cellError = new float[nO];
        weights = new float[nO];
        int i;
        for (i = 0; i < nO; i++) {
            weights[i] = r.nextFloat() - (float)0.5;
        }
        isGaussianSet = false;
    }

    /** Creates a new instance of GNGLLMCell */
    public GNGRBFCell(GNGRBFCell a, GNGRBFCell b) {
        int nI = a.position.length;
        int nO = a.weights.length;
//        if (nI != b.getPosition().length) {
//            throw new NullPointerException("Cells with different dimensions");
//        }

        edges = new MyLinkedList();

        position = new float[nI];
        dist = new float[nI];
        output = new float[nO];
        cellError = new float[nO];
        weights = new float[nO];

        int i;
        variance = 0;
        for (i = 0; i < nI; i++) {
            position[i] = (a.getPosition()[i] + b.getPosition()[i]) / 2;
            variance += (a.getPosition()[i] - b.getPosition()[i]) * (a.getPosition()[i] - b.getPosition()[i]);
        }
        variance /= 4;
        isGaussianSet = true;

        for (i = 0; i < nO; i++) {
            weights[i] = (a.weights[i] + b.weights[i])/2;
        }

    }

    public float[] calcOutput(float[] input) {
        int i;
        float[] temp;
        this.calcDistanceTo(input);
        
//        activation = (1 / stdDev) * (float)Math.exp(-distMag/(2*variance));
        activation = (float)Math.exp(-distMag/(2*variance));

//        System.err.print("O\t");
//        System.err.print(distMag);
//        System.err.print("\t");
//        System.err.print(variance);
//        System.err.print("\t");
//        System.err.print(activation);

        for (i = 0; i < output.length; i++) {
            output[i] = weights[i] * activation;
//            System.err.print("\t - ");
//            System.err.print(weights[i]);
//            System.err.print("\t");
//            System.err.print(output[i]);
        }
//        System.err.println();

        edges.head();
        while (edges.hasNode()) {
            temp = ((GNGRBFCell)getEdgeDoNext().getUnitConnectedTo(this)).calcOutputSingle(input);
            for (i = 0; i < output.length; i++) {
                output[i] += temp[i];
            }
        }

        return output;
    }

    private float[] calcOutputSingle(float[] input) {
        int i;
//        activation = (1 / stdDev) * (float)Math.exp(-distMag/(2*variance));
        activation = (float)Math.exp(-distMag/(2*variance));
        for (i = 0; i < output.length; i++) {
            output[i] = weights[i] * activation;
        }
        return output;
    }

    public float calcLocalError(float[] target) {
        int j;
        localError = 0;

        for (j = 0; j < target.length; j++) {
            cellError[j] = target[j] - output[j];
            localError += (cellError[j] * cellError[j]);
        }
        return localError;
    }

    public void updateSupervised(float eta, float[] input, float[] target) {
        int i;
        float actSum = activation + 1;
        float etaFrac = 0;

        calcOutput(input);
        calcLocalError(target);

        edges.head();
        while (edges.hasNode()) {
            actSum += ((GNGRBFCell)getEdgeDoNext().getUnitConnectedTo(this)).activation;
        }

//        System.err.print("W\t");
//        System.err.print(eta);
//        System.err.print("\t");
//        System.err.print(activation);
//        System.err.print("\t");
//        System.err.print(actSum);

        etaFrac = (eta * activation) / actSum;
//        System.err.print("\t");
//        System.err.print(etaFrac);

        for (i = 0; i < weights.length; i++) {
            weights[i] += etaFrac * cellError[i];
//            System.err.print("\t - ");
//            System.err.print(weights[i]);
        }
//        System.err.println();


        edges.head();
        while (edges.hasNode()) {
            ((GNGRBFCell)getEdgeDoNext().getUnitConnectedTo(this)).updateSupervised(eta, actSum, cellError);
        }

    }

    private void updateSupervised(float eta, float actSum, float[] cE) {
        float etaFrac = (eta * activation) / actSum;

//        System.err.print("\t");
//        System.err.print(activation);

        for (int i = 0; i < weights.length; i++) {
            weights[i] += etaFrac * cE[i];
        }
    }

    public void calcGaussian() {
        if (isGaussianSet) {
            return;
        }
        variance = 0;
        edges.head();
        while (edges.hasNode()) {
            variance += this.calcDistanceTo(((Edge)edges.getDoNext()).getUnitConnectedTo(this));
        }
        variance /= edges.size();
//        variance *= 2;

        isGaussianSet = true;
    }

    public void calcGaussianNeighbors() {
        calcGaussian();
        while (edges.hasNode()) {
            ((GNGRBFCell)((Edge)edges.getDoNext()).getUnitConnectedTo(this)).calcGaussian();
        }
    }

    public void connectTo(AbstractGNGCell neighbor) {
        super.connectTo(neighbor);

        this.isGaussianSet = false;
        this.calcGaussian();
        
        ((GNGRBFCell)neighbor).isGaussianSet = false;
        ((GNGRBFCell)neighbor).calcGaussian();
    }

    public void update(float e, float eb, float en, float wi, float wd, AbstractGNGCell u) {
        super.update(e, eb, en, wi, wd, u);

        calcGaussian();
        while (edges.hasNode()) {
            ((GNGRBFCell)((Edge)edges.getDoNext()).getUnitConnectedTo(this)).calcGaussianNeighbors();
        }
    }

    public void move(float e) {
        super.move(e);
        isGaussianSet = false;
    }

    public float getVariance() {
        return variance;
    }

    public float[] getWeights() {
        return weights;
    }

    private float activation;
    private float variance;
    private float[] weights;
    private float[] output;
    private float[] cellError;

    private boolean isGaussianSet;
}
