/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package cellGNG;

import data.MyLinkedList;
import model.localRepresentation.AbstractFNNXCell;

/**
 *
 * @author hsatizab
 */
public abstract class AbstractGNGCell extends AbstractFNNXCell {

    public void initCell(AbstractFNNXCell c1, AbstractFNNXCell c2) {
        int nI = c1.getPosition().length;
//        if (nI != c2.getPosition().length) {
//            throw new NullPointerException("Cells with different dimensions");
//        }

//        if (nI != this.getPosition().length) {
//            throw new NullPointerException("Cells with different dimensions");
//        }

        for (int i = 0; i < nI; i++) {
            position[i] = (c1.getPosition()[i] + c2.getPosition()[i]) / 2;
        }
        localError = 0;
        weight = 0;
    }

    public void incrementAge() {
        edges.head();
        while (edges.hasNode()) {
            ((Edge)edges.getDoNext()).incrementAge();
        }
    }

    public void update(float e, float eb, float en, float wi, float wd, AbstractGNGCell u) {
        Edge tempEdge;
        AbstractGNGCell tempUnit;
        boolean connectedToNearest = false;

        error += e;                                         //4. cumulating error

        this.move(eb * (float)Math.exp(-weight));           //5. moving nearest unit
        weight += wi;

        edges.head();
        while (edges.hasNode()) {
            tempEdge = (Edge) edges.getDoNext();
            
            tempUnit = tempEdge.getUnitConnectedTo(this);
            tempUnit.move(en * (float)Math.exp(-tempUnit.weight));    //5. moving neighbors

            tempUnit.subtractToWeight(wd);
            
            if (tempUnit == u) {                            //6. if it is connected to the nearest unit
                tempEdge.setAge(0);
                connectedToNearest = true;
            }
        }
        
        if (!connectedToNearest) {                          //6. if it is not connected to the nearest unit
//            tempEdge = new Edge(this, u);
//            edges.append(tempEdge);
//            u.edges.append(tempEdge);
            this.connectTo(u);
            u.subtractToWeight(wd);
        }
    }
    
    public boolean isConnectedTo(AbstractGNGCell u) {
        edges.head();
        while (edges.hasNode()) {
            if (((Edge)edges.getDoNext()).getUnitConnectedTo(this) == u) {
                return true;
            }
        }
        return false;
    }

    public AbstractGNGCell getNeighborWithLargestError() {
        float tempError = 0;
        float maxError = Float.NEGATIVE_INFINITY;
        AbstractGNGCell unitMaxError = null;
        AbstractGNGCell tempUnitMaxError = null;
        
        edges.head();
        while (edges.hasNode()) {
            tempUnitMaxError = ((Edge)edges.getDoNext()).getUnitConnectedTo(this);
            tempError = tempUnitMaxError.getError();
            if (tempError > maxError) {
                maxError = tempError;
                unitMaxError = tempUnitMaxError;
            }
        }
        return unitMaxError;
    }

    public void connectTo(AbstractGNGCell neighbor) {
        Edge tempEdge = new Edge(this, neighbor);
        this.edges.append(tempEdge);
        neighbor.edges.append(tempEdge);
    }

    public void deleteOlderConnections(int a) {
        Edge tempEdge;
        AbstractGNGCell tempCell;
        edges.head();
        while (edges.hasNode()) {
            tempEdge = (Edge)edges.getCurrent();
            if (tempEdge.getAge() > a) {
                tempCell = tempEdge.getUnitConnectedTo(this);
                tempCell.deleteConnectionTo(this);
                edges.deleteCurrentDoNext();
            }
            else {
                edges.next();
            }
        }
    }

    public void deleteConnectionTo(AbstractGNGCell ng) {
        edges.head();
        while (edges.hasNode()) {
            if (((Edge)edges.getCurrent()).getUnitConnectedTo(this) == ng) {
                edges.deleteCurrentDoNext();
                return;
            }
            else {
                edges.next();
            }
        }
    }

    public int getAmountOfEdges() {
        return edges.size();
    }

    public void firstEdge() {
        edges.head();
    }

    public boolean hasEdge() {
        return edges.hasNode();
    }

    public Edge getEdgeDoNext() {
        return (Edge)edges.getDoNext();
    }

    public void subtractToWeight(float d) {
        weight -= d;
        if (weight < 0) {
            weight = 0;
        }
    }

    public void anchorCell(float wf) {
        weight *= wf;
    }
    
    public void setWeight(float w) {
        weight = w;
    }

    public float getWeight() {
        return weight;
    }

    public float getLocalError() {
        return localError;
    }

    protected MyLinkedList edges;
    protected float localError;
    protected float weight;

}
