/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXCascadeCorrelation;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import model.distributedRepresentation.AbstractActivationFunction;
import functions.SigmoidalActivationFunction;
import functions.HyperbolicTangentActivationFunction;
import functions.LinearActivationFunction;
import networks.AbstractFNNXFeedForwardNetwork;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import toolFNNXMLP.FNNXMLP;
import visualFNNXCascadeCorrelation.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXCascadeCorrelation extends AbstractFNNXTrainingAlgorithm implements FNNXObject, FNNXFunction, FNNXTrainingAlgorithm {

    public FNNXCascadeCorrelation(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[12];
        parameterNames[0] = "network";
        parameterNames[1] = "neurons";
        parameterNames[2] = "dataset";
        parameterNames[3] = "actFuncOutput";
        parameterNames[4] = "inputTrainer";
        parameterNames[5] = "outputTrainer";
        parameterNames[6] = "tester";
        parameterNames[7] = "validator";
        parameterNames[8] = "stopIfTesterBelow";
        parameterNames[9] = "plotter";
        parameterNames[10] = "saver";
        parameterNames[11] = "delay";

        parameterTypes = new Class[12];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = Number.class;
        parameterTypes[2] = DataBlock.class;
        parameterTypes[3] = String.class;
        parameterTypes[4] = AbstractFNNXFunction.class;
        parameterTypes[5] = AbstractFNNXFunction.class;
        parameterTypes[6] = AbstractFNNXFunction.class;
        parameterTypes[7] = AbstractFNNXFunction.class;
        parameterTypes[8] = Number.class;
        parameterTypes[9] = AbstractFNNXParallelObserver.class;
        parameterTypes[10] = AbstractFNNXParallelObserver.class;
        parameterTypes[11] = Number.class;

        isParameterMandatory = new boolean[12];
        isParameterMandatory[0] = false;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = true;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = true;
        isParameterMandatory[6] = false;
        isParameterMandatory[7] = false;
        isParameterMandatory[8] = false;
        isParameterMandatory[9] = false;
        isParameterMandatory[10] = false;
        isParameterMandatory[11] = false;

        parameters = new Object[12];

        //default values
        delay = 0;
        stopIfTesterBelow = 0;
    }

    public FNNXCascadeCorrelation(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[12];
        parameterNames[0] = "network";
        parameterNames[1] = "neurons";
        parameterNames[2] = "dataset";
        parameterNames[3] = "actFuncOutput";
        parameterNames[4] = "inputTrainer";
        parameterNames[5] = "outputTrainer";
        parameterNames[6] = "tester";
        parameterNames[7] = "validator";
        parameterNames[8] = "stopIfTesterBelow";
        parameterNames[9] = "plotter";
        parameterNames[10] = "saver";
        parameterNames[11] = "delay";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        boolean flag;
        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXFeedForwardNetwork)parameters[i];
                        break;
                    case 1:
                        neuronsMax = ((Number)parameters[i]).intValue();
                        if (!testMin(i, neuronsMax, 0)) {
                            return false;
                        }
                        break;
                    case 2:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 3:
                        outputActivationFunction = (String)parameters[i];
                        break;
                    case 4:
                        inputTrainer = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 5:
                        outputTrainer = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 6:
                        tester = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 7:
                        validator = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 8:
                        stopIfTesterBelow = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, stopIfTesterBelow, 0)) {
                            return false;
                        }
                        break;
                    case 9:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 10:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 11:
                        delay = ((Number)parameters[i]).intValue();
                        if (!testMin(i, delay, 0)) {
                            return false;
                        }
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        if (dataset.getNParts() != 2) {
            monitor.appendErrorMessage("Error. Wrong number of dataset parts in "+ name + ". Inputs and outputs are required");
            return false;
        }

        int nInputs = dataset.getDataset(0).getNColumns();
        int nOutputs = dataset.getDataset(1).getNColumns();

        if (network == null) {

            AbstractActivationFunction oaf = null;
            switch (Names.activationFunction.valueOf(outputActivationFunction)) {
                case LINEAR:
                    oaf = new LinearActivationFunction();
                    break;
                case SIGM:
                    oaf = new SigmoidalActivationFunction();
                    break;
                case TANH:
                    oaf = new HyperbolicTangentActivationFunction();
                    break;
                default:
                    monitor.appendErrorMessage("Error. Illegal activation function.");
                    return false;
            }

            network = new FNNXMLP(nInputs, nOutputs, oaf);
            ((AbstractFNNXFeedForwardNetwork)network).connectFeedForward();
            network.initModel();
        }
        else {
            if (nInputs != ((AbstractFNNXFeedForwardNetwork)network).getAmountOfInputs()) {
                monitor.appendErrorMessage("Error. Wrong number of inputs.");
                return false;
            }
            if (nOutputs != ((AbstractFNNXFeedForwardNetwork)network).getAmountOfOutputs()) {
                monitor.appendErrorMessage("Error. Wrong number of outputs.");
                return false;
            }
        }

        int resBufferSize = 0;
        if (tester != null) {
            flag = true;
            flag &= tester.configure(tester.getParameterIndex("dataset"), dataset, monitor);
            flag &= tester.configure(tester.getParameterIndex("network"), network, monitor);
            flag &= tester.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasTester = true;
            resBufferSize += tester.getNumberOfReturningTraces();
        }
        else {
            hasTester = false;
        }

        if (validator != null) {
            flag = true;
            flag &= validator.configure(validator.getParameterIndex("network"), network, monitor);
            flag &= validator.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasValidator = true;
            resBufferSize += validator.getNumberOfReturningTraces();
        }
        else {
            hasValidator = false;
        }

        flag = true;
        flag &= inputTrainer.configure(inputTrainer.getParameterIndex("network"), network, monitor);
        flag &= inputTrainer.configure(inputTrainer.getParameterIndex("dataset"), dataset, monitor);
        flag &= inputTrainer.prepare(monitor);
        if (!flag) {
            monitor.appendErrorMessage("Error. Impossible to configure input trainer.");
            return false;
        }

        flag = true;
        flag &= outputTrainer.configure(outputTrainer.getParameterIndex("dataset"), dataset, monitor);
        flag &= outputTrainer.configure(outputTrainer.getParameterIndex("network"), network, monitor);
        flag &= outputTrainer.prepare(monitor);
        if (!flag) {
            monitor.appendErrorMessage("Error. Impossible to configure output trainer.");
            return false;
        }

        res = new float[resBufferSize];
        results = new InterThreadBuffer(resBufferSize);
        
        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }
    
    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int expectedTestLength = 0;
        int expectedValidationLength = 0;
        int neuronCounter = ((AbstractFNNXFeedForwardNetwork)network).getAmountOfLayers()-3;

        this.initPlotter(neuronsMax, 1);
        this.initSaver();
        
        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
            return null;
        }

        if (hasTester) {
            expectedTestLength = tester.getNumberOfReturningTraces();
        }
        if (hasValidator) {
            expectedValidationLength = validator.getNumberOfReturningTraces();
        }

        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        while ((neuronCounter < neuronsMax) && monitor.isRunning()) {
            while (monitor.isPaused()) {
            }
            
            //------------------------------------------------------------------
            //                  Train output connections
            //------------------------------------------------------------------
            outputTrainer.eval(monitor);
            if (monitor.hasErrorMessage()) {
                return null;
            }

            neuronCounter++;

            if (this.testAndLog(limit, expectedTestLength, expectedValidationLength)) {
                neuronCounter = neuronsMax;
            }

//            //------------------------------------------------------------------
//            if (hasTester) {
//                resultTest = (float[])tester.eval(monitor);
//                if (resultTest == null) {
//                    if (expectedTestLength != 0) {
//                        monitor.appendErrorMessage("Error. " + tester.toString() + " failed.");
//                        return null;
//                    }
//                }
//                else {
//                    System.arraycopy(resultTest, 0, res, 0, resultTest.length);
//                    if (resultTest[0] < stopIfTesterBelow) {
//                        stop = true;                                            //stops the algorithm
//                    }
//                }
//            }
//            if (hasValidator) {
//                resultVal = (float[])validator.eval(monitor);
//                if (resultVal == null) {
//                    if (expectedValidationLength != 0) {
//                        monitor.appendErrorMessage("Error. " + validator.toString() + "failed.");
//                        return null;
//                    }
//                }
//                else {
//                    System.arraycopy(resultVal, 0, res, posCopy, resultVal.length);
//                }
//                                                //this is necesary because the validator attachs the validation dataset to the network
//                if (!network.setDataset(dataset)) {
//                    monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
//                    return null;
//                }
//
//            }
//
//            results.newData(res, neuronCounter+1);
//            //----------------------------------------------
//            if (hasPlotter && (hasTester || hasValidator)) {
//                plotter.getMonitor().newResults(results);
//            }
//            if (hasSaver && (hasTester || hasValidator)) {
//                saver.getMonitor().newResults(results);
//            }
//            //----------------------------------------------

//            if (stop) {
//                neuronCounter = neuronsMax;                                     //stops the algorithm
//            }

            if (neuronCounter < neuronsMax) {
                //-------------------------------------------------------------------------
                //  Train input connections of hidden candidates and insert the best one
                //-------------------------------------------------------------------------
                inputTrainer.eval(monitor);
                if (monitor.hasErrorMessage()) {
                    return null;
                }
            }

            if (delay > 0) {
                try {
                    Thread.sleep(delay);
                }
                catch (InterruptedException ex) {
                }
            }

        }
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        this.endPlotter();
        this.endSaver();

        return network;
    }

    public void evaluateModel(int it) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void updateModel(int it) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    private int neuronsMax;
    private String outputActivationFunction;
    private AbstractFNNXFunction inputTrainer;
    private AbstractFNNXFunction outputTrainer;
    private int delay;

}
