/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXCrazyTeacher;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import model.AbstractFNNXModel;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import visualFNNXCrazyTeacher.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXCrazyTeacher extends AbstractFNNXTrainingAlgorithm implements FNNXObject, FNNXFunction, FNNXTrainingAlgorithm {

    public FNNXCrazyTeacher(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[9];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "trainer";
        parameterNames[3] = "epochs";
        parameterNames[4] = "tester";
        parameterNames[5] = "validator";
        parameterNames[6] = "logCounter";
        parameterNames[7] = "plotter";
        parameterNames[8] = "saver";

        parameterTypes = new Class[9];
        parameterTypes[0] = AbstractFNNXModel.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXTrainingAlgorithm.class;
        parameterTypes[3] = Number.class;
        parameterTypes[4] = AbstractFNNXFunction.class;
        parameterTypes[5] = AbstractFNNXFunction.class;
        parameterTypes[6] = Number.class;
        parameterTypes[7] = AbstractFNNXParallelObserver.class;
        parameterTypes[8] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[9];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = true;
        isParameterMandatory[4] = false;
        isParameterMandatory[5] = false;
        isParameterMandatory[6] = false;
        isParameterMandatory[7] = false;
        isParameterMandatory[8] = false;

        parameters = new Object[9];

        //default values
        stopIfTesterBelow = 0;
        logCounter = 100;

    }

    public FNNXCrazyTeacher(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[9];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "trainer";
        parameterNames[3] = "epochs";
        parameterNames[4] = "tester";
        parameterNames[5] = "validator";
        parameterNames[6] = "logCounter";
        parameterNames[7] = "plotter";
        parameterNames[8] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        boolean flag;
        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXModel)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        trainer = (AbstractFNNXTrainingAlgorithm)parameters[i];
                        break;
                    case 3:
                        epochs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, epochs, 0)) {
                            return false;
                        }
                        break;
                    case 4:
                        tester = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 5:
                        validator = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 6:
                        logCounter = ((Number)parameters[i]).intValue();
                        if (!testMin(i, logCounter, 0)) {
                            return false;
                        }
                        break;
                    case 7:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 8:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Impossible to set dataset in " + this.name);
            return false;
        }
        
        flag = true;
        flag &= trainer.configure(trainer.getParameterIndex("dataset"), dataset, monitor);
        flag &= trainer.configure(trainer.getParameterIndex("network"), network, monitor);
        flag &= trainer.configure(trainer.getParameterIndex("epochs"), epochs, monitor);
        flag &= trainer.prepare(monitor);
        if (!flag) {
            monitor.appendErrorMessage("Error. Impossible to configure trainer in " + this.name);
            return false;
        }

        int resBufferSize = 0;
        if (tester != null) {
            flag = true;
            flag &= tester.configure(tester.getParameterIndex("dataset"), dataset, monitor);
            flag &= tester.configure(tester.getParameterIndex("network"), network, monitor);
            flag &= tester.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure tester.");
                return false;
            }

            hasTester = true;
            resBufferSize += tester.getNumberOfReturningTraces();
        }
        else {
            hasTester = false;
        }

        if (validator != null) {
            flag = true;
            flag &= validator.configure(validator.getParameterIndex("network"), network, monitor);
            flag &= validator.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure validator.");
                return false;
            }

            hasValidator = true;
            resBufferSize += validator.getNumberOfReturningTraces();
        }
        else {
            hasValidator = false;
        }

        datasetLength = dataset.getNRows();
        limit = epochs * datasetLength;

        res = new float[resBufferSize];
        results = new InterThreadBuffer(resBufferSize);

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }
    
    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int it;
        int expectedTestLength = 0;
        int expectedValidationLength = 0;
       
        this.initPlotter(limit, 1);
        this.initSaver();

        if (hasTester) {
            expectedTestLength = tester.getNumberOfReturningTraces();
            tester.initSaver();
        }
        if (hasValidator) {
            expectedValidationLength = validator.getNumberOfReturningTraces();
            tester.initSaver();
        }

        for (it = 0; (it < limit) && (monitor.isRunning()); it++) {
            while (monitor.isPaused());

            //------------------------------------------------------------------
            network.nextExampleRandom();
            
            trainer.singleStep(it + 1);

            //------------------------------------------------------------------
            if ((it % logCounter) == 0) {

                if (this.testAndLog(it, expectedTestLength, expectedValidationLength)) {
                    it = limit;
                }

//                if (hasTester) {
//                    resultTest = (float[])tester.eval(monitor);
//                    if (resultTest == null) {
//                        if (expectedTestLength != 0) {
//                            monitor.appendErrorMessage("Error. " + tester.toString() + " failed.");
//                            return null;
//                        }
//                    }
//                    else {
//                        System.arraycopy(resultTest, 0, res, 0, resultTest.length);
//                        if (resultTest[0] < stopIfTesterBelow) {
//                            stop = true;                                            //stops the algorithm
//                        }
//                    }
//                }
//
//                if (hasValidator) {
//                    resultVal = (float[])validator.eval(monitor);
//                    if (resultVal == null) {
//                        if (expectedValidationLength != 0) {
//                            monitor.appendErrorMessage("Error. " + validator.toString() + "failed.");
//                            return null;
//                        }
//                    }
//                    else {
//                        System.arraycopy(resultVal, 0, res, posCopy, resultVal.length);
//                    }
//                                            //this is necesary because the validator attachs the validation dataset to the network
//                    if (!network.setDataset(dataset)) {
//                        monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
//                        return null;
//                    }
//                }
//
//                results.newData(res, it+1);
//                //----------------------------------------------
//                if (hasSaver && (hasTester || hasValidator)) {
//                    saver.getMonitor().newResults(results);
//                }
//                //----------------------------------------------
//
//                if (stop) {
//                    it = limit;                                                    //stops the algorithm
//                }
            }
        }

        this.endPlotter();
        this.endSaver();

        if (hasTester) {
            tester.endSaver();
        }
        if (hasValidator) {
            validator.endSaver();
        }

        return null;
    }

    private AbstractFNNXTrainingAlgorithm trainer;

    public void evaluateModel(int it) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void updateModel(int it) {
        throw new UnsupportedOperationException("Not supported yet.");
    }
}
