/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXDynamicNodeCreation;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import java.util.Random;
import javax.swing.JFrame;
import model.distributedRepresentation.AbstractActivationFunction;
import functions.SigmoidalActivationFunction;
import functions.HyperbolicTangentActivationFunction;
import functions.LinearActivationFunction;
import layers.HiddenLayer;
import networks.AbstractFNNXFeedForwardNetwork;
import neurons.HiddenUnit;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import toolFNNXMLP.FNNXMLP;
import visualFNNXDynamicNodeCreation.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXDynamicNodeCreation extends AbstractFNNXTrainingAlgorithm implements FNNXObject, FNNXFunction, FNNXTrainingAlgorithm {

    public FNNXDynamicNodeCreation(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[14];
        parameterNames[0] = "network";
        parameterNames[1] = "neurons";
        parameterNames[2] = "dataset";
        parameterNames[3] = "actFuncOutput";
        parameterNames[4] = "shortcut";
        parameterNames[5] = "insertEvery";
        parameterNames[6] = "trainer";
        parameterNames[7] = "epochs";
        parameterNames[8] = "tester";
        parameterNames[9] = "validator";
        parameterNames[10] = "stopIfTesterBelow";
        parameterNames[11] = "logCounter";
        parameterNames[12] = "plotter";
        parameterNames[13] = "saver";

        parameterTypes = new Class[14];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = Number.class;
        parameterTypes[2] = DataBlock.class;
        parameterTypes[3] = String.class;
        parameterTypes[4] = Boolean.class;
        parameterTypes[5] = Number.class;
        parameterTypes[6] = AbstractFNNXFunction.class;
        parameterTypes[7] = Number.class;
        parameterTypes[8] = AbstractFNNXFunction.class;
        parameterTypes[9] = AbstractFNNXFunction.class;
        parameterTypes[10] = Number.class;
        parameterTypes[11] = Number.class;
        parameterTypes[12] = AbstractFNNXParallelObserver.class;
        parameterTypes[13] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[14];
        isParameterMandatory[0] = false;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = true;
        isParameterMandatory[4] = false;
        isParameterMandatory[5] = true;
        isParameterMandatory[6] = true;
        isParameterMandatory[7] = true;
        isParameterMandatory[8] = false;
        isParameterMandatory[9] = false;
        isParameterMandatory[10] = false;
        isParameterMandatory[11] = false;
        isParameterMandatory[12] = false;
        isParameterMandatory[13] = false;

        parameters = new Object[14];

        //default values
        stopIfTesterBelow = 0;
        logCounter = 100;
        shortcut = false;
    }

    public FNNXDynamicNodeCreation(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[14];
        parameterNames[0] = "network";
        parameterNames[1] = "neurons";
        parameterNames[2] = "dataset";
        parameterNames[3] = "actFuncOutput";
        parameterNames[4] = "shortcut";
        parameterNames[5] = "insertEvery";
        parameterNames[6] = "trainer";
        parameterNames[7] = "epochs";
        parameterNames[8] = "tester";
        parameterNames[9] = "validator";
        parameterNames[10] = "stopIfTesterBelow";
        parameterNames[11] = "logCounter";
        parameterNames[12] = "plotter";
        parameterNames[13] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        boolean flag;
        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (FNNXMLP)parameters[i];
                        break;
                    case 1:
                        neuronsMax = ((Number)parameters[i]).intValue();
                        if (!testMin(i, neuronsMax, 0)) {
                            return false;
                        }
                        break;
                    case 2:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 3:
                        outputActivationFunction = (String)parameters[i];
                        break;
                    case 4:
                        shortcut = ((Boolean)parameters[i]).booleanValue();
                        break;
                    case 5:
                        insertEvery = ((Number)parameters[i]).intValue();
                        if (!testMin(i, insertEvery, 0)) {
                            return false;
                        }
                        break;
                    case 6:
                        trainer = (AbstractFNNXTrainingAlgorithm)parameters[i];
                        break;
                    case 7:
                        epochs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, epochs, 0)) {
                            return false;
                        }
                        break;
                    case 8:
                        tester = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 9:
                        validator = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 10:
                        stopIfTesterBelow = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, stopIfTesterBelow, 0)) {
                            return false;
                        }
                        break;
                    case 11:
                        logCounter = ((Number)parameters[i]).intValue();
                        if (!testMin(i, logCounter, 0)) {
                            return false;
                        }
                        break;
                    case 12:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 13:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        if (dataset.getNParts() != 2) {
            monitor.appendErrorMessage("Error. Wrong number of dataset parts in "+ name + ". Inputs and outputs are required");
            return false;
        }

        int nInputs = dataset.getDataset(0).getNColumns();
        int nOutputs = dataset.getDataset(1).getNColumns();

        if (network == null) {
            AbstractActivationFunction oaf = null;
            switch (Names.activationFunction.valueOf(outputActivationFunction)) {
                case LINEAR:
                    oaf = new LinearActivationFunction();
                    break;
                case SIGM:
                    oaf = new SigmoidalActivationFunction();
                    break;
                case TANH:
                    oaf = new HyperbolicTangentActivationFunction();
                    break;
                default:
                    monitor.appendErrorMessage("Error. Illegal activation function.");
                    return false;
            }

            network = new FNNXMLP(nInputs, nOutputs, oaf);
            ((AbstractFNNXFeedForwardNetwork)network).insertHiddenLayer(0, new HyperbolicTangentActivationFunction());
            if (shortcut) {
                ((AbstractFNNXFeedForwardNetwork)network).connectShortcut();
            }
            else {
                ((AbstractFNNXFeedForwardNetwork)network).connectFeedForward();
            }
            network.initModel();
        }
        else {
            if (nInputs != ((AbstractFNNXFeedForwardNetwork)network).getAmountOfInputs()) {
                monitor.appendErrorMessage("Error. The number of inputs of dataset and network must be the same");
                return false;
            }
            if (nOutputs != ((AbstractFNNXFeedForwardNetwork)network).getAmountOfOutputs()) {
                monitor.appendErrorMessage("Error. The number of outputs of dataset and network must be the same");
                return false;
            }
        }

        ((AbstractFNNXFeedForwardNetwork)network).firstLayer();
        ((AbstractFNNXFeedForwardNetwork)network).getLayerDoNext();
        neuronCounter = ((AbstractFNNXFeedForwardNetwork)network).getLayerDoNext().getAmountOfUnits() - 1;

        flag = true;
        flag &= trainer.configure(trainer.getParameterIndex("dataset"), dataset, monitor);
        flag &= trainer.configure(trainer.getParameterIndex("network"), network, monitor);
        flag &= trainer.configure(trainer.getParameterIndex("epochs"), epochs, monitor);
        flag &= trainer.prepare(monitor);
        if (!flag) {
            monitor.appendErrorMessage("Error. Impossible to configure trainer.");
            return false;
        }

        int resBufferSize = 0;
        if (tester != null) {
            flag = true;
            flag &= tester.configure(tester.getParameterIndex("dataset"), dataset, monitor);
            flag &= tester.configure(tester.getParameterIndex("network"), network, monitor);
            flag &= tester.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasTester = true;
            resBufferSize += tester.getNumberOfReturningTraces();
        }
        else {
            hasTester = false;
        }

        if (validator != null) {
            flag = true;
            flag &= validator.configure(validator.getParameterIndex("network"), network, monitor);
            flag &= validator.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasValidator = true;
            resBufferSize += validator.getNumberOfReturningTraces();
        }
        else {
            hasValidator = false;
        }
        
        datasetLength = dataset.getNRows();
        limit = epochs * datasetLength;
        gen = new Random();
        insertEvery *= datasetLength;

        res = new float[resBufferSize];
        results = new InterThreadBuffer(resBufferSize);

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

//        boolean flag;
//        HiddenUnit tempHiddenUnit;
//        HiddenLayer tempHiddenLayer;
//        float[] resultTest;
        int expectedTestLength = 0;
//        float[] resultVal;
        int expectedValidationLength = 0;
//        int posCopy;
//        boolean stop = false;

        this.initPlotter(limit, 1);
        this.initSaver();

        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
            return null;
        }

//        posCopy = 0;
        if (hasTester) {
//            posCopy = tester.getNumberOfReturningTraces();
            expectedTestLength = tester.getNumberOfReturningTraces();
        }
        if (hasValidator) {
            expectedValidationLength = validator.getNumberOfReturningTraces();
        }

        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        for (it = 0; ((it < limit) && monitor.isRunning()); it++) {
            while (monitor.isPaused()) {
            }
            network.nextExampleRandom();

            this.singleStep(it + 1);

            if ((it % logCounter) == 0) {

                if (this.testAndLog(it, expectedTestLength, expectedValidationLength)) {
                    it = limit;
                    neuronCounter = neuronsMax + 1;
                }
                
//            //------------------------------------------------------------------
//                if (hasTester) {
//                    resultTest = (float[])tester.eval(monitor);
//                    if (resultTest == null) {
//                        if (expectedTestLength != 0) {
//                            monitor.appendErrorMessage("Error. " + tester.toString() + " failed.");
//                            return null;
//                        }
//                    }
//                    else {
//                        System.arraycopy(resultTest, 0, res, 0, resultTest.length);
//                        if (resultTest[0] < stopIfTesterBelow) {
//                            stop = true;                                            //stops the algorithm
//                        }
//                    }
//                }
//
//                if (hasValidator) {
//                    resultVal = (float[])validator.eval(monitor);
//                    if (resultVal == null) {
//                        if (expectedValidationLength != 0) {
//                            monitor.appendErrorMessage("Error. " + validator.toString() + "failed.");
//                            return null;
//                        }
//                    }
//                    else {
//                        System.arraycopy(resultVal, 0, res, posCopy, resultVal.length);
//                    }
//                         //this is necesary because the validator attachs the validation dataset to the network
//                    if (!network.setDataset(dataset)) {
//                        monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
//                        return null;
//                    }
//                }
//
//                results.newData(res, neuronCounter+1);
//                //----------------------------------------------
//                if (hasPlotter && (hasTester || hasValidator)) {
//                    plotter.getMonitor().newResults(results);
//                }
//                if (hasSaver && (hasTester || hasValidator)) {
//                    saver.getMonitor().newResults(results);
//                }
//                //----------------------------------------------
//
//                if (stop) {
//                    neuronCounter = neuronsMax + 1;                                     //stops the algorithm
//                    it = limit;
//                }

            }

        }
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//        while ((neuronCounter < neuronsMax) && monitor.isRunning()) {
//            while (monitor.isPaused()) {
//            }

//            //------------------------------------------------------------------
//            if ((neuronCounter < neuronsMax) && (neuronCounter >= 0)) {
//                //--------------------------------------------------------------
//                // Insert unit in the hidden layer
//                tempHiddenUnit = new HiddenUnit(new HyperbolicTangentActivationFunction());
//
//                network.firstLayer();
//                network.getLayerDoNext();
//                tempHiddenLayer = (HiddenLayer)network.getLayerDoNext();
//                tempHiddenLayer.firstUnit();
//                tempHiddenLayer.insert(tempHiddenUnit);
//
//                tempHiddenUnit.connectBackward(network.getFirstLayer());
//                tempHiddenUnit.connectForward(network.getLastLayer());
//                tempHiddenUnit.initNeuron(gen);
//                System.err.print("Inserting... ");
//            }
//            System.err.println(neuronCounter);
//            neuronCounter++;
//
//            //------------------------------------------------------------------
//            // Train connections
//            trainer.eval(monitor);
//            if (monitor.hasErrorMessage()) {
//                return null;
//            }
//
//
//        }
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        this.endPlotter();
        this.endSaver();

        return network;
    }

    public void evaluateModel(int it) {
        trainer.evaluateModel(it);
    }

    public void updateModel(int it) {

        if ((it % insertEvery == 0) && (neuronCounter < neuronsMax)) {
            HiddenUnit tempHiddenUnit;
            HiddenLayer tempHiddenLayer;
            
            // Insert unit in the hidden layer
            tempHiddenUnit = new HiddenUnit(new HyperbolicTangentActivationFunction());

            ((AbstractFNNXFeedForwardNetwork)network).firstLayer();
            ((AbstractFNNXFeedForwardNetwork)network).getLayerDoNext();
            tempHiddenLayer = (HiddenLayer)((AbstractFNNXFeedForwardNetwork)network).getLayerDoNext();
            tempHiddenLayer.firstUnit();
            tempHiddenLayer.insert(tempHiddenUnit);

            tempHiddenUnit.connectBackward(((AbstractFNNXFeedForwardNetwork)network).getFirstLayer());
            tempHiddenUnit.connectForward(((AbstractFNNXFeedForwardNetwork)network).getLastLayer());
            tempHiddenUnit.initNeuron(gen);

//            System.err.println("inserting...");

            neuronCounter++;
        }
        trainer.updateModel(it);
    }

    private int neuronsMax;
    private String outputActivationFunction;
    private AbstractFNNXTrainingAlgorithm trainer;
    private int insertEvery;
    private boolean shortcut;

    private int it;
    private int neuronCounter;
    private Random gen;
}
