/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXDynamicNodeCreationCascade;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import java.util.Random;
import javax.swing.JFrame;
import model.distributedRepresentation.AbstractActivationFunction;
import functions.SigmoidalActivationFunction;
import functions.HyperbolicTangentActivationFunction;
import functions.LinearActivationFunction;
import layers.HiddenLayer;
import networks.AbstractFNNXFeedForwardNetwork;
import neurons.HiddenUnit;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import toolFNNXMLP.FNNXMLP;
import visualFNNXDynamicNodeCreationCascade.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXDynamicNodeCreationCascade extends AbstractFNNXTrainingAlgorithm implements FNNXObject, FNNXFunction, FNNXTrainingAlgorithm {

    public FNNXDynamicNodeCreationCascade(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[13];
        parameterNames[0] = "network";
        parameterNames[1] = "layers";
        parameterNames[2] = "dataset";
        parameterNames[3] = "actFuncOutput";
        parameterNames[4] = "insertEvery";
        parameterNames[5] = "trainer";
        parameterNames[6] = "epochs";
        parameterNames[7] = "tester";
        parameterNames[8] = "validator";
        parameterNames[9] = "stopIfTesterBelow";
        parameterNames[10] = "logCounter";
        parameterNames[11] = "plotter";
        parameterNames[12] = "saver";

        parameterTypes = new Class[13];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = Number.class;
        parameterTypes[2] = DataBlock.class;
        parameterTypes[3] = String.class;
        parameterTypes[4] = Number.class;
        parameterTypes[5] = AbstractFNNXFunction.class;
        parameterTypes[6] = Number.class;
        parameterTypes[7] = AbstractFNNXFunction.class;
        parameterTypes[8] = AbstractFNNXFunction.class;
        parameterTypes[9] = Number.class;
        parameterTypes[10] = Number.class;
        parameterTypes[11] = AbstractFNNXParallelObserver.class;
        parameterTypes[12] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[13];
        isParameterMandatory[0] = false;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = true;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = true;
        isParameterMandatory[6] = true;
        isParameterMandatory[7] = false;
        isParameterMandatory[8] = false;
        isParameterMandatory[9] = false;
        isParameterMandatory[10] = false;
        isParameterMandatory[11] = false;
        isParameterMandatory[12] = false;

        parameters = new Object[13];

        //default values
        stopIfTesterBelow = 0;
        logCounter = 100;
    }

    public FNNXDynamicNodeCreationCascade(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[13];
        parameterNames[0] = "network";
        parameterNames[1] = "layers";
        parameterNames[2] = "dataset";
        parameterNames[3] = "actFuncOutput";
        parameterNames[4] = "insertEvery";
        parameterNames[5] = "trainer";
        parameterNames[6] = "epochs";
        parameterNames[7] = "tester";
        parameterNames[8] = "validator";
        parameterNames[9] = "stopIfTesterBelow";
        parameterNames[10] = "logCounter";
        parameterNames[11] = "plotter";
        parameterNames[12] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        boolean flag;
        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (FNNXMLP)parameters[i];
                        break;
                    case 1:
                        layersMax = ((Number)parameters[i]).intValue();
                        if (!testMin(i, layersMax, 0)) {
                            return false;
                        }
                        break;
                    case 2:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 3:
                        outputActivationFunction = (String)parameters[i];
                        break;
                    case 4:
                        insertEvery = ((Number)parameters[i]).intValue();
                        if (!testMin(i, insertEvery, 0)) {
                            return false;
                        }
                        break;
                    case 5:
                        trainer = (AbstractFNNXTrainingAlgorithm)parameters[i];
                        break;
                    case 6:
                        epochs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, epochs, 0)) {
                            return false;
                        }
                        break;
                    case 7:
                        tester = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 8:
                        validator = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 9:
                        stopIfTesterBelow = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, stopIfTesterBelow, 0)) {
                            return false;
                        }
                        break;
                    case 10:
                        logCounter = ((Number)parameters[i]).intValue();
                        if (!testMin(i, logCounter, 0)) {
                            return false;
                        }
                        break;
                    case 11:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 12:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        if (dataset.getNParts() != 2) {
            monitor.appendErrorMessage("Error. Wrong number of dataset parts in "+ name + ". Inputs and outputs are required");
            return false;
        }

        int nInputs = dataset.getDataset(0).getNColumns();
        int nOutputs = dataset.getDataset(1).getNColumns();

        if (network == null) {
            AbstractActivationFunction oaf = null;
            switch (Names.activationFunction.valueOf(outputActivationFunction)) {
                case LINEAR:
                    oaf = new LinearActivationFunction();
                    break;
                case SIGM:
                    oaf = new SigmoidalActivationFunction();
                    break;
                case TANH:
                    oaf = new HyperbolicTangentActivationFunction();
                    break;
                default:
                    monitor.appendErrorMessage("Error. Illegal activation function.");
                    return false;
            }

            network = new FNNXMLP(nInputs, nOutputs, oaf);
            ((AbstractFNNXFeedForwardNetwork)network).connectFeedForward();
            network.initModel();
        }
        else {
            if (nInputs != ((AbstractFNNXFeedForwardNetwork)network).getAmountOfInputs()) {
                monitor.appendErrorMessage("Error. The number of inputs of dataset and network must be the same");
                return false;
            }
            if (nOutputs != ((AbstractFNNXFeedForwardNetwork)network).getAmountOfOutputs()) {
                monitor.appendErrorMessage("Error. The number of outputs of dataset and network must be the same");
                return false;
            }
        }

        layerCounter = ((AbstractFNNXFeedForwardNetwork)network).getAmountOfLayers() - 2;

        flag = true;
        flag &= trainer.configure(trainer.getParameterIndex("dataset"), dataset, monitor);
        flag &= trainer.configure(trainer.getParameterIndex("network"), network, monitor);
        flag &= trainer.configure(trainer.getParameterIndex("epochs"), epochs, monitor);
        flag &= trainer.prepare(monitor);
        if (!flag) {
            monitor.appendErrorMessage("Error. Impossible to configure trainer.");
            return false;
        }

        int resBufferSize = 0;
        if (tester != null) {
            flag = true;
            flag &= tester.configure(tester.getParameterIndex("dataset"), dataset, monitor);
            flag &= tester.configure(tester.getParameterIndex("network"), network, monitor);
            flag &= tester.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasTester = true;
            resBufferSize += tester.getNumberOfReturningTraces();
        }
        else {
            hasTester = false;
        }

        if (validator != null) {
            flag = true;
            flag &= validator.configure(validator.getParameterIndex("network"), network, monitor);
            flag &= validator.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasValidator = true;
            resBufferSize += validator.getNumberOfReturningTraces();
        }
        else {
            hasValidator = false;
        }
        
        datasetLength = dataset.getNRows();
        limit = epochs * datasetLength;
        gen = new Random();
        insertEvery *= datasetLength;

        res = new float[resBufferSize];
        results = new InterThreadBuffer(resBufferSize);

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int expectedTestLength = 0;
        int expectedValidationLength = 0;

        this.initPlotter(limit, 1);
        this.initSaver();

        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
            return null;
        }

        if (hasTester) {
            expectedTestLength = tester.getNumberOfReturningTraces();
        }
        if (hasValidator) {
            expectedValidationLength = validator.getNumberOfReturningTraces();
        }

        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        for (it = 0; ((it < limit) && monitor.isRunning()); it++) {
            while (monitor.isPaused()) {
            }
            network.nextExampleRandom();

            this.singleStep(it + 1);

            if ((it % logCounter) == 0) {
                if (this.testAndLog(it, expectedTestLength, expectedValidationLength)) {
                    it = limit;
                    layerCounter = layersMax + 1;
                }
            }

        }
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        this.endPlotter();
        this.endSaver();

        return network;
    }

    public void evaluateModel(int it) {
        trainer.evaluateModel(it);
    }

    public void updateModel(int it) {

        if ((it % insertEvery == 0) && (layerCounter < layersMax)) {
            HiddenUnit tempHiddenUnit;
            HiddenLayer tempHiddenLayer;
            
            // Insert hidden layer
            tempHiddenUnit = new HiddenUnit(new HyperbolicTangentActivationFunction());
            tempHiddenLayer = new HiddenLayer(tempHiddenUnit);

            ((AbstractFNNXFeedForwardNetwork)network).lastLayer();
            ((AbstractFNNXFeedForwardNetwork)network).getLayerDoPrevious();
            while (((AbstractFNNXFeedForwardNetwork)network).hasLayer()) {
                tempHiddenUnit.connectBackward(((AbstractFNNXFeedForwardNetwork)network).getLayerDoPrevious());
            }
            tempHiddenUnit.connectForward(((AbstractFNNXFeedForwardNetwork)network).getLastLayer());
            tempHiddenUnit.initNeuron(gen);

            ((AbstractFNNXFeedForwardNetwork)network).insertHiddenLayer(tempHiddenLayer);

            layerCounter++;
        }
        trainer.updateModel(it);
    }

    private int layersMax;
    private String outputActivationFunction;
    private AbstractFNNXTrainingAlgorithm trainer;
    private int insertEvery;

    private int it;
    private int layerCounter;
    private Random gen;
}
