/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXEasyTeacherMemory;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import java.util.Random;
import javax.swing.JFrame;
import model.AbstractFNNXModel;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import visualFNNXEasyTeacherMemory.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXEasyTeacherMemory extends AbstractFNNXTrainingAlgorithm implements FNNXObject, FNNXFunction, FNNXTrainingAlgorithm {

    public FNNXEasyTeacherMemory(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[12];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "pDiff";
        parameterNames[3] = "buffer";
        parameterNames[4] = "retry";
        parameterNames[5] = "trainer";
        parameterNames[6] = "epochs";
        parameterNames[7] = "tester";
        parameterNames[8] = "validator";
        parameterNames[9] = "logCounter";
        parameterNames[10] = "plotter";
        parameterNames[11] = "saver";

        parameterTypes = new Class[12];
        parameterTypes[0] = AbstractFNNXModel.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = Number.class;
        parameterTypes[3] = Number.class;
        parameterTypes[4] = Number.class;
        parameterTypes[5] = AbstractFNNXTrainingAlgorithm.class;
        parameterTypes[6] = Number.class;
        parameterTypes[7] = AbstractFNNXFunction.class;
        parameterTypes[8] = AbstractFNNXFunction.class;
        parameterTypes[9] = Number.class;
        parameterTypes[10] = AbstractFNNXParallelObserver.class;
        parameterTypes[11] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[12];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = true;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = true;
        isParameterMandatory[6] = true;
        isParameterMandatory[7] = false;
        isParameterMandatory[8] = false;
        isParameterMandatory[9] = false;
        isParameterMandatory[10] = false;
        isParameterMandatory[11] = false;

        parameters = new Object[12];

        //default values
        stopIfTesterBelow = 0;
        logCounter = 100;

    }

    public FNNXEasyTeacherMemory(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[12];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "pDiff";
        parameterNames[3] = "buffer";
        parameterNames[4] = "retry";
        parameterNames[5] = "trainer";
        parameterNames[6] = "epochs";
        parameterNames[7] = "tester";
        parameterNames[8] = "validator";
        parameterNames[9] = "logCounter";
        parameterNames[10] = "plotter";
        parameterNames[11] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        boolean flag;
        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXModel)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        pDiff = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, pDiff, 0)) {
                            return false;
                        }
                        if (!testMax(i, pDiff, 1)) {
                            return false;
                        }
                        break;
                    case 3:
                        bufferLength = ((Number)parameters[i]).intValue();
                        if (!testMin(i, bufferLength, 0)) {
                            return false;
                        }
                        break;
                    case 4:
                        retry = ((Number)parameters[i]).intValue();
                        if (!testMin(i, retry, 0)) {
                            return false;
                        }
                        break;
                    case 5:
                        trainer = (AbstractFNNXTrainingAlgorithm)parameters[i];
                        break;
                    case 6:
                        epochs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, epochs, 0)) {
                            return false;
                        }
                        break;
                    case 7:
                        tester = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 8:
                        validator = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 9:
                        logCounter = ((Number)parameters[i]).intValue();
                        if (!testMin(i, logCounter, 0)) {
                            return false;
                        }
                        break;
                    case 10:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 11:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Impossible to set dataset in " + this.name);
            return false;
        }
        
        flag = true;
        flag &= trainer.configure(trainer.getParameterIndex("dataset"), dataset, monitor);
        flag &= trainer.configure(trainer.getParameterIndex("network"), network, monitor);
        flag &= trainer.configure(trainer.getParameterIndex("epochs"), epochs, monitor);
        flag &= trainer.prepare(monitor);
        if (!flag) {
            monitor.appendErrorMessage("Error. Impossible to configure trainer in " + this.name);
            return false;
        }

        int resBufferSize = 0;
        if (tester != null) {
            flag = true;
            flag &= tester.configure(tester.getParameterIndex("dataset"), dataset, monitor);
            flag &= tester.configure(tester.getParameterIndex("network"), network, monitor);
            flag &= tester.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure tester.");
                return false;
            }

            hasTester = true;
            resBufferSize += tester.getNumberOfReturningTraces();
        }
        else {
            hasTester = false;
        }

        if (validator != null) {
            flag = true;
            flag &= validator.configure(validator.getParameterIndex("network"), network, monitor);
            flag &= validator.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure validator.");
                return false;
            }

            hasValidator = true;
            resBufferSize += validator.getNumberOfReturningTraces();
        }
        else {
            hasValidator = false;
        }

        buffer = new DataBlock(dataset.getNParts());
        for (int i = 0; i < dataset.getNParts(); i++) {
            buffer.configDataset(i, bufferLength, dataset.getDataset(i).getNColumns());
        }
        bufferLength = 0;

        rand = new Random();

        datasetLength = dataset.getNRows();
        limit = epochs * datasetLength;

        res = new float[resBufferSize];
        results = new InterThreadBuffer(resBufferSize);

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }
    
    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int it;
        int expectedTestLength = 0;
        int expectedValidationLength = 0;
        lastError = Float.MAX_VALUE;

        this.initPlotter(limit, 1);
        this.initSaver();

        if (hasTester) {
            expectedTestLength = tester.getNumberOfReturningTraces();
            tester.initSaver();
        }
        if (hasValidator) {
            expectedValidationLength = validator.getNumberOfReturningTraces();
            tester.initSaver();
        }

        network.initDatasetSequential();
        for (it = 0; (it < limit) && (monitor.isRunning()); it++) {
            while (monitor.isPaused());

            //------------------------------------------------------------------
            while (!network.hasNextExampleSequential()) {
                network.initDatasetSequential();
            }
            network.nextExampleSequential();

            trainer.evaluateModel(it + 1);
            currError = network.getError();
            backupData();                       //6
            if (currError < lastError) {
                trainer.updateModel(it + 1);
                lastError = currError;
//                backupData();                   //4
            }
            else {
//                backupData();                   //3
                if (pDiff > rand.nextFloat()) {
                    trainer.updateModel(it + 1);
                    lastError = currError;
//                    backupData();               //2
                }
                else {
//                    System.err.println("ignoring... " + String.valueOf(it));

//                    backupData();               //1
                    useMemory(it);

                    it--;
                    continue;
                }
            }

//            backupData();                       //5
            useMemory(it);
            
            //------------------------------------------------------------------
            if ((it % logCounter) == 0) {
                if (this.testAndLog(it, expectedTestLength, expectedValidationLength)) {
                    it = limit;
                }
            }
        }

        this.endPlotter();
        this.endSaver();

        if (hasTester) {
            tester.endSaver();
        }
        if (hasValidator) {
            validator.endSaver();
        }

        return null;
    }

    public void evaluateModel(int it) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void updateModel(int it) {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public void backupData() {
        int indexBuffer;
        if (bufferLength >= buffer.getNRows()) {
            indexBuffer = rand.nextInt(bufferLength);
        }
        else {
            indexBuffer = bufferLength;
            bufferLength++;
        }
        for (int i = 0; i < dataset.getNParts(); i++) {
            buffer.getDataset(i).setRow(indexBuffer, dataset.getDataset(i).getRow(network.getObsCounter()-1), false);
        }
    }

    public void useMemory(int it) {
        boolean update;
        int indexBuffer;

        for (int r = 0; (r < retry) && (bufferLength > 0); r++) {       //try several times with buffered data
            indexBuffer = rand.nextInt(buffer.getNRows());              //choose one observation at random
            if (indexBuffer < bufferLength) {                           //if it is within the buffer
                network.takeExampleFrom(indexBuffer, buffer);

                trainer.evaluateModel(it + 1);                          //evaluate observation
                currError = network.getError();
                update = false;
                if (currError < lastError) {
                    update = true;                                      //update the model if the current error is smaller than the previous
                }
                else {
                    if (pDiff > rand.nextFloat()) {                     //update the model anyway with a certain probability
                        update = true;
                    }
                }
                if (update) {
                    trainer.updateModel(it + 1);                        //update
                    lastError = currError;
                    for (int i = 0; i < buffer.getNParts(); i++) {      //remove observation from the dataset - copying the last in the current position
                        buffer.getDataset(i).setRow(indexBuffer, buffer.getDataset(i).getRow(bufferLength-1), false);
                    }
                    bufferLength--;
                    break;
                }
            }
        }
    }

/*
    private void insertCurrentOrdered(float error) {
        float[] currentInput = ((AbstractFNNXUnsupervisedModel)network).getCurrentInput();
        float [] currentTarget = null;

        if (network instanceof AbstractFNNXSupervisedModel) {
            currentTarget = ((AbstractFNNXSupervisedModel)network).getCurrentTarget();
        }

        BufferedObservation newObs = new BufferedObservation(currentInput, currentTarget, error);

        if (buffer.size() == 0) {
            buffer.append(newObs);
        }
        else {
            if (buffer.size() >= bufferLength) {
                buffer.deleteLast();
            }
            buffer.head();
            while (buffer.hasNode()) {
                if (error < ((BufferedObservation)buffer.getCurrent()).getError()) {
                    buffer.insert(newObs);
                    break;
                }
                else {
                    buffer.next();
                }
            }
            if (!buffer.hasNode()) {
                buffer.tail();
                buffer.append(newObs);
            }
        }
    }
*/
    private float pDiff;
    private int bufferLength;
    private DataBlock buffer;
    private int retry;
    private AbstractFNNXTrainingAlgorithm trainer;
    private Random rand;

    float lastError;
    float currError;

    
}
