/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXConnectUnits;

import visualFNNXConnectUnits.ConfigurationWindow;
import plugin.*;
import java.util.Hashtable;
import javax.swing.JFrame;
import model.distributedRepresentation.AbstractPerceptronBaseLayer;
import networks.AbstractFNNXFeedForwardNetwork;
import scripting.InterThreadMonitor;
import scripting.Names;


/**
 *
 * @author hsatizab
 */
public class FNNXConnectUnits extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXConnectUnits(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[5];
        parameterNames[0] = "network";
        parameterNames[1] = "fromLayer";
        parameterNames[2] = "fromUnits";
        parameterNames[3] = "toLayer";
        parameterNames[4] = "toUnits";

        parameterTypes = new Class[5];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = Number.class;
        parameterTypes[2] = Number[].class;
        parameterTypes[3] = Number.class;
        parameterTypes[4] = Number[].class;

        isParameterMandatory = new boolean[5];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;
        isParameterMandatory[3] = true;
        isParameterMandatory[4] = false;

        parameters = new Object[5];
    }

    public FNNXConnectUnits(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[5];
        parameterNames[0] = "network";
        parameterNames[1] = "fromLayer";
        parameterNames[2] = "fromUnits";
        parameterNames[3] = "toLayer";
        parameterNames[4] = "toUnits";

    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        int l;
        int layer;
        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXFeedForwardNetwork)parameters[i];
                        break;

                    case 1:
                        layer = ((Number)parameters[i]).intValue();

                        if (!testMin(i, layer, 0)) {
                            return false;
                        }
                        if (!testMax(i, layer, network.getAmountOfLayers()-1)) {
                            return false;
                        }

                        l = 0;
                        network.firstLayer();
                        while (l < layer) {
                            network.getLayerDoNext();
                            l++;
                        }
                        fromLayer = network.getLayerDoNext();
                        break;

                    case 2:
                        fromUnits = (Number[])parameters[i];

                        for (int j = 0; j < fromUnits.length; j++) {
                            if (!testMin(i, fromUnits[j].intValue(), 0)) {
                                return false;
                            }
                            if (!testMax(i, fromUnits[j].intValue(), fromLayer.getAmountOfUnits()-1)) {
                                return false;
                            }
                        }
                        break;

                    case 3:
                        layer = ((Number)parameters[i]).intValue();

                        if (!testMin(i, layer, 0)) {
                            return false;
                        }
                        if (!testMax(i, layer, network.getAmountOfLayers()-1)) {
                            return false;
                        }

                        l = 0;
                        network.firstLayer();
                        while (l < layer) {
                            network.getLayerDoNext();
                            l++;
                        }
                        toLayer = network.getLayerDoNext();
                        break;

                    case 4:
                        toUnits = (Number[])parameters[i];

                        for (int j = 0; j < toUnits.length; j++) {
                            if (!testMin(i, toUnits[j].intValue(), 0)) {
                                return false;
                            }
                            if (!testMax(i, toUnits[j].intValue(), toLayer.getAmountOfUnits()-1)) {
                                return false;
                            }
                        }
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int f, t;

        if (fromUnits == null) {
            fromUnits = new Number[fromLayer.getAmountOfUnits()];
            for (f = 0; f < fromUnits.length; f++) {
                fromUnits[f] = f;
            }
        }

        if (toUnits == null) {
            toUnits = new Number[toLayer.getAmountOfUnits()];
            for (t = 0; t < toUnits.length; t++) {
                toUnits[t] = t;
            }
        }

        for (f = 0; f < fromUnits.length; f++) {
            for (t = 0; t < toUnits.length; t++) {
                fromLayer.connectForward(fromUnits[f].intValue(), toLayer, toUnits[t].intValue());
            }
        }
        return null;

    }

    private AbstractFNNXFeedForwardNetwork network;
    private AbstractPerceptronBaseLayer fromLayer;
    private Number[] fromUnits;
    private AbstractPerceptronBaseLayer toLayer;
    private Number[] toUnits;

}
