/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXCorrelationBatchTrainer;

import plugin.*;
import data.DataBlock;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.Random;
import javax.swing.JFrame;
import layers.FloatingLayer;
//import model.distributedRepresentation.Connector;
import networks.AbstractFNNXFeedForwardNetwork;
import neurons.FloatingUnit;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import visualFNNXCorrelationBatchTrainer.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXCorrelationBatchTrainer extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXCorrelationBatchTrainer(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[8];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "pool";
        parameterNames[3] = "eta";
        parameterNames[4] = "mu";
        parameterNames[5] = "epochs";
        parameterNames[6] = "plotter";
        parameterNames[7] = "saver";

        parameterTypes = new Class[8];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = Number.class;
        parameterTypes[3] = Number.class;
        parameterTypes[4] = Number.class;
        parameterTypes[5] = Number.class;
        parameterTypes[6] = AbstractFNNXParallelObserver.class;
        parameterTypes[7] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[8];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = true;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = true;
        isParameterMandatory[6] = false;
        isParameterMandatory[7] = false;

        parameters = new Object[8];

    }

    public FNNXCorrelationBatchTrainer(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[8];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "pool";
        parameterNames[3] = "eta";
        parameterNames[4] = "mu";
        parameterNames[5] = "epochs";
        parameterNames[6] = "plotter";
        parameterNames[7] = "saver";

    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        boolean flag;
        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXFeedForwardNetwork)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        poolSize = ((Number)parameters[i]).intValue();
                        if (!testMin(i, poolSize, 1)) {
                            return false;
                        }
                        break;
                    case 3:
                        eta = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, eta, 0)) {
                            return false;
                        }
                        break;
                    case 4:
                        mu = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, mu, 0)) {
                            return false;
                        }
                        break;
                    case 5:
                        epochs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, epochs, 0)) {
                            return false;
                        }
                        break;
                    case 6:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 7:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        res = new float[1];
        results = new InterThreadBuffer(res.length);
        
        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }
    
    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        float[][] eDev = new float[dataset.getNRows()][network.getAmountOfOutputs()];
        float[] cov = new float[network.getAmountOfOutputs()];
        float sCov[] = new float[poolSize];
        float[] oAv = new float[network.getAmountOfOutputs()];
        float[] eAv = new float[network.getAmountOfOutputs()];
        int i, j, n, k;
        FloatingUnit[] pool;
        FloatingLayer newLayer;
        Random gen = new Random();

        this.initPlotter(epochs, 1);
        this.initSaver();
        
        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
            return null;
        }

        //------------------------------------------------------------------
        //                  Calculate error deviations
        //------------------------------------------------------------------
        Arrays.fill(eAv, 0);
        for (network.initDatasetSequential(); network.hasNextExampleSequential(); network.nextExampleSequential()) {
            while (monitor.isPaused()) {
            }
            network.feedForward();
            network.calcModelError();
            for (network.getLastLayer().firstUnit(), k = 0; network.getLastLayer().hasUnit(); network.getLastLayer().nextUnit(), k++) {
                eAv[k] += network.getLastLayer().getError();
            }
        }
        for (network.getLastLayer().firstUnit(), k = 0; network.getLastLayer().hasUnit(); network.getLastLayer().nextUnit(), k++) {
            eAv[k] /= dataset.getNRows();
        }

        for (network.initDatasetSequential(), n = 0; network.hasNextExampleSequential(); network.nextExampleSequential(), n++) {
            while (monitor.isPaused()) {
            }
            network.feedForward();
            network.calcModelError();
            for (network.getLastLayer().firstUnit(), k = 0; network.getLastLayer().hasUnit(); network.getLastLayer().nextUnit(), k++) {
                eDev[n][k] = (float)network.getLastLayer().getError() - eAv[k];
            }
        }

        //--------------------------------------------------------------
        //          Train input connections of hidden candidates
        //--------------------------------------------------------------
        pool = new FloatingUnit[poolSize];
        for (i = 0; (i < pool.length) && monitor.isRunning(); i++) {
            while (monitor.isPaused()) {
            }

            pool[i] = new FloatingUnit(i);

            network.lastLayer();
            network.getLayerDoPrevious();
            while (network.hasLayer()) {
                pool[i].connectBackward(network.getLayerDoPrevious());
            }

            pool[i].initNeuron(gen);

            //--------------------------------------------------------------
            //              Update input connections
            //--------------------------------------------------------------
            for (j = 0; (j < epochs) && monitor.isRunning(); j++) {
                while (monitor.isPaused()) {
                }
                //--------------------------------------------------------------
                //              Calculate covariance sign
                //--------------------------------------------------------------
                Arrays.fill(oAv, 0);
                for (network.initDatasetSequential(); network.hasNextExampleSequential(); network.nextExampleSequential()) {
                    network.feedForward();
                    pool[i].calcOutput();
                    for (network.getLastLayer().firstUnit(), k = 0; network.getLastLayer().hasUnit(); network.getLastLayer().nextUnit(), k++) {
//                        if (Float.compare(pool[i].getOutput(), Float.NaN) == 0) {
//                            System.err.print("\t\t\t\t");
//                            System.err.print(pool[i].getOutput());
//                            System.err.print(" += ");
//                            System.err.print(oAv[k]);
//                            System.err.print("\t-> ");
//                            System.err.print(pool[i].getNeta());
//                            System.err.print(" = ");
//                            Connector temp;
//                            pool[i].firstBackwardConnection();
//                            while (pool[i].hasBackwardConnection()) {
//                                temp = pool[i].getBackwardConnectorDoNext();
//                                System.err.print(" ");
//                                System.err.print(temp.getRefNeuron().getOutput());
//                                System.err.print("*");
//                                System.err.print(temp.getWeightValue());
//                                System.err.print(",");
//                            }
//                            System.err.print("\tS -> ");
//                            System.err.print(pool[i].getActivation());
//                            System.err.println();
//                        }
                        oAv[k] += pool[i].getOutput();
                    }
                }
                for (network.getLastLayer().firstUnit(), k = 0; network.getLastLayer().hasUnit(); network.getLastLayer().nextUnit(), k++) {
                    oAv[k] /= dataset.getNRows();
                }

                Arrays.fill(cov, 0);
                for (network.initDatasetSequential(), n = 0; network.hasNextExampleSequential(); network.nextExampleSequential(), n++) {
                    network.feedForward();
                    pool[i].calcOutput();
                    for (network.getLastLayer().firstUnit(), k = 0; network.getLastLayer().hasUnit(); network.getLastLayer().nextUnit(), k++) {
//                        if ((Float.compare(pool[i].getOutput(), Float.NaN) == 0) || (Float.compare(oAv[k], Float.NaN) == 0) || (Float.compare(eDev[n][k], Float.NaN) == 0)) {
//                            System.err.print("\t\t\t");
//                            System.err.print(pool[i].getOutput());
//                            System.err.print(" - ");
//                            System.err.print(oAv[k]);
//                            System.err.print(" * ");
//                            System.err.print(eDev[n][k]);
//                            System.err.print(" = ");
//                            System.err.println(cov[k]);
//                        }
                        cov[k] += (pool[i].getOutput() - oAv[k]) * eDev[n][k];
                    }
                }
                sCov[i] = 0;
                for (k = 0; k < cov.length; k++) {
                    sCov[i] += Math.abs(cov[k]);
                }
                sCov[i] /= dataset.getNRows();

                //--------------------------------------------------------------
                //              One epoch - batch learning
                //--------------------------------------------------------------
                for (network.initDatasetSequential(), n = 0; network.hasNextExampleSequential() && monitor.isRunning(); network.nextExampleSequential(), n++) {
                    while (monitor.isPaused()) {
                    }
                    network.feedForward();
                    pool[i].calcOutput();

                    for (network.getLastLayer().firstUnit(), k = 0; network.getLastLayer().hasUnit(); network.getLastLayer().nextUnit(), k++) {
//                        if ((Float.compare(cov[k], Float.NaN) == 0) || (Float.compare(eDev[n][k], Float.NaN) == 0)) {
//                            System.err.print("\t\t");
//                            System.err.print(cov[k]);
//                            System.err.print(" * ");
//                            System.err.println(eDev[n][k]);
//                        }
                        pool[i].calcDelta(Math.signum(cov[k])*eDev[n][k]);
                        pool[i].addToWeightIncrements();
                    }
                }
                pool[i].updateWeightsDeltaRuleBatch(eta, mu, 0);

                res[0] = sCov[i];
                results.newData(res, j+1);
                //----------------------------------------------
                if (hasPlotter) {
                    plotter.getMonitor().newResults(results);
                }
                if (hasSaver) {
                    saver.getMonitor().newResults(results);
                }
                //----------------------------------------------

            }

            //--------------------------------------------------------------
            //                  Calculate covariance
            //--------------------------------------------------------------
            Arrays.fill(oAv, 0);
            for (network.initDatasetSequential(); network.hasNextExampleSequential(); network.nextExampleSequential()) {
                network.feedForward();
                pool[i].calcOutput();
                for (network.getLastLayer().firstUnit(), k = 0; network.getLastLayer().hasUnit(); network.getLastLayer().nextUnit(), k++) {
                    oAv[k] += pool[i].getOutput();
                }
            }
            for (network.getLastLayer().firstUnit(), k = 0; network.getLastLayer().hasUnit(); network.getLastLayer().nextUnit(), k++) {
                oAv[k] /= dataset.getNRows();
            }

//            System.err.println("---------------------------");
            Arrays.fill(cov, 0);
            for (network.initDatasetSequential(), n = 0; network.hasNextExampleSequential(); network.nextExampleSequential(), n++) {
                network.feedForward();
                pool[i].calcOutput();
                for (network.getLastLayer().firstUnit(), k = 0; network.getLastLayer().hasUnit(); network.getLastLayer().nextUnit(), k++) {
                    cov[k] += (pool[i].getOutput() - oAv[k]) * eDev[n][k];

//                    if (Float.compare(pool[i].getOutput(), Float.NaN) == 0) {
//                        System.err.println("ooops! output is NaN");
//                    }
//                    if (Float.compare(eDev[n][k], Float.NaN) == 0) {
//                        System.err.println("ooops! eDev is NaN");
//                    }
//                    System.err.print(pool[i].getOutput());
//                    System.err.print("\t");
//                    System.err.println(eDev[n][k]);
                }
            }
            sCov[i] = 0;
            for (k = 0; k < cov.length; k++) {
                sCov[i] += Math.abs(cov[k]);
            }
            sCov[i] /= dataset.getNRows();

//            System.err.println(sCov[i]);
        }
//        System.err.println("-----------");

        if (monitor.isRunning()) {
            //--------------------------------------------------------------
            //  Select the best candidate and disconnect discarded units
            //--------------------------------------------------------------
            for (i = 1, j = 0; i < pool.length; i++) {
                if (sCov[i] > sCov[j]) {
                    pool[j].disconnectBackward();
                    j = i;
                }
                else {
                    pool[i].disconnectBackward();
                }
            }
    //        //--------------------------------------------------------------
    //        //                  Disconnect discarded units
    //        //--------------------------------------------------------------
    //        for (i = 0; i < pool.length; i++) {
    //            if (i != j) {
    //                pool[i].disconnectBackward();
    //            }
    //        }
            //--------------------------------------------------------------
            //  Insert the best candidate in a new layer of the network
            //--------------------------------------------------------------
            newLayer = new FloatingLayer(pool[j]);
            network.insertHiddenLayer(newLayer);
            newLayer.connectForward(network.getLastLayer());
        }

        this.endPlotter();
        this.endSaver();

        return null;
    }

    private int poolSize;
    private AbstractFNNXFeedForwardNetwork network;
    private DataBlock dataset;
    private float eta;
    private float mu;
    private int epochs;

}
