/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXExportFeedForwardNetwork;

import visualFNNXExportFeedForwardNetwork.ConfigurationWindow;
import plugin.*;
import java.util.Hashtable;
import javax.swing.JFrame;
import model.distributedRepresentation.AbstractPerceptronBaseLayer;
import networks.AbstractFNNXFeedForwardNetwork;
import model.distributedRepresentation.AbstractPerceptronBase;
import model.distributedRepresentation.Connector;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;


/**
 *
 * @author hsatizab
 */
public class FNNXExportFeedForwardNetwork extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXExportFeedForwardNetwork(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[2];
        parameterNames[0] = "network";
        parameterNames[1] = "saver";

        parameterTypes = new Class[2];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[2];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;

        parameters = new Object[2];
    }

    public FNNXExportFeedForwardNetwork(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[2];
        parameterNames[0] = "network";
        parameterNames[1] = "saver";

    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXFeedForwardNetwork)parameters[0];
                        break;
                    case 1:
                        saver = (AbstractFNNXParallelObserver)parameters[1];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

//        res = new float[network.inventory()];
//        results = new InterThreadBuffer(res.length);

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        AbstractPerceptronBaseLayer tempLayer;
        AbstractPerceptronBase tempNeuron;
        Connector tempConnector;
        int numberOfNeurons;
        float[][] weights;
        String message;
        int c = 0;

        numberOfNeurons = network.inventory();
        res = new float[numberOfNeurons];
        results = new InterThreadBuffer(res.length);
        this.initSaver();

        weights = new float[numberOfNeurons][numberOfNeurons];

//        System.err.println(network.getAmountOfLayers());
        
        network.firstLayer();
        while (network.hasLayer()) {
            tempLayer = network.getLayerDoNext();
            message = "Layer " + String.valueOf(tempLayer.getNumber()) + ":";
            tempLayer.firstUnit();
            while (tempLayer.hasUnit()) {
                tempNeuron = (AbstractPerceptronBase)tempLayer.getUnitDoNext();
                message += ("\t" + String.valueOf(tempNeuron.getNumber()));
                tempNeuron.firstBackwardConnection();
                while (tempNeuron.hasBackwardConnection()) {
                    tempConnector = tempNeuron.getBackwardConnectorDoNext();
                    weights[tempConnector.getRefNeuron().getNumber()][tempNeuron.getNumber()] = (float)tempConnector.getWeightValue();
                }
            }
//            System.out.println(message);
            while (saver.getMonitor().hasErrorMessage()) {
                try {
                    Thread.sleep(10);
                }
                catch (InterruptedException ex) {
                }
            }

            saver.getMonitor().newMessage(message);
            results.setCommand(Names.interThreadCommands.writeMessage, 0);
            saver.getMonitor().newResults(results);
            
        }

        while (saver.getMonitor().hasErrorMessage()) {
            try {
                Thread.sleep(10);
            }
            catch (InterruptedException ex) {
            }
        }

        message = "";
        for (c = 0; c < numberOfNeurons; c++) {
            message += ("\t" + String.valueOf(c));
        }
        saver.getMonitor().newMessage(message);
        results.setCommand(Names.interThreadCommands.writeMessage, 0);
        saver.getMonitor().newResults(results);

        for (c = 0; c < numberOfNeurons; c++) {
            res = weights[c];
            results.newData(res, c);
            saver.getMonitor().newResults(results);
        }

        this.endSaver();
        
        return null;

    }

    private AbstractFNNXFeedForwardNetwork network;

}
