/*
 * MLP.java
 *
 * Created on 10. mai 2006, 10:10
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package toolFNNXMLP;

import functions.ExponentialActivationFunction;
import java.util.Hashtable;
import java.util.Random;
import javax.swing.JFrame;
import model.distributedRepresentation.AbstractActivationFunction;
import functions.HyperbolicTangentActivationFunction;
import functions.LinearActivationFunction;
import functions.SigmoidalActivationFunction;
import layers.InputLayer;
import networks.AbstractFNNXFeedForwardNetwork;
import networks.LayerList;
import neurons.BiasUnit;
import plugin.FNNXObject;
import scripting.InterThreadMonitor;
import scripting.Names;
import visualFNNXMLP.ConfigurationWindow;

/**
 *
 * @author hsatiza
 */
public class FNNXMLP extends AbstractFNNXFeedForwardNetwork implements FNNXObject {
    

    public FNNXMLP(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;

        parameterNames = new String[6];
        parameterNames[0] = "inputs";
        parameterNames[1] = "hidden";
        parameterNames[2] = "actFuncHidden";
        parameterNames[3] = "outputs";
        parameterNames[4] = "actFuncOutput";
        parameterNames[5] = "shortcut";

        parameterTypes = new Class[6];
        parameterTypes[0] = Number.class;
        parameterTypes[1] = Number[].class;
        parameterTypes[2] = String[].class;
        parameterTypes[3] = Number.class;
        parameterTypes[4] = String.class;
        parameterTypes[5] = Boolean.class;

        isParameterMandatory = new boolean[6];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = false;
        isParameterMandatory[2] = false;
        isParameterMandatory[3] = true;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = false;

        parameters = new Object[6];

        //default values
        nHidden = null;
        actFuncHidden = null;
        shortcut = false;
    }

    public FNNXMLP(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[6];
        parameterNames[0] = "inputs";
        parameterNames[1] = "hidden";
        parameterNames[2] = "actFuncHidden";
        parameterNames[3] = "outputs";
        parameterNames[4] = "actFuncOutput";
        parameterNames[5] = "shortcut";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        nInputs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, nInputs, 0)) {
                            return false;
                        }
                        break;

                    case 1:
                        nHidden = (Number[])parameters[i];
                        for (int j = 0; j < nHidden.length; j++) {
                            if (!testMin(i, nHidden[j].intValue(), 0)) {
                                return false;
                            }
                        }
                        break;

                    case 2:
                        actFuncHidden = ((String[])parameters[i]);
                        break;

                    case 3:
                        nOutputs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, nOutputs, 0)) {
                            return false;
                        }
                        break;

                    case 4:
                        actFuncOutput = (String)parameters[i];
                        break;

                    case 5:
                        shortcut = ((Boolean)parameters[i]).booleanValue();
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        layers = new LayerList();

        layers.addInput(nInputs);

        AbstractActivationFunction oaf = null;
        switch (Names.activationFunction.valueOf(actFuncOutput)) {
            case LINEAR:
                oaf = new LinearActivationFunction();
                break;
            case SIGM:
                oaf = new SigmoidalActivationFunction();
                break;
            case TANH:
                oaf = new HyperbolicTangentActivationFunction();
                break;
            case EXP:
                oaf = new ExponentialActivationFunction();
                break;
            default:
                monitor.appendErrorMessage("Error. Illegal activation function.");
                return false;

        }
        layers.addOutput(nOutputs, oaf);

        if ((nHidden != null) && (actFuncHidden != null)) {
            int i;
            if (nHidden.length != actFuncHidden.length) {
                if (actFuncHidden.length == 1) {
                    String temp = actFuncHidden[0];
                    actFuncHidden = new String[nHidden.length];
                    for (i = 0; i < nHidden.length; i++) {
                        actFuncHidden[i] = temp;
                    }
                }
                else {
                    monitor.appendErrorMessage("Error. The array of activation function does not coincide with the amount of hidden units.");
                    return false;
                }
            }

            for (i = 0; i < nHidden.length; i++) {
                AbstractActivationFunction haf = null;
                switch (Names.activationFunction.valueOf(actFuncHidden[i])) {
                    case LINEAR:
                        haf = new LinearActivationFunction();
                        break;
                    case SIGM:
                        haf = new SigmoidalActivationFunction();
                        break;
                    case TANH:
                        haf = new HyperbolicTangentActivationFunction();
                        break;
                    case EXP:
                        haf = new ExponentialActivationFunction();
                        break;
                    default:
                        monitor.appendErrorMessage("Error. Illegal activation function.");
                        return false;
                }
                this.insertHiddenLayer(nHidden[i].intValue(), haf);
            }
        }

        if (shortcut) {
            this.connectShortcut();
        }
        else {
            this.connectFeedForward();
        }
        gen = new Random();
        this.initModel();

        configured = true;
        return true;
    }

    public String toString() {
        String temp;
        
        temp = name;
        try {
            this.firstLayer();
            temp += String.valueOf(" [" + this.getLayerDoNext().getAmountOfUnits()) + "-";
            while (this.hasLayer()) {
                if (this.isLastLayer()) {
                    temp += String.valueOf(this.getLayerDoNext().getAmountOfUnits()) + "]";
                    break;
                }
                else {
                    temp += String.valueOf(this.getLayerDoNext().getAmountOfUnits()) + "-";
                }
            }
        }
        catch (NullPointerException ne) {
        }
        return temp;
    }

    public FNNXMLP(int nI, int nO, AbstractActivationFunction oaf) {
        this.setName();
        configurationWindow = null;
        variables = null;
        monitor = null;

        nInputs = nI;
        nOutputs = nO;

        if ((nInputs < 1) || (nOutputs < 1)) {
            nInputs = 0;
            nOutputs = 0;
            return;
        }

        layers = new LayerList();

        layers.addInput(nInputs);

        layers.addOutput(nOutputs, oaf);

        gen = new Random();
    }

    public FNNXMLP(InputLayer iL, int nO, AbstractActivationFunction oaf) {
        this.setName();
        configurationWindow = null;
        variables = null;
        monitor = null;

        if (iL.getLast() instanceof BiasUnit) {
            nInputs = iL.getAmountOfUnits()-1;
        }
        else {
            nInputs = iL.getAmountOfUnits();
        }
        nOutputs = nO;

        if ((nInputs < 1) || (nOutputs < 1)) {
            nInputs = 0;
            nOutputs = 0;
            return;
        }

        layers = new LayerList();

        layers.addInput(iL);

        layers.addOutput(nOutputs, oaf);

        gen = new Random();
    }

    private Number[] nHidden;
    private String[] actFuncHidden;
    private String actFuncOutput;
    private boolean shortcut;

}
