/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXMLPActivationTester;

import visualFNNXMLPActivationTester.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
//import layers.AbstractPerceptronLayer;
//import model.AbstractFNNXSupervisedModel;
import model.distributedRepresentation.AbstractPerceptronBaseLayer;
import networks.AbstractFNNXFeedForwardNetwork;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;

/**
 *
 * @author hsatizab
 */
public class FNNXMLPActivationTester extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXMLPActivationTester(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[3];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";

        parameterTypes = new Class[3];
//        parameterTypes[0] = AbstractFNNXSupervisedModel.class;
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[3];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;

        parameters = new Object[3];
    }

    public FNNXMLPActivationTester(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[3];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXFeedForwardNetwork)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }


        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }
        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Dimensions mismatch.");
            return null;
        }

        int j;
        int i;
        AbstractPerceptronBaseLayer tempLayer;

        res = new float[network.inventory()];
        results = new InterThreadBuffer(res.length);

        this.initSaver();

        for (network.initDatasetSequential(), i = 1; network.hasNextExampleSequential() && monitor.isRunning(); i++, network.nextExampleSequential()) {
            while (monitor.isPaused()) {
            }

            j = 0;
            network.feedForward();
            network.firstLayer();
            while (network.hasLayer()) {
                tempLayer = network.getLayerDoNext();
                tempLayer.firstUnit();
                while (tempLayer.hasUnit()) {
                    res[j] = tempLayer.getUnitDoNext().getOutput();
                    j++;
                }
            }
            
            //--------------------------------------------
            if (hasSaver) {
                results.newData(res, i);
                saver.getMonitor().newResults(results);
            }
            //--------------------------------------------
        }

        this.endSaver();
        
        return null;
    }

    private AbstractFNNXFeedForwardNetwork network;
    private DataBlock dataset;

}
