/*
 * MLP.java
 *
 * Created on 10. mai 2006, 10:10
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package toolFNNXMLPEnsemble;

import functions.ExponentialActivationFunction;
import java.util.Hashtable;
import java.util.Random;
import javax.swing.JFrame;
import model.distributedRepresentation.AbstractActivationFunction;
import functions.HyperbolicTangentActivationFunction;
import functions.LinearActivationFunction;
import functions.SigmoidalActivationFunction;
import networks.AbstractFNNXFeedForwardEnsemble;
import networks.AbstractFNNXFeedForwardNetwork;
import networks.LayerList;
import plugin.FNNXObject;
import scripting.InterThreadMonitor;
import scripting.Names;
import toolFNNXMLP.FNNXMLP;
import visualFNNXMLPEnsemble.ConfigurationWindow;

/**
 *
 * @author hsatiza
 */
public class FNNXMLPEnsemble extends AbstractFNNXFeedForwardEnsemble implements FNNXObject {
    

    public FNNXMLPEnsemble(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;

        parameterNames = new String[6];
        parameterNames[0] = "networks";
        parameterNames[1] = "inputs";
        parameterNames[2] = "hidden";
        parameterNames[3] = "actFuncHidden";
        parameterNames[4] = "outputs";
        parameterNames[5] = "actFuncOutput";

        parameterTypes = new Class[6];
        parameterTypes[0] = Number.class;
        parameterTypes[1] = Number.class;
        parameterTypes[2] = Number[].class;
        parameterTypes[3] = String[].class;
        parameterTypes[4] = Number.class;
        parameterTypes[5] = String.class;

        isParameterMandatory = new boolean[6];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;
        isParameterMandatory[3] = false;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = true;

        parameters = new Object[6];

    }

    public FNNXMLPEnsemble(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[6];
        parameterNames[0] = "networks";
        parameterNames[1] = "inputs";
        parameterNames[2] = "hidden";
        parameterNames[3] = "actFuncHidden";
        parameterNames[4] = "outputs";
        parameterNames[5] = "actFuncOutput";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        nNetworks = ((Number)parameters[i]).intValue();
                        if (!testMin(i, nNetworks, 0)) {
                            return false;
                        }
                        break;

                    case 1:
                        nInputs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, nInputs, 0)) {
                            return false;
                        }
                        break;

                    case 2:
                        nHidden = (Number[])parameters[i];
                        for (int j = 0; j < nHidden.length; j++) {
                            if (!testMin(i, nHidden[j].intValue(), 0)) {
                                return false;
                            }
                        }
                        break;

                    case 3:
                        actFuncHidden = ((String[])parameters[i]);
                        break;

                    case 4:
                        nOutputs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, nOutputs, 0)) {
                            return false;
                        }
                        break;

                    case 5:
                        actFuncOutput = (String)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                monitor.appendErrorMessage("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        layers = new LayerList();

        networks = new AbstractFNNXFeedForwardNetwork[nNetworks];

        layers.addInput(nInputs);

        AbstractActivationFunction oaf = null;
        switch (Names.activationFunction.valueOf(actFuncOutput)) {
            case LINEAR:
                oaf = new LinearActivationFunction();
                break;
            case SIGM:
                oaf = new SigmoidalActivationFunction();
                break;
            case TANH:
                oaf = new HyperbolicTangentActivationFunction();
                break;
            case EXP:
                oaf = new ExponentialActivationFunction();
                break;
            default:
                monitor.appendErrorMessage("Error. Illegal activation function.");
                return false;
        }
        layers.addOutput(nOutputs, new LinearActivationFunction());

        for (int i = 0; i < networks.length; i++) {
            networks[i] = new FNNXMLP(layers.getInputLayer(), nOutputs, oaf);

            networks[i].getLastLayer().firstUnit();         //connect last layers in order to calculate the average
            this.getLastLayer().firstUnit();
            while (networks[i].getLastLayer().hasUnit() && this.getLastLayer().hasUnit()) {
                networks[i].getLastLayer().getUnitDoNext().connectForward(this.getLastLayer().getUnitDoNext(), (float)1/networks.length);
//                System.err.println((float)1/networks.length);
            }
        }

        if (nHidden != null) {
            int i;
            if (nHidden.length != actFuncHidden.length) {
                if (actFuncHidden.length == 1) {
                    String temp = actFuncHidden[0];
                    actFuncHidden = new String[nHidden.length];
                    for (i = 0; i < nHidden.length; i++) {
                        actFuncHidden[i] = temp;
                    }
                }
                else {
                    monitor.appendErrorMessage("Error. The array of activation function does not coincide with the amount of hidden units.");
                    return false;
                }
            }

            for (i = 0; i < nHidden.length; i++) {
                AbstractActivationFunction haf = null;
                switch (Names.activationFunction.valueOf(actFuncHidden[i])) {
                    case LINEAR:
                        haf = new LinearActivationFunction();
                        break;
                    case SIGM:
                        haf = new SigmoidalActivationFunction();
                        break;
                    case TANH:
                        haf = new HyperbolicTangentActivationFunction();
                        break;
                    case EXP:
                        haf = new ExponentialActivationFunction();
                        break;
                    default:
                        monitor.appendErrorMessage("Error. Illegal activation function.");
                        return false;
                }
                this.addHiddenLayer(nHidden[i].intValue(), haf);
            }
        }

        this.insertHiddenOutputs();

        this.connectFeedForward();
        gen = new Random();
        this.initModel();

        configured = true;
        return true;
    }

    public String toString() {
        String temp;

        temp = this.name;
        try {
            temp += " = " + String.valueOf(this.getNumberOfNetworks()) + "xFNNXMLP [";
            this.firstLayer();
            temp += String.valueOf(this.getLayerDoNext().getAmountOfUnits()) + "-";
            while (this.hasLayer()) {
                if (this.isLastLayer()) {
                    temp += String.valueOf(this.getLayerDoNext().getAmountOfUnits()) + "]";
                    break;
                }
                else {
                    temp += String.valueOf(this.getLayerDoNext().getAmountOfUnits()) + "-";
                }
            }
            for (int i = 0; i < networks.length; i++) {
                temp += "\n" + this.getNetwork(i).toString();
            }
        }
        catch (NullPointerException ne) {
        }
        return temp;
    }

    private int nNetworks;
    private Number[] nHidden;
    private String[] actFuncHidden;
    private String actFuncOutput;


}
