/*
 * OutputLayer.java
 *
 * Created on 1. mai 2006, 16:36
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package layers;

import model.distributedRepresentation.AbstractActivationFunction;
import neurons.OutputUnit;

/**
 *
 * @author Hector
 */
public class OutputLayer extends AbstractPerceptronLayer {
    
    public OutputLayer() {
    }

    /**
     * Creates a new instance of OutputLayer
     *
     * @param   n   the number of units in the layer
     * @param   af  the activation function of the units in the layer
     */
    public OutputLayer(int n, AbstractActivationFunction af) {
        
        for(int i = 0; i < n; i++){
            this.append(new OutputUnit(af));
        }
    }
    
    /**
     * Returns a copy of the oobject.
     */
    public OutputLayer clone() {
        return (OutputLayer) super.clone();
    }

    /**
     * Returns the output of the current unit.
     *
     * @return  the output value of the unit pointed by the window
     */
    public double getOutput() {
        return ((OutputUnit) this.getCurrent()).getOutput();
    }

    /**
     * Returns the value of the current target.
     *
     * @return the value of the current target
     */
    public float getTarget() {
        return ((OutputUnit) this.getCurrent()).getTarget();
    }

    /**
     * Returns the error of the current unit.
     *
     * @return  the error value of the unit pointed by the window
     */
    public double getError() {
        return ((OutputUnit) this.getCurrent()).getError();
    }

    public float getSSE() {
        return SSE;
    }

    /**
     * Moves the window to the next unit.
     */
    public void nextUnit() {
        this.next();
    }
    
    /**
     * Calculate the error for every unit in the layer.
     *
     * @return   the error for this layer (SSE)
     */
    public float calcError() {
        float er;
        OutputUnit tempo;
        SSE = 0;
        isLearnt = true;
        
        this.head();
        while (this.hasNode()) {
            tempo = (OutputUnit) this.getDoNext();
            er = tempo.calcError();
            SSE += (er * er);
            isLearnt &= tempo.isLearnt();
        }
        SSE /= this.nNodes;
        
        return SSE;
    }
   
    /**
     * Calculate the error for every unit in the layer.
     *
     * @return   the error for this layer (SSE)
     */
    public float[] calcErrorByOutput() {
        float er;
        OutputUnit tempo;
        float[] SSEO = new float[this.nNodes];
        int i = 0;
        SSE = 0;
        isLearnt = true;

        this.head();
        while (this.hasNode()) {
            tempo = (OutputUnit) this.getDoNext();
            er = tempo.calcError();
            SSEO[i] = (er * er);
            SSE += SSEO[i];
            isLearnt &= tempo.isLearnt();
            i++;
        }
        SSE /= this.nNodes;

        return SSEO;
    }

    /**
     * Return the state of the {@link OutputUnit#isLearnt} flag.
     *
     * @return  true, if the actual pattern is well modelled, false otherwise
     */
    public boolean getLearntFlag() {
        return ((OutputUnit) this.getCurrent()).isLearnt();
    }
    
    /**
     * Return the state of the {@link #isLearnt} flag.
     *
     * @return  true, if the actual pattern is well modelled, false otherwise
     */
    public boolean isLearnt() {
        return isLearnt;
    }

    public void setTarget(float[] obs) {
        int s;

        if (obs.length < this.size()) {
            s = obs.length;
        }
        else {
            s = this.size();
        }

        this.head();
        for (int i = 0; i < s; i++) {
            ((OutputUnit)this.getUnitDoNext()).setTargetData(obs[i]);
        }
    }

    public float getForwardWeightSum() {
        return 0;
    }

    private float SSE;
    private boolean isLearnt;
}
