/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package networks;

import data.DataBlock;
import model.distributedRepresentation.AbstractActivationFunction;
import layers.HiddenLayer;
import layers.OutputLayer;
import model.FNNXLayeredNetwork;
import model.FNNXNeuralNetwork;
import model.distributedRepresentation.AbstractPerceptronBaseLayer;

/**
 *
 * @author hsatizab
 */
public abstract class AbstractFNNXFeedForwardEnsemble extends AbstractFNNXFeedForwardNetwork implements FNNXNeuralNetwork, FNNXLayeredNetwork {

    public void addHiddenLayer(int h, AbstractActivationFunction af) {
        HiddenLayer hiddenAll = new HiddenLayer();
        HiddenLayer temp;
        for (int i = 0; i < networks.length; i++) {
            temp = new HiddenLayer(h, af);
            networks[i].insertHiddenLayer(temp);
            hiddenAll.addUnits(temp);
        }
        super.insertHiddenLayer(hiddenAll);
    }

    public void insertHiddenOutputs() {
        OutputLayer outputAll = new OutputLayer();
        for (int i = 0; i < networks.length; i++) {
            outputAll.addUnits(networks[i].getLastLayer());
        }
        lastLayer();
        layers.insert(outputAll);
    }

    public void connectFeedForward() {
        if (this.connected) {
            return;
        }

        for (int i = 0; i < networks.length; i++) {
            networks[i].connectFeedForward();
        }

        this.connected = true;
    }

    public AbstractFNNXFeedForwardNetwork getNetwork(int n) {
        return networks[n];
    }

    public void feedForward() {
        layers.getInputLayer().setInput(currentInput);
        layers.getOutputLayer().setTarget(currentTarget);

        for (int i = 0; i < networks.length; i++) {
            networks[i].feedForward();
        }
        this.getLastLayer().calcOutput();

    }

    public float calcModelError() {
        for (int i = 0; i < networks.length; i++) {
            networks[i].calcModelError();
        }
        error = layers.getOutputLayer().calcError();
        return error;
    }

    public void setMinErrPercent(int c) {
        for (int i = 0; i < networks.length; i++) {
            networks[i].setMinErrPercent(c);
        }
        super.setMinErrPercent(c);
    }

    public int getNumberOfNetworks() {
        return networks.length;
    }

    public void initModel() {
        for (int i = 0; i < networks.length; i++) {
            networks[i].initModel();
        }
    }

    public boolean setDataset(DataBlock data) {
        if(!super.setDataset(data)) {
            return false;
        }

        for (int i = 0; i < networks.length; i++) {
            if (!networks[i].setDataset(data)) {
                return false;
            }
        }

        return true;
    }

    public void nextExampleRandom() {
        float[] i, t;
        super.nextExampleRandom();
        t = this.getCurrentTarget();
        i = this.getCurrentInput();
        for (int n = 0; n < networks.length; n++) {
            networks[n].setCurrentInput(i);
            networks[n].setCurrentTarget(t);
        }
    }

    public boolean hasNextExampleSequential() {
        float[] i, t;

        if (super.hasNextExampleSequential()) {
            t = this.getCurrentTarget();
            i = this.getCurrentInput();
            for (int n = 0; n < networks.length; n++) {
                networks[n].setCurrentInput(i);
                networks[n].setCurrentTarget(t);
            }
            return true;
        }
        else {
            return false;
        }
    }

    public void setInput(int i, float value) {
        super.setInput(i, value);

        for (int n = 0; n < networks.length; n++) {
            networks[n].setInput(i, value);
        }
    }

    public void calcDelta() {
        for (int nC = 0; nC < networks.length; nC++) {
//            if (!networks[nC].isLastPatternLearnt()) {
                networks[nC].calcDelta();                //network deltas calculation
//            }
        }
    }

    public void updateWeightsDeltaRuleOnline(float eta, float mu, float wd) {
        for (int nC = 0; nC < networks.length; nC++) {
//            if (!networks[nC].isLastPatternLearnt()) {
                networks[nC].updateWeightsDeltaRuleOnline(eta, mu, wd);            //update weights
//            }
        }
    }

    public void updateWeightsDeltaRuleBatch(float eta, float mu, float wd) {
        for (int nC = 0; nC < networks.length; nC++) {
            networks[nC].updateWeightsDeltaRuleBatch(-eta, mu, wd);            //update weights
        }
    }

    public void addToWeightIncrements() {
        for (int nC = 0; nC < networks.length; nC++) {
            if (!networks[nC].isLastPatternLearnt()) {
                networks[nC].addToWeightIncrements();
            }
        }
    }

    public void reset() {
        for (int n = 0; n < networks.length; n++) {
            networks[n].reset();
        }
    }

    public int inventory() {
        int l = 0;
        int c = 0;
        AbstractPerceptronBaseLayer tempLayer;

        this.firstLayer();
        tempLayer = this.getLayerDoNext();
        tempLayer.firstUnit();
        tempLayer.setNumber(l);
//        System.err.print(tempLayer.getNumber());
//        System.err.print(":");
//        System.err.println(tempLayer.getAmountOfUnits());
        while (tempLayer.hasUnit()) {
            tempLayer.getUnitDoNext().setNumber(c);
            c++;
        }
        l++;

        for (int i = 0; i < networks.length; i++) {
            networks[i].firstLayer();
            networks[i].getLayerDoNext();
        }

        while (networks[0].hasLayer()) {
            this.getLayerDoNext().setNumber(l);
            for (int i = 0; i < networks.length; i++) {
                tempLayer = networks[i].getLayerDoNext();
                tempLayer.firstUnit();
                tempLayer.setNumber(l);
//                System.err.print(tempLayer.getNumber());
//                System.err.print(":");
//                System.err.println(tempLayer.getAmountOfUnits());
                while (tempLayer.hasUnit()) {
                    tempLayer.getUnitDoNext().setNumber(c);
                    c++;
                }
            }
            l++;
        }

        tempLayer = this.getLastLayer();
        tempLayer.firstUnit();
        tempLayer.setNumber(l);
//        System.err.print(tempLayer.getNumber());
//        System.err.print(":");
//        System.err.println(tempLayer.getAmountOfUnits());
        while (tempLayer.hasUnit()) {
            tempLayer.getUnitDoNext().setNumber(c);
            c++;
        }

        return c;
    }

    protected AbstractFNNXFeedForwardNetwork[] networks;

}