/*
 * Network.java
 *
 * Created on 8. mai 2006, 11:14
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package networks;

import java.util.Random;
import model.distributedRepresentation.AbstractActivationFunction;
import model.distributedRepresentation.AbstractPerceptronBaseLayer;
import layers.AbstractPerceptronLayer;
import layers.HiddenLayer;
import layers.InputLayer;
import layers.OutputLayer;
import model.AbstractFNNXSupervisedModel;
import model.FNNXLayeredNetwork;
import model.FNNXNeuralNetwork;
import neurons.AbstractPerceptron;
import model.distributedRepresentation.Connector;
import neurons.OutputUnit;
import scripting.Names;

/**
 *
 * @author Hector
 */
//public abstract class AbstractFNNXFeedForwardNetwork extends AbstractFNNXObject implements FNNXNeuralNetwork {
public abstract class AbstractFNNXFeedForwardNetwork extends AbstractFNNXSupervisedModel implements FNNXNeuralNetwork, FNNXLayeredNetwork {
    
    public AbstractFNNXFeedForwardNetwork clone() {
        AbstractFNNXFeedForwardNetwork temp;

        try {
            temp = (AbstractFNNXFeedForwardNetwork) super.clone();
            temp.layers = (LayerList) this.layers.clone();
            temp.cloneConnections(this.layers);
            return temp;
        }
        catch (CloneNotSupportedException ex) {
            ex.printStackTrace();
            return null;
        }
    }

    public void cloneConnections(LayerList toCopy) {
        AbstractPerceptronLayer d1, d2;
        AbstractPerceptronLayer s1, s2;
        
        layers.tail();
        toCopy.tail();
        
        if (layers.hasNode() && toCopy.hasNode()) {
            d1 = (AbstractPerceptronLayer) layers.getDoPrevious();
            s1 = (AbstractPerceptronLayer) toCopy.getDoPrevious();
            while (layers.hasNode() && toCopy.hasNode()) {
                d2 = (AbstractPerceptronLayer) layers.getDoPrevious();
                s2 = (AbstractPerceptronLayer) toCopy.getDoPrevious();
                
                d1.connectBackwardClone(d2, s1);      //It connects d1 to d2 creating new connections with the values in s1
                d1 = d2;
                s1 = s2;
            }
        }
    }
    
    public void deleteConnections() {
        AbstractPerceptronLayer temp;
        layers.head();
        while (layers.hasNode()) {
            temp = ((AbstractPerceptronLayer)layers.getDoNext());
            temp.firstUnit();
            while (temp.hasUnit()) {
                temp.getUnitDoNext().deleteConnections();
            }
        }
        connected = false;
    }
    
    public void copyConnections(LayerList toCopy) {
        layers.tail();
        toCopy.tail();
        while (layers.hasNode() && toCopy.hasNode()) {
            ((AbstractPerceptronLayer)layers.getDoPrevious()).copyConnections(((AbstractPerceptronLayer)toCopy.getDoPrevious()));
        }
    }
    
    /** 
    * Connects the layers
    *
    */
    
    public void connectFeedForward() {
        AbstractPerceptronLayer temp1, temp2;
        if (connected) {
            return;
        }
        
        layers.head();
        if (layers.hasNode()) {
            temp1 = (AbstractPerceptronLayer) layers.getDoNext();
            while (layers.hasNode()) {
                temp2 = (AbstractPerceptronLayer) layers.getDoNext();
                temp1.connectForward(temp2);    //It connects temp1 to temp2
                temp1 = temp2;
            }
        }
        connected = true;
    }
    
    public void connectShortcut() {
        if (connected) {
            return;
        }
        
        layers.getInputLayer().connectForward(layers.getOutputLayer());
        this.connectFeedForward();
    }
    
    public void initModel() {
        AbstractPerceptronLayer temp;
        gen = new Random();
        
        layers.head();
        while (layers.hasNode()) {
            temp = (AbstractPerceptronLayer) layers.getDoNext();
            temp.initLayer(gen);
        }

    }
    
    public void loadCurrentInput() {
        layers.getInputLayer().setInput(currentInput);
    }
    
    public void loadCurrentTarget() {
        layers.getOutputLayer().setTarget(currentTarget);
    }
    
    public void feedForward() {
        AbstractPerceptronLayer temp;

        this.loadCurrentInput();
        this.loadCurrentTarget();
        
        layers.head();
        while (layers.hasNode()) {
            temp = (AbstractPerceptronLayer) layers.getDoNext();
            temp.calcOutput();
        }
    }
    
    public void calcDelta() {
        AbstractPerceptronLayer temp;

        if (!this.isLastPatternLearnt()) {
            layers.tail();
            while (layers.hasNode()) {
                temp = (AbstractPerceptronLayer) layers.getDoPrevious();
                temp.calcDelta();
            }
        }
        
    }

    public void updateWeightsDeltaRuleOnline(float eta, float mu, float wd) {
        AbstractPerceptronLayer temp;
        
        if (!this.isLastPatternLearnt()) {
            layers.tail();
            while (layers.hasNode()) {                                        //update weights backward
                temp = (AbstractPerceptronLayer) layers.getDoPrevious();
                temp.updateWeightsDeltaRuleOnline(eta, mu, wd);
            }
        }
    }

    public void addToWeightIncrements() {
        AbstractPerceptronLayer temp;

        layers.tail();
        while (layers.hasNode()) {                                        //update weights backward
            temp = (AbstractPerceptronLayer) layers.getDoPrevious();
            temp.addToWeightIncrements();
        }
    }

    public void updateWeightsDeltaRuleBatch(float eta, float mu, float wd) {
        AbstractPerceptronLayer temp;

        layers.tail();
        while (layers.hasNode()) {                                        //update weights backward
            temp = (AbstractPerceptronLayer) layers.getDoPrevious();
            temp.updateWeightsDeltaRuleBatch(eta, mu, wd);
        }
    }

    public void updateWeightsQuickprop(float eta, float mu) {
        AbstractPerceptronLayer temp;

        layers.tail();
        while (layers.hasNode()) {                                        //update weights backward
            temp = (AbstractPerceptronLayer) layers.getDoPrevious();
            temp.updateWeightsQuickprop(eta, mu);
        }
    }

    public int getAmountOfLayers () {
        return layers.size();
    }

    public void insertHiddenLayer(int h, AbstractActivationFunction af) {
        layers.insertHidden(h, af);
    }

    public void insertHiddenLayer(HiddenLayer h) {
        layers.insertHidden(h);
    }

    public void setMinErrPercent(int c) {
        int i = 0;
        minErrPercent = c;

        layers.getOutputLayer().head();
        while (layers.getOutputLayer().hasNode()) {
            ((OutputUnit)layers.getOutputLayer().getDoNext()).setMinErr((float)minErrPercent*outputDataset.getRange()[i]/100);
            i++;
        }

    }
    
    public int getMinErrPercent() {
        return minErrPercent;
    }

    public float calcModelError() {
        error = layers.getOutputLayer().calcError();
        return error;
    }

    public float[] calcModelErrorByOutput() {
        float[] e = layers.getOutputLayer().calcErrorByOutput();

        error = layers.getOutputLayer().getSSE();
        
        return e;
    }

    public boolean isLastPatternLearnt() {
        return layers.getOutputLayer().isLearnt();
    }

    public void firstLayer() {
        layers.head();
    }

    public void lastLayer() {
        layers.tail();
    }
    
    public AbstractPerceptronLayer getLayerDoNext() {
        return (AbstractPerceptronLayer) layers.getDoNext();
    }

    public AbstractPerceptronLayer getLayerDoPrevious() {
        return (AbstractPerceptronLayer) layers.getDoPrevious();
    }

    public boolean hasLayer() {
        return layers.hasNode();
    }

    public OutputLayer getLastLayer() {
        return (OutputLayer) layers.getLast();
    }

    public InputLayer getFirstLayer() {
        return (InputLayer) layers.getFirst();
    }
    
    public AbstractPerceptronLayer getCurrentLayer() {
        return (AbstractPerceptronLayer) layers.getCurrent();
    }

    public void calcInputsRelevance(Names.relevanceMethod m) {
        layers.tail();
        
        switch (m) {
            case GARSON:
            case OVERALL_CONNECTIONS:
            case ANNIGMA:
                while(layers.hasNode()) {
                    ((AbstractPerceptronLayer)layers.getDoPrevious()).calcRelevance(m);
                }
                break;
                
            default:
                while(layers.hasNode()) {
                    ((AbstractPerceptronLayer)layers.getDoPrevious()).calcRelevance(Names.relevanceMethod.OTHER);
                }
                break;
        }
    }

    public void useAllInputs() {
        layers.getInputLayer().useAllUnits();
    }
    
    public boolean isConnected() {
        return connected;
    }
    
    public void setSensitivity(double s) {
        sensitivity = s;
    }
    
    public double getSensitivity() {
        return sensitivity;
    }
    
    public boolean isLastLayer() {
        return layers.isLast();
    }
    
    public int inventory() {
        int l = 0;
        int c = 0;
        AbstractPerceptronBaseLayer tempLayer;
        
        this.firstLayer();
        while (this.hasLayer()) {
            tempLayer = this.getLayerDoNext();
            tempLayer.firstUnit();
            tempLayer.setNumber(l);
            while (tempLayer.hasUnit()) {
                tempLayer.getUnitDoNext().setNumber(c);
                c++;
            }
            l++;
        }
        return c;
    }

    public void showNetworkWeights() {
        AbstractPerceptronBaseLayer tempLayer;
        AbstractPerceptron tempNeuron;
        Connector tempConnector;
        int numberOfNeurons;
        float[][] weights;
        
        numberOfNeurons = this.inventory();
        weights = new float[numberOfNeurons][numberOfNeurons];

        System.err.print("FENNIX Network Weights\n\n");
        System.err.print("Indexes\n");

        this.firstLayer();
        while (this.hasLayer()) {
            tempLayer = this.getLayerDoNext();
            tempLayer.firstUnit();
            System.err.print("Layer\t" + String.valueOf(tempLayer.getNumber()) + "\n");
            tempLayer.firstUnit();
            while (tempLayer.hasUnit()) {
                tempNeuron = (AbstractPerceptron)tempLayer.getUnitDoNext();
                System.err.print(String.valueOf(tempNeuron.getNumber()) + "\t");
                tempNeuron.firstBackwardConnection();
                while (tempNeuron.hasBackwardConnection()) {
                    tempConnector = tempNeuron.getBackwardConnectorDoNext();
                    weights[tempConnector.getRefNeuron().getNumber()][tempNeuron.getNumber()] = (float)tempConnector.getWeightValue();
                }

            }
            System.err.println();
        }
        System.err.println();

        for (int from = 0; from < numberOfNeurons; from++) {
            for (int to = 0; to < numberOfNeurons; to++) {
                 System.err.print(String.valueOf(weights[from][to]) + "\t");
            }
            System.err.println();
        }

    }

    public float[] calcOutput() {
        this.feedForward();
        int o;

        for (o = 0, this.layers.getOutputLayer().firstUnit(); o < this.nOutputs && this.layers.getOutputLayer().hasUnit(); o++, layers.getOutputLayer().nextUnit()) {
            output[o] = (float)this.layers.getOutputLayer().getOutput();
        }

        return output;
    }

    public float getWeightSum() {
        float sum = 0;

        layers.head();
        while (layers.hasNode()) {
            sum += ((AbstractPerceptronBaseLayer)layers.getDoNext()).getForwardWeightSum();
        }
        return sum/layers.size();
    }

    public void reset() {
        layers.head();
        while (layers.hasNode()) {
            ((AbstractPerceptronBaseLayer)layers.getDoNext()).reset();
        }
    }

    protected LayerList layers;
    protected int minErrPercent;
    protected boolean connected;

    private double sensitivity;
}
