/*
 * AbstractPerceptron.java
 *
 * Created on 1. mai 2006, 10:39
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package neurons;

import model.distributedRepresentation.Connector;
import model.distributedRepresentation.AbstractPerceptronBase;
import java.io.Serializable;
import scripting.Names;

/**
 *
 * @author Hector
 */
public abstract class AbstractPerceptron extends AbstractPerceptronBase implements Cloneable, Serializable {
    
    /**
     * Creates a new instance of AbstractPerceptron
     */
    public AbstractPerceptron() {
        relevance = 1;
        receiveConnections = true;
    }

    /**
     * Returns a copy of the object.
     */
    public AbstractPerceptron clone() {
        AbstractPerceptron temp = (AbstractPerceptron) super.clone();
        return temp;
    }

    /**
     * Calculates the relevance of the unit.
     *
     * @param   m   the relevance metric to be used
     */
    public void calcRelevance(Names.relevanceMethod m) {
        if (!use) {
            relevance = Float.NaN;
            return;
        }
        
        Connector temp;
        relevance = 0;

        switch (m) {
            case GARSON:
                double acc;
                
                forwardLinks.head();
                while (forwardLinks.hasNode()) {
                    temp = (Connector) forwardLinks.getDoNext();

                    temp.getRefNeuron().firstBackwardConnection();
                    acc = 0;
                    while (temp.getRefNeuron().hasBackwardConnection()) {
                        acc += Math.abs(((Connector) temp.getRefNeuron().getBackwardConnectorDoNext()).getWeightValue());
                    }
                    
                    //relevance += ((Math.abs(temp.getWeightValue()) * temp.getRefNeuron().getRelevance()) / acc);
                    relevance += (Math.abs(temp.getWeightValue()) / acc);
                }
                break;
                
            case OVERALL_CONNECTIONS:
                forwardLinks.head();
                while (forwardLinks.hasNode()) {
                    temp = (Connector) forwardLinks.getDoNext();
                    relevance += (temp.getWeightValue() * ((AbstractPerceptron)temp.getRefNeuron()).relevance);
                }
                break;
                
            case ANNIGMA:
                forwardLinks.head();
                while (forwardLinks.hasNode()) {
                    temp = (Connector) forwardLinks.getDoNext();
                    relevance += Math.abs(temp.getWeightValue() * ((AbstractPerceptron)temp.getRefNeuron()).relevance);
                }
                break;

            default:
                relevance  = 0;
                break;
        }
    }
    
    /**
     * Returns the relevance of the unit.
     *
     * @return  the relevance value of the unit
     */
    public float getRelevance() {
        return relevance;
    }
    
    /**
     * Calculates the sensitivity of this unit with respect to a given unit.
     *
     * @param   n   the sensitivity is obtained with respect to this unit
     * @return  the value of sensitivity
     */
    public double calcSensitivity(AbstractPerceptronBase n) {

        if (this.equals(n)) {
            return 1.0;
        }
        else {
            double s = 0;
            Connector w;
            backwardLinks.head();
            while (backwardLinks.hasNode()) {
                w = (Connector)backwardLinks.getDoNext();
                s += w.getWeightValue() * ((AbstractPerceptron)w.getRefNeuron()).calcSensitivity(n);
            }
            s = s * d_output;
            return s;
        }
        
    }

    protected float relevance;

}
