/*
 * OutputUnit.java
 *
 * Created on 1. mai 2006, 11:19
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package neurons;

import model.distributedRepresentation.AbstractActivationFunction;
import model.distributedRepresentation.AbstractOutputFunction;
import functions.LinearOutputFunction;
import scripting.Names;

/**
 *
 * @author Hector
 */
public class OutputUnit extends AbstractPerceptron {
    
    /**
     * Creates a new instance of OutputUnit.
     *
     * @param   af  the activation function of the unit
     */
    public OutputUnit(AbstractActivationFunction af) {
        activationFunction = af;
        outputFunction = new LinearOutputFunction();
        minErr = 0;
        error = 0;
        relevance = 1;
        receiveConnections = true;
    }

    /**
     * Returns a copy of the object.
     */
    public OutputUnit clone() {
        OutputUnit temp = (OutputUnit) super.clone();
        temp.activationFunction = (AbstractActivationFunction) this.activationFunction.clone();
        temp.outputFunction = (AbstractOutputFunction) this.outputFunction.clone();
        return temp;
    }
    
    /**
     * Sets the object that is used as target.
     *
     * @param   ta  the object containing the target data
     */
    public void setTargetData(float ta) {
        target = ta;
    }

    /**
     *  Calculates the parameter delta of the unit.
     */
    public void calcDelta() {
        delta = this.d_output * error;
    }
    
    /**
     * Returns the target of the unit.
     *
     * @return  the target value of this unit.
     */
    public float getTarget() {
        return target;
    }
    
    /**
     * Sets the minimum error value of this unit. Below this value the prediction id considered as correct.
     *
     * @param   e   the new value of minimum error
     */
    public void setMinErr(float e) {
        minErr = e;
    }

    /**
     * Calculates the output error.
     *
     * @return  the error value
     */
    public float calcError() {
        error = output - target;
        if ((error < minErr) && (error > -minErr)) {
            isLearnt = true;
        }
        else {
            isLearnt = false;
        }

        return error;
    }

    /**
     * Returns the error of this unit.
     *
     * @return  the error value of this unit
     */
    public double getError() {
        return error;
    }
    
    /**
     * Returns the state of the {@link #isLearnt} flag.
     *
     * @return  true, if the error is lower than {@link #minErr}, false otherwise
     */
    public boolean isLearnt() {
        return isLearnt;
    }
    
    /**
     * Calculates the relevance of the unit.
     *
     * @param   m   the relevance metric to be used
     */
    public void calcRelevance(Names.relevanceMethod m) {
        switch (m) {
            case GARSON:
            case OVERALL_CONNECTIONS:
            case ANNIGMA:
                if (this.forwardLinks.size() == 0) {
                    relevance = 1;
                }
                else {
                    relevance = (float) 1 / this.forwardLinks.size();
                }
                break;
                
            default:
                relevance = 0;
                break;
        }
    }

    public float getForwardWeightSum() {
        return 0;
    }

    private float target;
    private float minErr;
    private float error;
    private boolean isLearnt;
}
