/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXOnlineBackprop;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
//import networks.AbstractFNNXFeedForwardEnsemble;
import networks.AbstractFNNXFeedForwardNetwork;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import visualFNNXOnlineBackprop.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXOnlineBackprop extends AbstractFNNXTrainingAlgorithm implements FNNXObject, FNNXFunction, FNNXTrainingAlgorithm {

    public FNNXOnlineBackprop(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[15];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "tester";
        parameterNames[3] = "validator";
        parameterNames[4] = "etaIni";
        parameterNames[5] = "etaEnd";
        parameterNames[6] = "mu";
        parameterNames[7] = "epochs";
        parameterNames[8] = "weightDecay";
        parameterNames[9] = "minErrorPercent";
        parameterNames[10] = "stopIfTesterBelow";
        parameterNames[11] = "logCounter";
        parameterNames[12] = "delay";
        parameterNames[13] = "plotter";
        parameterNames[14] = "saver";

        parameterTypes = new Class[15];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXFunction.class;
        parameterTypes[3] = AbstractFNNXFunction.class;
        parameterTypes[4] = Number.class;
        parameterTypes[5] = Number.class;
        parameterTypes[6] = Number.class;
        parameterTypes[7] = Number.class;
        parameterTypes[8] = Number.class;
        parameterTypes[9] = Number.class;
        parameterTypes[10] = Number.class;
        parameterTypes[11] = Number.class;
        parameterTypes[12] = Number.class;
        parameterTypes[13] = AbstractFNNXParallelObserver.class;
        parameterTypes[14] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[15];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;
        isParameterMandatory[3] = false;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = true;
        isParameterMandatory[6] = true;
        isParameterMandatory[7] = true;
        isParameterMandatory[8] = true;
        isParameterMandatory[9] = true;
        isParameterMandatory[10] = false;
        isParameterMandatory[11] = false;
        isParameterMandatory[12] = false;
        isParameterMandatory[13] = false;
        isParameterMandatory[14] = false;

        parameters = new Object[15];

        //default values
        delay = 0;
        stopIfTesterBelow = Float.NEGATIVE_INFINITY;
        logCounter = 100;
    }

    public FNNXOnlineBackprop(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[15];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "tester";
        parameterNames[3] = "validator";
        parameterNames[4] = "etaIni";
        parameterNames[5] = "etaEnd";
        parameterNames[6] = "mu";
        parameterNames[7] = "epochs";
        parameterNames[8] = "weightDecay";
        parameterNames[9] = "minErrorPercent";
        parameterNames[10] = "stopIfTesterBelow";
        parameterNames[11] = "logCounter";
        parameterNames[12] = "delay";
        parameterNames[13] = "plotter";
        parameterNames[14] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        boolean flag;
        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXFeedForwardNetwork)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        tester = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 3:
                        validator = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 4:
                        etaIni = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, etaIni, 0)) {
                            return false;
                        }
                        eta = etaIni;
                        break;
                    case 5:
                        etaEnd = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, etaEnd, 0)) {
                            return false;
                        }
                        if (!testMax(i, etaEnd, etaIni)) {
                            return false;
                        }
                        break;
                    case 6:
                        mu = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, mu, 0)) {
                            return false;
                        }
                        break;
                    case 7:
                        epochs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, epochs, 0)) {
                            return false;
                        }
                        break;
                    case 8:
                        weightDecay = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, weightDecay, 0)) {
                            return false;
                        }
                        break;
                    case 9:
                        minErrorPercent = ((Number)parameters[i]).intValue();
                        if (!testMin(i, minErrorPercent, 0)) {
                            return false;
                        }
                        break;
                    case 10:
                        stopIfTesterBelow = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, stopIfTesterBelow, 0)) {
                            return false;
                        }
                        break;
                    case 11:
                        logCounter = ((Number)parameters[i]).intValue();
                        if (!testMin(i, logCounter, 0)) {
                            return false;
                        }
                        break;
                    case 12:
                        delay = ((Number)parameters[i]).intValue();
                        if (!testMin(i, delay, 0)) {
                            return false;
                        }
                        break;
                    case 13:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 14:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        int resBufferSize = 0;
        if (tester != null) {
            flag = true;
            flag &= tester.configure(tester.getParameterIndex("dataset"), dataset, monitor);
            flag &= tester.configure(tester.getParameterIndex("network"), network, monitor);
            flag &= tester.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasTester = true;
            resBufferSize += tester.getNumberOfReturningTraces();
        }
        else {
            hasTester = false;
        }

        if (validator != null) {
            flag = true;
            flag &= validator.configure(validator.getParameterIndex("network"), network, monitor);
            flag &= validator.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation validator.");
                return false;
            }

            hasValidator = true;
            resBufferSize += validator.getNumberOfReturningTraces();
        }
        else {
            hasValidator = false;
        }

        datasetLength = dataset.getNRows();
        limit = epochs * datasetLength;

//        if (network instanceof AbstractFNNXFeedForwardEnsemble) {
//            ensemble = true;
//            nN = ((AbstractFNNXFeedForwardEnsemble)network).getNumberOfNetworks();
//        }
//        else {
//            ensemble = false;
//            nN = 1;
//        }
//        currentNetwork = (AbstractFNNXFeedForwardNetwork)network;


        res = new float[resBufferSize];
        results = new InterThreadBuffer(resBufferSize);

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }
    
    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int expectedTestLength = 0;
        int expectedValidationLength = 0;

        this.initPlotter(limit, 1);
        this.initSaver();

        if (!((AbstractFNNXFeedForwardNetwork)network).setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
            return null;
        }
        ((AbstractFNNXFeedForwardNetwork)network).setMinErrPercent(minErrorPercent);

        if (hasTester) {
            expectedTestLength = tester.getNumberOfReturningTraces();
        }
        if (hasValidator) {
            expectedValidationLength = validator.getNumberOfReturningTraces();
        }

        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        for (it = 0; ((it < limit) && monitor.isRunning()); it++) {
            while (monitor.isPaused()) {
            }

            network.nextExampleRandom();

            //------------------------------------------------------------------
            this.singleStep(it);
            //------------------------------------------------------------------

            if ((it % logCounter) == 0) {

                if (this.testAndLog(it, expectedTestLength, expectedValidationLength)) {
                    it = limit;
                }

                if (delay > 0) {
                    try {
                        Thread.sleep(delay);
                    }
                    catch (InterruptedException ex) {
                    }
                }
            }

        }
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        this.endPlotter();
        this.endSaver();

        return null;
    }

    public void evaluateModel(int it) {

        ((AbstractFNNXFeedForwardNetwork)network).feedForward();                       //network outputs calculation
        ((AbstractFNNXFeedForwardNetwork)network).calcModelError();

//        System.err.print("in: ");
//        System.err.print(Arrays.toString(network.getCurrentInput()));
//        System.err.print(" | target: ");
////        System.err.print(Arrays.toString(network.getCurrentTarget()));
//        network.getLastLayer().firstUnit();
//        System.err.print(network.getLastLayer().getTarget());
//        System.err.print(" | out: ");
//        network.getLastLayer().firstUnit();
//        System.err.print(network.getLastLayer().getOutput());
//        System.err.print(" | error: ");
//        System.err.print(network.getError());
//        System.err.println();

//        if (ensemble) {
//            for (int nC = 0; nC < nN; nC++) {
//                currentNetwork = ((AbstractFNNXFeedForwardEnsemble)network).getNetwork(nC);
//
//                System.err.print("in: ");
//                System.err.print(Arrays.toString(currentNetwork.getCurrentInput()));
//                System.err.print(" | target: ");
////                System.err.print(Arrays.toString(currentNetwork.getCurrentTarget()));
//                currentNetwork.getLastLayer().firstUnit();
//                System.err.print(currentNetwork.getLastLayer().getTarget());
//                System.err.print(" | out: ");
//                System.err.print(currentNetwork.getLastLayer().getOutput());
//                System.err.print(" | error: ");
//                System.err.print(currentNetwork.getError());
//                System.err.println();
//            }
//
//            System.err.println();
//
//        }


    }

    public void updateModel(int it) {
        if ((it % datasetLength) == 0) {
//            eta = etaIni + (float)( (etaIni - etaEnd) * ( 1 - 1 / ( 1 + Math.exp(10*(0.5-((float)it/limit))) ) ) );
            eta = etaEnd + (float)( (etaIni - etaEnd) * ( 1 - 1 / ( 1 + Math.exp(10*(0.5-((float)it/limit))) ) ) );
        }

        ((AbstractFNNXFeedForwardNetwork)network).calcDelta();
        ((AbstractFNNXFeedForwardNetwork)network).updateWeightsDeltaRuleOnline(-eta, mu, weightDecay);
        
//        if (ensemble) {
//            for (int nC = 0; nC < nN; nC++) {
//                currentNetwork = ((AbstractFNNXFeedForwardEnsemble)network).getNetwork(nC);
//                if (!currentNetwork.isLastPatternLearnt()) {
//                    currentNetwork.calcDelta();                //network deltas calculation
//                    currentNetwork.updateWeightsDeltaRuleOnline(-eta, mu, weightDecay);            //update weights
//                }
//            }
//        }
//        else {
//            if (!currentNetwork.isLastPatternLearnt()) {
//                currentNetwork.calcDelta();                //network deltas calculation
//                currentNetwork.updateWeightsDeltaRuleOnline(-eta, mu, weightDecay);            //update weights
//            }
//        }
    }
    
    private float etaIni;
    private float etaEnd;
    private float mu;
    private float weightDecay;
    private int minErrorPercent;
    private int delay;

    private float eta;
    private int it;
//    private AbstractFNNXFeedForwardNetwork currentNetwork;

}
