/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXSequentialBackprop;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import networks.AbstractFNNXFeedForwardNetwork;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import visualFNNXSequentialBackprop.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXSequentialBackprop extends AbstractFNNXTrainingAlgorithm implements FNNXObject, FNNXFunction, FNNXTrainingAlgorithm {

    public FNNXSequentialBackprop(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[13];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "tester";
        parameterNames[3] = "validator";
        parameterNames[4] = "eta";
        parameterNames[5] = "mu";
        parameterNames[6] = "epochs";
        parameterNames[7] = "weightDecay";
        parameterNames[8] = "minErrorPercent";
        parameterNames[9] = "stopIfTesterBelow";
        parameterNames[10] = "delay";
        parameterNames[11] = "plotter";
        parameterNames[12] = "saver";

        parameterTypes = new Class[13];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXFunction.class;
        parameterTypes[3] = AbstractFNNXFunction.class;
        parameterTypes[4] = Number.class;
        parameterTypes[5] = Number.class;
        parameterTypes[6] = Number.class;
        parameterTypes[7] = Number.class;
        parameterTypes[8] = Number.class;
        parameterTypes[9] = Number.class;
        parameterTypes[10] = Number.class;
        parameterTypes[11] = AbstractFNNXParallelObserver.class;
        parameterTypes[12] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[13];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;
        isParameterMandatory[3] = false;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = true;
        isParameterMandatory[6] = true;
        isParameterMandatory[7] = true;
        isParameterMandatory[8] = true;
        isParameterMandatory[9] = false;
        isParameterMandatory[10] = false;
        isParameterMandatory[11] = false;
        isParameterMandatory[12] = false;

        parameters = new Object[13];

        //default values
        delay = 0;
        stopIfTesterBelow = 0;
        logCounter = 100;
    }

    public FNNXSequentialBackprop(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[13];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "tester";
        parameterNames[3] = "validator";
        parameterNames[4] = "eta";
        parameterNames[5] = "mu";
        parameterNames[6] = "epochs";
        parameterNames[7] = "weightDecay";
        parameterNames[8] = "minErrorPercent";
        parameterNames[9] = "stopIfTesterBelow";
        parameterNames[10] = "delay";
        parameterNames[11] = "plotter";
        parameterNames[12] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        boolean flag;
        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXFeedForwardNetwork)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        tester = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 3:
                        validator = (AbstractFNNXFunction)parameters[i];
                        break;
                    case 4:
                        eta = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, eta, 0)) {
                            return false;
                        }
                        break;
                    case 5:
                        mu = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, mu, 0)) {
                            return false;
                        }
                        break;
                    case 6:
                        epochs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, epochs, 0)) {
                            return false;
                        }
                        break;
                    case 7:
                        weightDecay = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, weightDecay, 0)) {
                            return false;
                        }
                        break;
                    case 8:
                        minErrorPercent = ((Number)parameters[i]).intValue();
                        if (!testMin(i, minErrorPercent, 0)) {
                            return false;
                        }
                        break;
                    case 9:
                        stopIfTesterBelow = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, stopIfTesterBelow, 0)) {
                            return false;
                        }
                        break;
                    case 10:
                        delay = ((Number)parameters[i]).intValue();
                        if (!testMin(i, delay, 0)) {
                            return false;
                        }
                        break;
                    case 11:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 12:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        int resBufferSize = 0;
        if (tester != null) {
            flag = true;
            flag &= tester.configure(tester.getParameterIndex("dataset"), dataset, monitor);
            flag &= tester.configure(tester.getParameterIndex("network"), network, monitor);
            flag &= tester.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasTester = true;
            resBufferSize += tester.getNumberOfReturningTraces();
        }
        else {
            hasTester = false;
        }

        if (validator != null) {
            flag = true;
            flag &= validator.configure(validator.getParameterIndex("network"), network, monitor);
            flag &= validator.prepare(monitor);
            if (!flag) {
                monitor.appendErrorMessage("Error. Impossible to configure backpropagation tester.");
                return false;
            }

            hasValidator = true;
            resBufferSize += validator.getNumberOfReturningTraces();
        }
        else {
            hasValidator = false;
        }

        limit = epochs;

        res = new float[resBufferSize];
        results = new InterThreadBuffer(resBufferSize);

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }
    
    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int expectedTestLength = 0;
        int expectedValidationLength = 0;

        this.initPlotter(limit, 1);
        this.initSaver();

        if (!((AbstractFNNXFeedForwardNetwork)network).setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
            return null;
        }
        ((AbstractFNNXFeedForwardNetwork)network).setMinErrPercent(minErrorPercent);

        if (hasTester) {
            expectedTestLength = tester.getNumberOfReturningTraces();
        }
        if (hasValidator) {
            expectedValidationLength = validator.getNumberOfReturningTraces();
        }

        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        for (it = 0; ((it < limit) && monitor.isRunning()); it++) {
            while (monitor.isPaused()) {
            }

            for (network.initDatasetSequential(); network.hasNextExampleSequential(); network.nextExampleSequential()) {
                //------------------------------------------------------------------
                this.singleStep(it);
                //------------------------------------------------------------------
            }

            if (this.testAndLog(it, expectedTestLength, expectedValidationLength)) {
                it = limit;
            }

            if (delay > 0) {
                try {
                    Thread.sleep(delay);
                }
                catch (InterruptedException ex) {
                }
            }

        }
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        this.endPlotter();
        this.endSaver();

        return null;
    }

    public void evaluateModel(int it) {
        ((AbstractFNNXFeedForwardNetwork)network).feedForward();                       //network outputs calculation
        ((AbstractFNNXFeedForwardNetwork)network).calcModelError();
    }

    public void updateModel(int it) {
        ((AbstractFNNXFeedForwardNetwork)network).calcDelta();
        ((AbstractFNNXFeedForwardNetwork)network).updateWeightsDeltaRuleOnline(-eta, mu, weightDecay);
    }
    
    private float eta;
    private float mu;
    private float weightDecay;
    private int minErrorPercent;
    private int delay;

    private int it;
//    private AbstractFNNXFeedForwardNetwork currentNetwork;

}
