/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXSaver;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Date;
import java.util.Hashtable;
import javax.swing.JFrame;
import plugin.AbstractFNNXParallelObserver;
import plugin.FNNXObject;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import visualFNNXSaver.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXSaver extends AbstractFNNXParallelObserver implements FNNXObject {

    public FNNXSaver (Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;

        parameterNames = new String[2];
        parameterNames[0] = "file";
        parameterNames[1] = "ending";

        parameterTypes = new Class[2];
        parameterTypes[0] = String.class;
        parameterTypes[1] = Object.class;

        isParameterMandatory = new boolean[2];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = false;

        parameters = new Object[2];
    }

    public FNNXSaver (JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[2];
        parameterNames[0] = "file";
        parameterNames[1] = "ending";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        try {
            fileName = (String)parameters[0];
            if (fileName == null) {
                monitor.appendErrorMessage("Error. Missing file name.");
                return false;
            }
            file = null;
            if (parameters[1] != null) {
                ending = parameters[1].toString().replace('.', '_');

                int dot = fileName.lastIndexOf('.');
                if (dot >= 0) {
                    String path = fileName.substring(0, dot);
                    String ext = fileName.substring(dot);
                    fileName = path + "_" + ending + ext;
                }
                else {
                    fileName = fileName + "_" + ending + ".fxr";
                }
            }
            else {
                ending = "";
            }

        }
        catch(NullPointerException ne) {
            monitor.appendErrorMessage("Internal error. Null pointer.");
            return false;
        }
        catch(IndexOutOfBoundsException ie) {
            monitor.appendErrorMessage("Internal error. Index out of bounds.");
            return false;
        }
        catch(ClassCastException ce) {
            monitor.appendErrorMessage("Internal error. Class cast.");
            return false;
        }

        configured = true;
        return true;
    }

    public void run() {
        String dataString;
        float[] dataFloat;
        Object parameter;
        int i;

        System.out.println("Saver initialized.");
        while (observerMonitor.isRunning()) {
            observerResults = observerMonitor.getResults();

            switch (observerResults.getCommand()) {
                case doNothing:
                    break;

                case init:
//                    super.initialize(observerResults.getParameter());
                    observerResults = new InterThreadBuffer(observerResults.getParameter());
                    break;

                case openFile:
                    Date date = new Date();
                    try {
                        if (file == null) {
                            file = new BufferedWriter(new FileWriter(fileName, true));
                        }
                        else {
                            file.newLine();
                        }
                        file.write("//" + date.toString());
                        file.newLine();
                        file.write("//" + observerMonitor.getMessage());
                        file.newLine();
                        file.flush();
                    }
                    catch (IOException ie) {
                        file = null;
                        System.out.println("Error. Unable to open file \"" + fileName + "\"");
                    }
                    break;

                case writeMessage:
                    try {
//                        if (observerMonitor.hasErrorMessage()) {
                            file.write("//" + observerMonitor.getMessage());
                            file.newLine();
                            file.flush();
//                        }
                    }
                    catch (IOException ie) {
                        file = null;
                        System.out.println("Error. Unable to open file \"" + fileName + "\"");
                    }
                    break;

                case closeFile:
                    try {
                        file.close();
                    }
                    catch (IOException ie) {
                        file = null;
                        System.out.println("Error. Unable to open file \"" + fileName + "\"");
                    }
                    catch (NullPointerException ne) {
                    }
                    break;

                case newData:
                case lastData:
                    try {
                        dataFloat = observerResults.getData();
                        parameter = observerResults.getParameter();
                        if (parameter != null) {
                            dataString = String.valueOf(parameter);
                        }
                        else {
                            dataString = "";
                        }
                        for (i = 0; i < dataFloat.length; i++) {
                            dataString += "\t" + String.valueOf(dataFloat[i]);
                        }
                        if (file != null) {
                            file.write(dataString);
                            if (observerMonitor.hasMessage()) {
                                file.write("\t//" + observerMonitor.getMessage());
                            }
                            file.newLine();
                            file.flush();
                        }
                        else {
                            System.out.println("Error. File \"" + fileName + "\" is not opened.");
                        }

                    }
                    catch (IOException ex) {
                        System.out.println("Error. Unable to write in file \"" + fileName + "\"");
                    }
                    break;
                    
                default:
                    break;
            }
            if (!monitor.isRunning()) {
                observerMonitor.notRunning();
            }
        }

        try {
            file.close();
            System.out.println("Saver finished.");
        }
        catch (IOException ie) {
            file = null;
            System.out.println("Error. Unable to open file \"" + fileName + "\"");
        }
        catch (NullPointerException ne) {
        }
    }

//    public void initialize(int lines) {
//        super.initialize(lines);
//        observerResults = new InterThreadBuffer(lines);
//    }

    private String fileName;
    private String ending;
    private BufferedWriter file;
}
