/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXContrastiveDivergence;

import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import networks.AbstractFNNXRBM;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import visualFNNXContrastiveDivergence.ConfigurationWindow;

/**
 *
 * @author hsatizab
 */
public class FNNXContrastiveDivergence extends AbstractFNNXTrainingAlgorithm implements FNNXObject, FNNXFunction, FNNXTrainingAlgorithm {

    public FNNXContrastiveDivergence(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[11];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "eta";
        parameterNames[3] = "mu";
        parameterNames[4] = "weightDecay";
        parameterNames[5] = "nGibbs";
        parameterNames[6] = "epochs";
        parameterNames[7] = "stopIfTesterBelow";
        parameterNames[8] = "delay";
        parameterNames[9] = "plotter";
        parameterNames[10] = "saver";

        parameterTypes = new Class[11];
        parameterTypes[0] = AbstractFNNXRBM.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = Number.class;
        parameterTypes[3] = Number.class;
        parameterTypes[4] = Number.class;
        parameterTypes[5] = Number.class;
        parameterTypes[6] = Number.class;
        parameterTypes[7] = Number.class;
        parameterTypes[8] = Number.class;
        parameterTypes[9] = AbstractFNNXParallelObserver.class;
        parameterTypes[10] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[11];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = true;
        isParameterMandatory[4] = true;
        isParameterMandatory[5] = true;
        isParameterMandatory[6] = true;
        isParameterMandatory[7] = true;
        isParameterMandatory[8] = true;
        isParameterMandatory[9] = false;
        isParameterMandatory[10] = false;

        parameters = new Object[11];

        //default values
        nGibbs = 1;
        delay = 0;
        stopIfTesterBelow = Float.NEGATIVE_INFINITY;
    }

    public FNNXContrastiveDivergence(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[11];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "eta";
        parameterNames[3] = "mu";
        parameterNames[4] = "weightDecay";
        parameterNames[5] = "nGibbs";
        parameterNames[6] = "epochs";
        parameterNames[7] = "stopIfTesterBelow";
        parameterNames[8] = "delay";
        parameterNames[9] = "plotter";
        parameterNames[10] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXRBM)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        logCounter = dataset.getNRows();
                        break;
                    case 2:
                        eta = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, eta, 0)) {
                            return false;
                        }
                        break;
                    case 3:
                        mu = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, mu, 0)) {
                            return false;
                        }
                        break;
                    case 4:
                        weightDecay = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, weightDecay, 0)) {
                            return false;
                        }
                        break;
                    case 5:
                        nGibbs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, nGibbs, 1)) {
                            return false;
                        }
                        break;
                    case 6:
                        epochs = ((Number)parameters[i]).intValue();
                        if (!testMin(i, epochs, 0)) {
                            return false;
                        }
                        break;
                    case 7:
                        stopIfTesterBelow = ((Number)parameters[i]).floatValue();
                        if (!testMin(i, stopIfTesterBelow, 0)) {
                            return false;
                        }
                        break;
                    case 8:
                        delay = ((Number)parameters[i]).intValue();
                        if (!testMin(i, delay, 0)) {
                            return false;
                        }
                        break;
                    case 9:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 10:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        res = new float[1];
        results = new InterThreadBuffer(1);

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }
    
    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        this.initPlotter(epochs, 1);
        this.initSaver();

        if (!((AbstractFNNXRBM)network).setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Impossible to attach dataset to network.");
            return null;
        }

        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        eta /= dataset.getNRows();
        for (it = 0; ((it < epochs) && monitor.isRunning()); it++) {
            while (monitor.isPaused()) {
            }

            //------------------------------------------------------------------
            this.singleStep(it);
            //------------------------------------------------------------------

            if ((network.getAccError()/logCounter) < stopIfTesterBelow) {
                it = epochs;
            }
            this.log(it);

            if (delay > 0) {
                try {
                    Thread.sleep(delay);
                }
                catch (InterruptedException ex) {
                }
            }

        }
        //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        this.endPlotter();
        this.endSaver();

        return null;
    }

    public void evaluateModel(int it) {
        for (network.initDatasetSequential(); network.hasNextExampleSequential(); network.nextExampleSequential()) {
            ((AbstractFNNXRBM)network).clampVisibleUnits();
            ((AbstractFNNXRBM)network).computeOutputProductsReal();     //Provided by the data
            ((AbstractFNNXRBM)network).gibbsSampling(nGibbs);           //Gibbs sampling
            ((AbstractFNNXRBM)network).computeOutputProductsFantasy();  //Provided by the model
        }
        network.calcModelError();
    }

    public void updateModel(int it) {
        ((AbstractFNNXRBM)network).updateWeightsContrastiveDivergence(eta, mu, weightDecay);
    }
    
    
    private float eta;
    private float mu;
    private float weightDecay;
    private int nGibbs;
    private int delay;

    private int it;

}
