/*
 * MLP.java
 *
 * Created on 10. mai 2006, 10:10
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package toolFNNXRBM;

import java.util.Hashtable;
import java.util.Random;
import javax.swing.JFrame;
import networks.AbstractFNNXRBM;
import networks.RBMLayerList;
import plugin.FNNXObject;
import scripting.InterThreadMonitor;
import visualRBM.ConfigurationWindow;

/**
 *
 * @author hsatiza
 */
public class FNNXRBM extends AbstractFNNXRBM implements FNNXObject {
    

    public FNNXRBM(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;

        parameterNames = new String[2];
        parameterNames[0] = "layers";
        parameterNames[1] = "visible";

        parameterTypes = new Class[2];
        parameterTypes[0] = Number[].class;
        parameterTypes[1] = Boolean[].class;

        isParameterMandatory = new boolean[2];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;

        parameters = new Object[2];

        //default values
        layersSize = null;
        visible = null;

    }

    public FNNXRBM(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[2];
        parameterNames[0] = "layers";
        parameterNames[1] = "visible";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        layersSize = (Number[])parameters[i];
                        for (int j = 0; j < layersSize.length; j++) {
                            if (!testMin(i, layersSize[j].intValue(), 1)) {
                                return false;
                            }
                        }
                        break;

                    case 1:
                        visible = ((Boolean[])parameters[i]);
                        break;
                }
            }
            catch(NullPointerException ne) {
                System.out.println("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        if (layersSize.length != visible.length) {
            monitor.appendErrorMessage("Error. The array of visibility does not coincide with the amount of units.");
            return false;
        }
        
        if (!visible[0]) {
            monitor.appendErrorMessage("Error. First layer must be visible.");
            return false;
        }
        
        layers = new RBMLayerList();
        for (int i = 0; i < layersSize.length; i++) {
            this.appendLayer(layersSize[i].intValue(), visible[i]);
        }

        connect();
        gen = new Random();
        this.initModel();

        configured = true;
        return true;
    }

    public String toString() {
        String temp;
        
        temp = name;
        try {
            this.firstLayer();
            temp += String.valueOf(" [" + this.getLayerDoNext().getAmountOfUnits()) + "-";
            while (this.hasLayer()) {
                if (this.isLastLayer()) {
                    temp += String.valueOf(this.getLayerDoNext().getAmountOfUnits()) + "]";
                    break;
                }
                else {
                    temp += String.valueOf(this.getLayerDoNext().getAmountOfUnits()) + "-";
                }
            }
        }
        catch (NullPointerException ne) {
        }
        return temp;
    }

    public FNNXRBM(int[] nN, boolean[] v) {
        this.setName();
        configurationWindow = null;
        variables = null;
        monitor = null;

        if (nN.length != v.length) {
            layers = null;
            return;
        }
        
        nInputs = 0;
        layers = new RBMLayerList();
        for (int i = 0; i < nN.length; i++) {
            if (nN[i] < 1) {
                layers = null;
                return;
            }
            this.appendLayer(nN[i], v[i]);
        }
        
        gen = new Random();
    }

    private Number[] layersSize;
    private Boolean[] visible;

}
