/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package layers;

import model.distributedRepresentation.AbstractPerceptronBaseLayer;
import neurons.AbstractRBMNeuron;
import neurons.BiasRBMNeuron;
import neurons.RBMNeuron;

/**
 *
 * @author hsatizab
 */
public class RBMLayer extends AbstractPerceptronBaseLayer implements Cloneable {
    
    /**
     * Returns a copy of the object.
     */
    public RBMLayer clone() {
        RBMLayer tempc = null;
        tempc = (RBMLayer) super.clone();
        return tempc;
    }

    
    /**
     * Creates a new instance of RBMLayer.
     */
    public RBMLayer() {
        visible = false;
    }

    /**
     * Creates a new instance of RBMLayer.
     *
     * @param   n   the number of units in the layer
     */
    public RBMLayer(int n, boolean v) {
        
        int i;
        for(i = 0; i < n; i++){
            this.append(new RBMNeuron(v));
        }
        this.append(new BiasRBMNeuron());
        
        visible = v;
    }
    
    public boolean isVisible() {
        return visible;
    }
    
    public void clampUnits(float[] input) {
        if (visible) {
            int i;
            for (this.head(), i=0; (this.hasNode() && (i < input.length)); i++) {
                ((AbstractRBMNeuron)this.getDoNext()).setInput(input[i]);
            }
        }
        else {
            return;
        }
    }

    public void computeOutputProductsReal() {
        this.head();
        while (this.hasNode()) {
            ((AbstractRBMNeuron)this.getDoNext()).computeOutputProductsReal();
        }
    }
    
    public void computeOutputProductsFantasy() {
        this.head();
        while (this.hasNode()) {
            ((AbstractRBMNeuron)this.getDoNext()).computeOutputProductsFantasy();
        }
    }

    /**
     * Calculates the output of every unit in the layer
     */
    public void calcOutput(boolean fantasy) {
        this.head();
        while (this.hasNode()) {
            ((AbstractRBMNeuron)this.getDoNext()).calcNeta(fantasy);
        }
        this.head();
        while (this.hasNode()) {
            ((AbstractRBMNeuron)this.getDoNext()).calcOutput(fantasy);
        }
    }
    
    public void updateWeightsContrastiveDivergence(float eta, float mu, float weightDecay) {
        this.head();
        while (this.hasNode()) {
            ((AbstractRBMNeuron)this.getDoNext()).updateWeightsContrastiveDivergence(eta, mu, weightDecay);
        }
    }
    
    public float calcError() {
        float e = 0;
        this.head();
        while (this.hasUnit()) {
            e += ((AbstractRBMNeuron)this.getUnitDoNext()).calcError();
        }
        return e;
    }
    
    public void reset() {
        this.head();
        while (this.hasNode()) {
            ((AbstractRBMNeuron)this.getDoNext()).reset();
        }
    }
    
    private boolean visible;

}
