/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package networks;

import java.util.Random;
import layers.RBMLayer;
import model.AbstractFNNXUnsupervisedModel;
import model.FNNXNeuralNetwork;
import model.distributedRepresentation.Connector;
import neurons.AbstractRBMNeuron;

/**
 *
 * @author hsatizab
 */
public abstract class AbstractFNNXRBM extends AbstractFNNXUnsupervisedModel implements FNNXNeuralNetwork {
    
    public AbstractFNNXRBM clone() {
        AbstractFNNXRBM temp;

        try {
            temp = (AbstractFNNXRBM) super.clone();
            temp.layers = (RBMLayerList) this.layers.clone();
            temp.cloneConnections(this.layers);
            return temp;
        }
        catch (CloneNotSupportedException ex) {
            ex.printStackTrace();
            return null;
        }
    }

    public void cloneConnections(RBMLayerList toCopy) {
        RBMLayer d1, d2;
        RBMLayer s1, s2;
        
        layers.tail();
        toCopy.tail();
        
        if (layers.hasNode() && toCopy.hasNode()) {
            d1 = (RBMLayer) layers.getDoPrevious();
            s1 = (RBMLayer) toCopy.getDoPrevious();
            while (layers.hasNode() && toCopy.hasNode()) {
                d2 = (RBMLayer) layers.getDoPrevious();
                s2 = (RBMLayer) toCopy.getDoPrevious();
                
                d1.connectBackwardClone(d2, s1);      //It connects d1 to d2 creating new connections with the values in s1
                d1 = d2;
                s1 = s2;
            }
        }
    }
    
    public void deleteConnections() {
        RBMLayer temp;
        layers.head();
        while (layers.hasNode()) {
            temp = ((RBMLayer)layers.getDoNext());
            temp.firstUnit();
            while (temp.hasUnit()) {
                temp.getUnitDoNext().deleteConnections();
            }
        }
        connected = false;
    }
    
    public void copyConnections(RBMLayerList toCopy) {
        layers.tail();
        toCopy.tail();
        while (layers.hasNode() && toCopy.hasNode()) {
            ((RBMLayer)layers.getDoPrevious()).copyConnections(((RBMLayer)toCopy.getDoPrevious()));
        }
    }
    
    /** 
    * Connects the layers
    *
    */
    public void connect() {
        RBMLayer temp1, temp2;
        if (connected) {
            return;
        }
        
        layers.head();
        if (layers.hasNode()) {
            temp1 = (RBMLayer) layers.getDoNext();
            while (layers.hasNode()) {
                temp2 = (RBMLayer) layers.getDoNext();
                temp1.connectForward(temp2);    //It connects temp1 to temp2
                temp1 = temp2;
            }
        }
        connected = true;
    }
    
    public void initModel() {
        RBMLayer temp;
        gen = new Random();
        
        layers.head();
        while (layers.hasNode()) {
            temp = (RBMLayer) layers.getDoNext();
            temp.initLayer(gen);
        }
    }

    public int getAmountOfLayers () {
        return layers.size();
    }

    public void appendLayer(int h, boolean v) {
        layers.appendLayer(h, v);
        if (v) {
            nInputs += h;
        }
    }

    public void appendLayer(RBMLayer h, boolean v) {
        layers.appendLayer(h);
        if (v) {
            nInputs += h.size();
        }
    }

    public float calcModelError() {
        layers.head();
        error = 0;
        while (layers.hasNode()) {
            if (layers.isCurrentVisible()) {
                error += ((RBMLayer)layers.getDoNext()).calcError();
            }
            else {
                layers.next();
            }
        }
        
        error /= this.nInputs;
        this.accError += error;
        return error;
    }

    public void firstLayer() {
        layers.head();
    }

    public void lastLayer() {
        layers.tail();
    }
    
    public RBMLayer getLayerDoNext() {
        return (RBMLayer) layers.getDoNext();
    }

    public RBMLayer getLayerDoPrevious() {
        return (RBMLayer) layers.getDoPrevious();
    }

    public boolean hasLayer() {
        return layers.hasNode();
    }
    
    public boolean isLastLayer() {
        return layers.isLast();
    }
    
    public RBMLayer getLastLayer() {
        return (RBMLayer) layers.getLast();
    }

    public RBMLayer getFirstLayer() {
        return (RBMLayer) layers.getFirst();
    }

    public boolean isConnected() {
        return connected;
    }
    
    public int inventory() {
        int l = 0;
        int c = 0;
        RBMLayer tempLayer;
        
        this.firstLayer();
        while (this.hasLayer()) {
            tempLayer = this.getLayerDoNext();
            tempLayer.firstUnit();
            tempLayer.setNumber(l);
            while (tempLayer.hasUnit()) {
                tempLayer.getUnitDoNext().setNumber(c);
                c++;
            }
            l++;
        }
        return c;
    }

    public void showNetworkWeights() {
        RBMLayer tempLayer;
        AbstractRBMNeuron tempNeuron;
        Connector tempConnector;
        int numberOfNeurons;
        float[][] weights;
        
        numberOfNeurons = this.inventory();
        weights = new float[numberOfNeurons][numberOfNeurons];

        System.err.print("FENNIX Network Weights\n\n");
        System.err.print("Indexes\n");

        this.firstLayer();
        while (this.hasLayer()) {
            tempLayer = this.getLayerDoNext();
            tempLayer.firstUnit();
            System.err.print("Layer\t" + String.valueOf(tempLayer.getNumber()) + "\n");
            tempLayer.firstUnit();
            while (tempLayer.hasUnit()) {
                tempNeuron = (AbstractRBMNeuron)tempLayer.getUnitDoNext();
                System.err.print(String.valueOf(tempNeuron.getNumber()) + "\t");
                tempNeuron.firstBackwardConnection();
                while (tempNeuron.hasBackwardConnection()) {
                    tempConnector = tempNeuron.getBackwardConnectorDoNext();
                    weights[tempConnector.getRefNeuron().getNumber()][tempNeuron.getNumber()] = (float)tempConnector.getWeightValue();
                }

            }
            System.err.println();
        }
        System.err.println();

        for (int from = 0; from < numberOfNeurons; from++) {
            for (int to = 0; to < numberOfNeurons; to++) {
                 System.err.print(String.valueOf(weights[from][to]) + "\t");
            }
            System.err.println();
        }
    }

    public float getWeightSum() {
        float sum = 0;

        layers.head();
        while (layers.hasNode()) {
            sum += ((RBMLayer)layers.getDoNext()).getForwardWeightSum();
        }
        return sum/layers.size();
    }

    public void reset() {
        layers.head();
        while (layers.hasNode()) {
            ((RBMLayer)layers.getDoNext()).reset();
        }
    }
    
    public void clampVisibleUnits() {
        int clampedIndex = 0;
        int size;
        float[] input;
        
        layers.head();
        while (layers.hasNode()) {
            if (layers.isCurrentVisible()) {
                size = ((RBMLayer)layers.getCurrent()).size() - 1;          //subtract bias unit
                input = new float[size];
                System.arraycopy(currentInput, clampedIndex, input, 0, size);
                ((RBMLayer)layers.getDoNext()).clampUnits(input);
                clampedIndex += size;
            }
            else {
                layers.next();
            }
        }
    }

    public void computeOutputProductsReal() {
        layers.head();
        while (layers.hasNode()) {
            ((RBMLayer)layers.getCurrent()).calcOutput(false);
            ((RBMLayer)layers.getDoNext()).computeOutputProductsReal();
        }
    }
    
    public void computeOutputProductsFantasy() {
        layers.head();
        while (layers.hasNode()) {
            ((RBMLayer)layers.getDoNext()).calcOutput(true);
        }
        
        layers.head();
        while (layers.hasNode()) {
            ((RBMLayer)layers.getCurrent()).calcOutput(false);
            ((RBMLayer)layers.getDoNext()).computeOutputProductsFantasy();
        }
    }

    public void calcOutput(boolean fantasy) {
        layers.head();
        while (layers.hasNode()) {
            ((RBMLayer)layers.getDoNext()).calcOutput(fantasy);
        }
    }

    public void updateWeightsContrastiveDivergence(float eta, float mu, float weightDecay) {
        layers.head();
        while (layers.hasNode()) {
            ((RBMLayer)layers.getDoNext()).updateWeightsContrastiveDivergence(eta, mu, weightDecay);
        }
    }
    
    public void gibbsSampling(int n) {
        for (int i = 1; i < n; i++) {
            calcOutput(false);
            calcOutput(true);
        }
    }
    
    protected RBMLayerList layers;
    protected boolean connected;

}
