/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package neurons;

import data.MyFloat;
import java.io.Serializable;
import java.util.Random;
import model.distributedRepresentation.AbstractPerceptronBase;
import model.distributedRepresentation.AbstractPerceptronBaseLayer;
import model.distributedRepresentation.Connector;

/**
 *
 * @author hsatizab
 */
public class AbstractRBMNeuron extends AbstractPerceptronBase implements Cloneable, Serializable {

    /**
     * Creates a new instance of AbstractRBMNeuron
     */
    public AbstractRBMNeuron() {
        receiveConnections = true;
        visible = true;
    }
    
    /**
     * Creates a new instance of AbstractRBMNeuron
     */
    public AbstractRBMNeuron(boolean v) {
        receiveConnections = true;
        visible = v;
    }

    /**
     * Returns a copy of the object.
     */
    public AbstractRBMNeuron clone() {
        AbstractRBMNeuron temp = (AbstractRBMNeuron) super.clone();
        return temp;
    }
   
    /**
     * Connects this unit with the given unit.
     *
     * @param   an  the unit to connect with
     * @param   w   the new value of the connection
     */
    public void connectForward(AbstractPerceptronBase an, float w) {
        if (this.isAlreadyConnectedTo(an)) {
            return;
        }

        MyFloat temp1 = new MyFloat(w);
        MyFloat temp2 = new MyFloat(0);

        this.forwardLinks.append(new RBMConnector(an, temp1, temp2));
        an.appendToBackwardLinks(new RBMConnector(this, temp1, temp2));
    }

    /**
     * Connects this unit with the given unit.
     *
     * @param   an  the unit to connect with
     */
    public void connectForward(AbstractPerceptronBase an) {
        System.out.println("Connecting units");
        
        if (this.isAlreadyConnectedTo(an)) {
            return;
        }

        MyFloat temp1 = new MyFloat(1);
        MyFloat temp2 = new MyFloat(0);

        this.forwardLinks.append(new RBMConnector(an, temp1, temp2));
        an.appendToBackwardLinks(new RBMConnector(this, temp1, temp2));
    }

    /**
     * Connects this unit with the given units.
     *
     * @param   ld  the layer containing the units to connect with
     */
    public void connectForward(AbstractPerceptronBaseLayer ld) {
        ld.head();
        while (ld.hasNode()) {
            this.connectForward((AbstractRBMNeuron)ld.getDoNext());
        }
    }

    /**
     * Connects this unit with the given unit.
     *
     * @param   an  the unit to connect with
     * @param   w   the new value of the connection
     */
    public void connectBackward(AbstractPerceptronBase an, float w) {
        if (this.isAlreadyConnectedFrom(an)) {
            return;
        }

        MyFloat temp1 = new MyFloat(w);
        MyFloat temp2 = new MyFloat(0);

        this.backwardLinks.append(new RBMConnector(an, temp1, temp2));
        an.appendToForwardLinks(new RBMConnector(this, temp1, temp2));
    }
    
    /**
     * Connects this unit with the given unit.
     *
     * @param   an  the unit to connect with
     */
    public void connectBackward(AbstractPerceptronBase an) {
        if (this.isAlreadyConnectedFrom(an)) {
            return;
        }

        MyFloat temp1 = new MyFloat(1);
        MyFloat temp2 = new MyFloat(0);

        this.backwardLinks.append(new RBMConnector(an, temp1, temp2));
        an.appendToForwardLinks(new RBMConnector(this, temp1, temp2));
    }

    /**
     * Connects this unit with the given units.
     *
     * @param   ld  the layer containing the units to connect with
     */
    public void connectBackward(AbstractPerceptronBaseLayer ld) {
        ld.head();
        while (ld.hasNode()) {
            this.connectBackward((AbstractRBMNeuron)ld.getDoNext());
        }
    }

    /**
     * Copies the information saved in a connector in order to connect forward this unit with a given one.
     *
     * @param   an  the unit to connect with
     * @param   tc  the connector from where the weight values are copied
     */
    public void connectForwardClone(AbstractPerceptronBase an, Connector tc) {
        if (!an.receiveConnections()) {
            return;
        }

        MyFloat temp1 = new MyFloat(tc.getWeightValue());
        MyFloat temp2 = new MyFloat(tc.getLastWeightIncrement());

        this.forwardLinks.append(new RBMConnector(an, temp1, temp2));
        an.appendToBackwardLinks(new RBMConnector(this, temp1, temp2));
    }

    /**
     * Copies the information saved in a unit in order to connect forward this unit with a given layer.
     *
     * @param   ld  the layer to connect with
     * @param   tc  the unit from where the weight values are copied
     */
    public void connectForwardClone(AbstractPerceptronBaseLayer ld, AbstractPerceptronBase tc) {
        ld.head();
        tc.firstForwardConnection();

        while (ld.hasNode() && tc.hasForwardConnection()) {
            this.connectForwardClone((AbstractPerceptronBase)ld.getDoNext(), (Connector)tc.getForwardConnectorDoNext());   //It connect this to ld.unit using information in tc
        }
    }

    /**
     * Copies the information saved in a connector in order to connect backward this unit with a given one.
     *
     * @param   an  the unit to connect with
     * @param   cInfo   the connector from where the weight values are copied
     */
    public void connectBackwardClone(AbstractPerceptronBase an, Connector cInfo) {
        if (!this.receiveConnections) {
            return;
        }

        MyFloat temp1 = new MyFloat(cInfo.getWeightValue());
        MyFloat temp2 = new MyFloat(cInfo.getLastWeightIncrement());

        this.backwardLinks.append(new RBMConnector(an, temp1, temp2));
        an.appendToForwardLinks(new RBMConnector(this, temp1, temp2));
    }

    /**
     * Copies the information saved in a unit in order to connect backward this unit with a given layer.
     *
     * @param   lTo the layer to connect with
     * @param   uInfo   the unit from where the weight values are copied
     */
    public void connectBackwardClone(AbstractPerceptronBaseLayer lTo, AbstractPerceptronBase uInfo) {
        lTo.head();
        uInfo.firstBackwardConnection();

        while (lTo.hasNode() && uInfo.hasBackwardConnection()) {
            this.connectBackwardClone((AbstractPerceptronBase)lTo.getDoNext(), (Connector)uInfo.getBackwardConnectorDoNext());   //It connect this to lTo.units using information in uInfo
        }
    }

    public void setVisible(boolean v) {
        visible = v;
    }
    
    public boolean getVisible() {
        return visible;
    }
    
    public void calcOutput(boolean fantasy) {
        if (fantasy == visible) {
            activation = (float) (1 / (1 + Math.exp(neta)));
            
//            System.out.println("Activation");
//            System.out.println(activation);
            System.out.println();
            
            if (Math.random() < activation) {
                output = 1;
            }
            else {
                output = 0;
            }
        }
    }
    
    /**
     * Calculates the neta of the unit.
     */
    public void calcNeta(boolean fantasy) {
        RBMConnector temp;

        if (fantasy == visible) {
            if (use) {
                neta = 0;
                backwardLinks.head();
                while (backwardLinks.hasNode()) {
                    temp = (RBMConnector) backwardLinks.getDoNext();
                    neta += temp.getWeightValue() * temp.getRefNeuron().getActivation();
                }
                forwardLinks.head();
                while (forwardLinks.hasNode()) {
                    temp = (RBMConnector) forwardLinks.getDoNext();
                    neta += temp.getWeightValue() * temp.getRefNeuron().getActivation();
                }
                neta *= -1;
            }
        }
    }
    
    public void computeOutputProductsReal() {
        backwardLinks.head();
        while (backwardLinks.hasNode()) {
            ((RBMConnector)backwardLinks.getDoNext()).addToActivationProductReal(this.activation);
        }
    }

    public void computeOutputProductsFantasy() {
        backwardLinks.head();
        while (backwardLinks.hasNode()) {
            ((RBMConnector)backwardLinks.getDoNext()).addToActivationProductFantasy(this.activation);
        }
    }
    
    public void reset() {
        backwardLinks.head();
        while (backwardLinks.hasNode()) {
            ((RBMConnector)backwardLinks.getDoNext()).reset();
        }
    }
    
    public void updateWeightsContrastiveDivergence(float eta, float mu, float weightDecay) {
        RBMConnector temp;
        backwardLinks.head();
        while (backwardLinks.hasNode()) {
            temp = (RBMConnector)backwardLinks.getDoNext();
            temp.addToWeight( (mu * temp.getLastWeightIncrement()) + (eta * ((temp.getActivationProductReal() - temp.getActivationProductFantasy()) - (weightDecay * temp.getWeightValue()))) );
        }
    }
    
    public boolean isVisible() {
        return visible;
    }
    
    public void setInput(float d) {
        input = d;
        activation = d;
        output = d;
    }
    
    public float getInput() {
        return input;
    }
    
    public float calcError() {
        float e;
        e = input - output;
        e *= e;
        return e;
    }
    
    public boolean isBias() {
        return false;
    }
    
        /**
     * Initializes the unit.
     *
     * @param   gen the random number generator to use in the initialization
     */
    public void initNeuron(Random gen) {
        Connector temp;
        float r = (float)(2 / Math.sqrt(backwardLinks.size()));
        
        backwardLinks.head();
        while (backwardLinks.hasNode()) {
            temp = (Connector) backwardLinks.getDoNext();
            temp.setWeight((float)(2 * (gen.nextFloat() - 0.5) * r));     // uniform distribution [-r, r]
        }
        reset();
    }
    
    private boolean visible;
    private float input;

}
