/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXInputRelevance;

import visualFNNXInputRelevance.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import data.DataMatrix;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.Random;
import javax.swing.JFrame;
import networks.AbstractFNNXFeedForwardNetwork;
import model.distributedRepresentation.AbstractPerceptronBase;
import neurons.AbstractPerceptron;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;

/**
 *
 * @author hsatizab
 */
public class FNNXInputRelevance extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXInputRelevance(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.bars;

        parameterNames = new String[5];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "method";
        parameterNames[3] = "plotter";
        parameterNames[4] = "saver";

        parameterTypes = new Class[5];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = String.class;
        parameterTypes[3] = AbstractFNNXParallelObserver.class;
        parameterTypes[4] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[5];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = false;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = false;
        isParameterMandatory[4] = false;

        parameters = new Object[5];
    }

    public FNNXInputRelevance(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[5];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "method";
        parameterNames[3] = "plotter";
        parameterNames[4] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        try {
            network = (AbstractFNNXFeedForwardNetwork)parameters[0];
            dataset = (DataBlock)parameters[1];
            method = Names.relevanceMethod.valueOf((String)parameters[2]);
            plotter = (AbstractFNNXParallelObserver)parameters[3];
            saver = (AbstractFNNXParallelObserver)parameters[4];

            if (network == null) {
                monitor.appendErrorMessage("Error. No network to test output.");
                return false;
            }
            res = new float[dataset.getDataset(0).getNColumns()];
            results = new InterThreadBuffer(res.length);

            switch (method) {
                case SENSITIVITY:
                case INPUT_CANCELLATION:
                case INPUT_PERTURBATION_FIX:
                case INPUT_PERTURBATION_NOISE:
                    if (dataset == null) {
                        monitor.appendErrorMessage("Error. No dataset to test output.");
                        return false;
                    }
                break;

                default:
                    break;
            }

            rand = new Random();
        }
        catch(NullPointerException ne) {
            monitor.appendErrorMessage("Internal error. Null pointer.");
            return false;
        }
        catch(IndexOutOfBoundsException ie) {
            monitor.appendErrorMessage("Internal error. Index out of bounds.");
            return false;
        }
        catch(ClassCastException ce) {
            monitor.appendErrorMessage("Internal error. Class cast.");
            return false;
        }

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        this.initPlotter(-1, 1);
        this.initSaver();

        int i, p;

//        if (hasPlotter) {
//            results.setCommand(interThreadCommands.clearPlot, null);
//            plotter.getMonitor().newResults(results);
//        }

        float[][] map;

        switch (method) {
            case GARSON:
            case OVERALL_CONNECTIONS:
            case ANNIGMA:
                network.calcInputsRelevance(method);
                res = network.getFirstLayer().getRelevance();
                break;

            case INPUT_PERTURBATION_NOISE:
            case INPUT_PERTURBATION_FIX:
            case INPUT_CANCELLATION:
            case SENSITIVITY:
                if (!network.setDataset(dataset)) {
                    monitor.appendErrorMessage("Error. Impossible to calculate relevance. Dimensions mismatch.");
                    return null;
                }
                
                map = this.calcInputPatternRelevance();
                if (map == null) {
                    monitor.appendErrorMessage("Error. Impossible to calculate relevance.");
                    return null;
                }
                else {
                    res = new float[dataset.getDataset(0).getNColumns()];
                    for (i = 0; i < res.length; i++) {
                        if (Double.isNaN(map[0][i])) {
                            res[i] = Float.NaN;
                        }
                        else {
                            res[i] = 0;
                            for (p = 0; p < map.length; p++) {
                                res[i] += map[p][i];
                            }
                        }
                    }
                }
                break;
                
            default:
                monitor.appendErrorMessage("Error. Unknown relevance method.");
                return null;
        }

        double acc = 0;
        for (i = 0; i < res.length; i++) {
            if (Double.compare(res[i], Double.NaN) != 0) {
                acc += Math.abs(res[i]);
            }
        }
        network.setSensitivity(acc);

//        if (Double.compare(acc, 0.0) != 0) {
//            for (i = 0; i < res.length; i++) {
//                if (Double.compare(res[i], Float.NaN) != 0) {
//                    res[i] = res[i] / (float)acc;
//                }
//            }
//        }
//        else {
//            monitor.appendErrorMessage("Error. No active inputs found. Impossible to calculate relevance.");
//            return null;
//        }

        if (Double.compare(acc, 0.0) == 0) {
            acc = Float.MAX_VALUE;
        }
        for (i = 0; i < res.length; i++) {
            if (Double.compare(res[i], Float.NaN) != 0) {
                res[i] = (float) (res[i] / acc);
            }
        }
        
        results.lastData(res, dataset.getNRows());
        //--------------------------------------------
        if (hasPlotter) {
//            System.err.println("BEFORE relevance!!");
            plotter.getMonitor().newResults(results);
//            System.err.println("AFTER relevance!!");
        }
        //--------------------------------------------
        if (hasSaver) {
            saver.getMonitor().newResults(results);
        }
        //--------------------------------------------

        this.endPlotter();
        this.endSaver();

        return res;
    }

    /**
     * Calculates the relevance of each input in each observation.
     *
     * @param   m   the relevance metric to be used
     * @return  a   bidimensional array containing the relevance of each input for every observation
     */
    private float[][] calcInputPatternRelevance() {
        float[][] relMap;
        float f;
        int i, j, p = 0;
        double error;

        if (!network.setDataset(dataset)) {
            relMap = null;
            return relMap;
        }

        DataMatrix inputDataset = dataset.getDataset(0);
        AbstractPerceptronBase [] n = new AbstractPerceptronBase[inputDataset.getNColumns()];
        relMap = new float[inputDataset.getNRows()][inputDataset.getNColumns()];

        network.getFirstLayer().firstUnit();
        for (i = 0; i < n.length; i++) {
            n[i] = network.getFirstLayer().getUnitDoNext();
        }
//        network.initDatasetSequential();

        switch (method) {
            case INPUT_PERTURBATION_NOISE:
                for (network.initDatasetSequential(), p = 0; monitor.isRunning() && network.hasNextExampleSequential(); network.nextExampleSequential(), p++) {
                    while (monitor.isPaused()) {
                    }
                    network.feedForward();
                    error = network.calcModelError();
                    for (i = 0; (i < inputDataset.getNColumns()) && monitor.isRunning(); i++) {
                        while (monitor.isPaused()) {
                        }
                        network.hasNextExampleSequential();                     //to reset input values
                        if (!n[i].getUse()) {
                            relMap[p][i] = Float.NaN;
                        }
                        else {
                            relMap[p][i] = 0;
                            for (j = 0; j < 100; j++) {
                                f = (float)(((2*rand.nextDouble())-1)*(0.025*inputDataset.getRange()[i]));
                                network.setInput(i, f);
                                network.feedForward();
                                relMap[p][i] += Math.pow((network.calcModelError() - error), 2);
                            }
                        }
                    }
                }
                break;
                
            case INPUT_PERTURBATION_FIX:
                for (network.initDatasetSequential(), p = 0; monitor.isRunning() && network.hasNextExampleSequential(); network.nextExampleSequential(), p++) {
                    while (monitor.isPaused()) {
                    }
                    network.feedForward();
                    error = network.calcModelError();
                    for (i = 0; (i < inputDataset.getNColumns()) && monitor.isRunning(); i++) {
                        while (monitor.isPaused()) {
                        }
                        network.hasNextExampleSequential();                     //to reset input values
                        f = inputDataset.getRange()[i] * (float)0.025;
                        if (!n[i].getUse()) {
                            relMap[p][i] = Float.NaN;
                        }
                        else {
                            relMap[p][i] = 0;
                            network.setInput(i, f);
                            network.feedForward();
                            relMap[p][i] += Math.pow((network.calcModelError() - error), 2);

                            network.setInput(i, -f);
                            network.feedForward();
                            relMap[p][i] += Math.pow((network.calcModelError() - error), 2);
                        }
                    }
                }
                break;

            case INPUT_CANCELLATION:
                for (network.initDatasetSequential(), p = 0; monitor.isRunning() && network.hasNextExampleSequential(); network.nextExampleSequential(), p++) {
                    while (monitor.isPaused()) {
                    }
                    network.feedForward();
                    error = network.calcModelError();
                    for (i = 0; (i < inputDataset.getNColumns()) && monitor.isRunning(); i++) {
                        while (monitor.isPaused()) {
                        }
                        network.hasNextExampleSequential();                     //to reset input values
                        if (!n[i].getUse()) {
                            relMap[p][i] = Float.NaN;
                        }
                        else {
                            f = inputDataset.getMean()[i];
                            network.setInput(i, f);
                            network.feedForward();
                            relMap[p][i] = (float)Math.pow((network.calcModelError() - error), 2);
                        }
                    }
                }
                break;

            case SENSITIVITY:
                for (network.initDatasetSequential(), p = 0; monitor.isRunning() && network.hasNextExampleSequential(); network.nextExampleSequential(), p++) {
                    while (monitor.isPaused()) {
                    }
                    network.feedForward();
                    for (i = 0; (i < inputDataset.getNColumns()) && monitor.isRunning(); i++) {
                        while (monitor.isPaused()) {
                        }
                        if (!n[i].getUse()) {
                            relMap[p][i] = Float.NaN;
                        }
                        else {
                            network.getLastLayer().firstUnit();
                            relMap[p][i] = 0;
                            while (network.getLastLayer().hasUnit()) {
                                relMap[p][i] += Math.abs(((AbstractPerceptron)network.getLastLayer().getUnitDoNext()).calcSensitivity(n[i]));
                            }
                            relMap[p][i] = relMap[p][i] / network.getLastLayer().getAmountOfUnits();
                        }
                    }
                }
                break;

            default:
                relMap = null;
        }

        if (!monitor.isRunning()) {
            monitor.appendErrorMessage("Algorithm stopped by user.");
            return null;
        }
        return relMap;
    }


    private AbstractFNNXFeedForwardNetwork network;
    private DataBlock dataset;
    private Names.relevanceMethod method;
    private Random rand;

}
