/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXOutputTester;

import visualFNNXOutputTester.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import model.AbstractFNNXSupervisedModel;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import scripting.Names.interThreadCommands;

/**
 *
 * @author hsatizab
 */
public class FNNXOutputTester extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXOutputTester(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_multi;

        parameterNames = new String[4];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";
        parameterNames[3] = "plotter";

        parameterTypes = new Class[4];
        parameterTypes[0] = AbstractFNNXSupervisedModel.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXParallelObserver.class;
        parameterTypes[3] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[4];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;
        isParameterMandatory[3] = false;

        parameters = new Object[4];
    }

    public FNNXOutputTester(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[4];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";
        parameterNames[3] = "plotter";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXSupervisedModel)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 3:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                monitor.appendErrorMessage("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        res = new float[2*network.getAmountOfOutputs()];
        results = new InterThreadBuffer(res.length);

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }
        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Dimensions mismatch.");
            return null;
        }

        this.initPlotter(dataset.getNRows(), 2);
        this.initSaver();

        int j;
        int i;
        float[] output;
        float[] target;

        float temp;

        if (hasPlotter) {
            results.setCommand(interThreadCommands.clearPlot, 0);
            plotter.getMonitor().newResults(results);
        }

        for (network.initDatasetSequential(), i = 1; network.hasNextExampleSequential() && monitor.isRunning(); i++, network.nextExampleSequential()) {
            while (monitor.isPaused()) {
            }

            output = network.calcOutput();
            target = network.getCurrentTarget();

//            ((AbstractFNNXFeedForwardNetwork)network).firstLayer();
//            ((AbstractFNNXFeedForwardNetwork)network).getLayerDoNext();
//            temp = ((AbstractPerceptron)((AbstractFNNXFeedForwardNetwork)network).getLayerDoNext().getFirst()).getOutput();
//            System.err.println(temp);

            for (j = 0; (2*j) < res.length; j++) {
                res[2*j] = target[j];
                res[(2*j)+1] = output[j];
            }
            //--------------------------------------------
            if (hasPlotter) {
                results.newData(res, i);
                plotter.getMonitor().newResults(results);
            }
            //--------------------------------------------
            if (hasSaver) {
                results.newData(res, i);
                saver.getMonitor().newResults(results);
            }
            //--------------------------------------------
        }

        this.endPlotter();
        this.endSaver();
        
        return null;
    }

    private AbstractFNNXSupervisedModel network;
    private DataBlock dataset;

}
