/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXPerformanceTester;

import visualFNNXPerformanceTester.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import model.AbstractFNNXSupervisedModel;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;

/**
 *
 * @author hsatizab
 */
public class FNNXPerformanceTester extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXPerformanceTester(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.bars;

        parameterNames = new String[5];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";
        parameterNames[3] = "plotter";
        parameterNames[4] = "allowedError";

        parameterTypes = new Class[5];
        parameterTypes[0] = AbstractFNNXSupervisedModel.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXParallelObserver.class;
        parameterTypes[3] = AbstractFNNXParallelObserver.class;
        parameterTypes[4] = Number.class;

        isParameterMandatory = new boolean[5];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;
        isParameterMandatory[3] = false;
        isParameterMandatory[4] = true;

        parameters = new Object[5];
    }

    public FNNXPerformanceTester(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[5];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";
        parameterNames[3] = "plotter";
        parameterNames[4] = "allowedError";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXSupervisedModel)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 3:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 4:
                        allowedError = ((Number)parameters[i]).intValue();
                        if (!testMin(i, allowedError, 0)) {
                            return false;
                        }
                        if (!testMax(i, allowedError, 50)) {
                            return false;
                        }
                        break;
                }
            }
            catch(NullPointerException ne) {
                monitor.appendErrorMessage("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        if (network.getAmountOfOutputs() == 1) {
            res = new float[1];
        }
        else {
            res = new float[network.getAmountOfOutputs() + 1];
        }
        results = new InterThreadBuffer(res.length);

//        try {
//            network = (AbstractFNNXFeedForwardNetwork)parameters[0];
//            dataset = (DataBlock)parameters[1];
//            saver = (AbstractFNNXParallelObserver)parameters[2];
//            plotter = (AbstractFNNXParallelObserver)parameters[3];
//            allowedError = ((Number)parameters[4]).intValue();
//
//            if (network == null) {
//                monitor.appendErrorMessage("Error. No network to test performance.");
//                return false;
//            }
//            if (network.getAmountOfOutputs() == 1) {
//                res = new float[network.getAmountOfOutputs()];
//            }
//            else {
//                res = new float[network.getAmountOfOutputs() + 1];
//            }
//            results = new InterThreadBuffer(res.length);
//
//            if (dataset == null) {
//                monitor.appendErrorMessage("Error. No dataset to test performance.");
//                return false;
//            }
//
//            if ((allowedError < 0) || (allowedError > 100)) {
//                monitor.appendErrorMessage("Error. Allowed error must be a percentage of the range.");
//                return false;
//            }
//        }
//        catch(NullPointerException ne) {
//            monitor.appendErrorMessage("Internal error. Null pointer.");
//            return false;
//        }
//        catch(IndexOutOfBoundsException ie) {
//            monitor.appendErrorMessage("Internal error. Index out of bounds.");
//            return false;
//        }
//        catch(ClassCastException ce) {
//            monitor.appendErrorMessage("Internal error. Class cast.");
//            return false;
//        }

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int up = 0;
        int[] byOutput = new int[res.length];
        float[] currentOutput;
        float[] currentTarget;
        float[] error = new float[network.getAmountOfOutputs()];
        boolean[] isOLearnt = new boolean[error.length];
        boolean allOLearnt;
        float[] threshold = new float[error.length];

        this.initPlotter(-1, 1);
        this.initSaver();

        network.setDataset(dataset);
        for (up = 0; up < error.length; up++) {
            threshold[up] = network.getOutputDataset().getRange()[up] * allowedError / 100;
        }

        for (network.initDatasetSequential(); network.hasNextExampleSequential() && monitor.isRunning(); network.nextExampleSequential()) {
            while (monitor.isPaused()) {
            }

            currentOutput = network.calcOutput();
            currentTarget = network.getCurrentTarget();

            allOLearnt = true;
            for (up = 0; up < error.length; up++) {
                error[up] = Math.abs(currentTarget[up] - currentOutput[up]);
                if (error[up] > threshold[up]) {
                    isOLearnt[up] = false;
                    allOLearnt = false;
                }
                else {
                    isOLearnt[up] = true;
                }
            }

            up = 0;
            if (allOLearnt) {
                byOutput[up]++;
            }

            up++;
            while(up < byOutput.length) {
                if (isOLearnt[up-1]) {
                    byOutput[up]++;
                }
                up++;
            }
        }

        up = dataset.getNRows();
        if (up == 0) {
            up = 1;
        }

        for (int i = 0; i < res.length; i++) {
            res[i] = (float)byOutput[i] / up;
        }

        if (hasPlotter) {
            results.lastData(res, 1);
            plotter.getMonitor().newResults(results);
        }
        if (hasSaver) {
            results.newData(res, 1);
            saver.getMonitor().newResults(results);
        }

        this.endPlotter();
        this.endSaver();

        return res;
    }

    private AbstractFNNXSupervisedModel network;
    private DataBlock dataset;
    private int allowedError;

}
