/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXProfileTester;

import visualFNNXProfileTester.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import java.util.Arrays;
import java.util.Hashtable;
import javax.swing.JFrame;
import networks.AbstractFNNXFeedForwardNetwork;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;
import scripting.Names.interThreadCommands;

/**
 *
 * @author hsatizab
 */
public class FNNXProfileTester extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXProfileTester(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_multi;

        parameterNames = new String[5];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "input";
        parameterNames[3] = "saver";
        parameterNames[4] = "plotter";

        parameterTypes = new Class[5];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = Number.class;
        parameterTypes[3] = AbstractFNNXParallelObserver.class;
        parameterTypes[4] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[5];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = true;
        isParameterMandatory[3] = false;
        isParameterMandatory[4] = false;

        parameters = new Object[5];
    }

    public FNNXProfileTester(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[5];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "input";
        parameterNames[3] = "saver";
        parameterNames[4] = "plotter";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        try {
            network = (AbstractFNNXFeedForwardNetwork)parameters[0];
            dataset = (DataBlock)parameters[1];
            input = ((Number)parameters[2]).intValue();
            saver = (AbstractFNNXParallelObserver)parameters[3];
            plotter = (AbstractFNNXParallelObserver)parameters[4];

            if (input < 1) {
                System.out.println("Error. Select a valid input.");
                return false;
            }
            input--;
            if (input >= dataset.getDataset(0).getNColumns()) {
                System.out.println("Error. Input doesn't exist.");
                return false;
            }

            if (network == null) {
                monitor.appendErrorMessage("Error. No network to test output.");
                return false;
            }
            res = new float[network.getAmountOfOutputs()];
            results = new InterThreadBuffer(res.length);

            if (dataset == null) {
                monitor.appendErrorMessage("Error. No dataset to test output.");
                return false;
            }
        }
        catch(NullPointerException ne) {
            monitor.appendErrorMessage("Internal error. Null pointer.");
            return false;
        }
        catch(IndexOutOfBoundsException ie) {
            monitor.appendErrorMessage("Internal error. Index out of bounds.");
            return false;
        }
        catch(ClassCastException ce) {
            monitor.appendErrorMessage("Internal error. Class cast.");
            return false;
        }

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        float inputValue;
        float range = dataset.getDataset(0).getRange()[input];
        float min = dataset.getDataset(0).getMin()[input];
        int i, o;
        
        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Dimensions mismatch.");
            return null;
        }

        this.initPlotter(100, 1);
        this.initSaver();

        if (hasPlotter) {
            results.setCommand(interThreadCommands.clearPlot, 0);
            plotter.getMonitor().newResults(results);
        }

        for (i = 0; (i < 100) && (monitor.isRunning()); i++) {
            while (monitor.isPaused()) {
            }
//            network.initDatasetSequential();
            Arrays.fill(res, 0);

            inputValue = min + range * ((float)i / 99);

            for (network.initDatasetSequential(); network.hasNextExampleSequential(); network.nextExampleSequential()) {
//            while (network.hasNextExampleSequential()) {
//                network.nextExampleSequential();
                network.setInput(input, inputValue);
                network.feedForward();

                network.getLastLayer().firstUnit();
                o = 0;
                while (network.getLastLayer().hasUnit()) {
                    res[o] += network.getLastLayer().getUnitDoNext().getOutput();
                    o++;
                }
            }
            for (o = 0; o < res.length; o++) {
                res[o] /= dataset.getNRows();
            }

            //--------------------------------------------
            if (hasPlotter) {
                results.newData(res, i);
                plotter.getMonitor().newResults(results);
            }
            //--------------------------------------------
            if (hasSaver) {
                results.newData(res, i);
                saver.getMonitor().newResults(results);
            }
            //--------------------------------------------

        }

        this.endPlotter();
        this.endSaver();
        
        return null;
    }

    private AbstractFNNXFeedForwardNetwork network;
    private DataBlock dataset;
    private int input;

}
