/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXQuantizerTester;

import visualFNNXQuantizerTester.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import model.AbstractFNNXUnsupervisedModel;
import networkGNG.AbstractFNNXGNGNetwork;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;

/**
 *
 * @author hsatizab
 */
public class FNNXQuantizerTester extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXQuantizerTester(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.YvsT_single;

        parameterNames = new String[4];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";
        parameterNames[3] = "plotter";

        parameterTypes = new Class[4];
        parameterTypes[0] = AbstractFNNXUnsupervisedModel.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXParallelObserver.class;
        parameterTypes[3] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[4];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;
        isParameterMandatory[3] = false;

        parameters = new Object[4];
    }

    public FNNXQuantizerTester(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[4];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";
        parameterNames[3] = "plotter";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXUnsupervisedModel)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                    case 3:
                        plotter = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                monitor.appendErrorMessage("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        res = new float[3];
        results = new InterThreadBuffer(res.length);

        plotRes = new float[1];
        plotResults = new InterThreadBuffer(plotRes.length);

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        int it = 1;
        this.initPlotter(dataset.getNRows(), 1);
        this.initSaver();

        network.inventory();

        network.setDataset(dataset);
        for (network.initDatasetSequential(); network.hasNextExampleSequential() && monitor.isRunning(); network.nextExampleSequential(), it++) {
            while (monitor.isPaused()) {
            }
            res[0] = ((AbstractFNNXGNGNetwork)network).getNearest().getNumber();
            res[1] = network.calcModelError();
            res[2] = ((AbstractFNNXGNGNetwork)network).getNearest().getMaxDistance();

            plotRes[0] = res[0];

            if (hasPlotter) {
                plotResults.newData(plotRes, it);
                plotter.getMonitor().newResults(plotResults);
            }
            if (hasSaver) {
                results.newData(res, it);
                saver.getMonitor().newResults(results);
            }

        }

        this.endPlotter();
        this.endSaver();

        return null;
    }

    private AbstractFNNXUnsupervisedModel network;
    private DataBlock dataset;

    private float[] plotRes;
    private InterThreadBuffer plotResults;

}
