/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXR2Tester;

import visualFNNR2Tester.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import model.AbstractFNNXSupervisedModel;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;

/**
 *
 * @author hsatizab
 */
public class FNNXR2Tester extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXR2Tester(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[3];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";

        parameterTypes = new Class[3];
        parameterTypes[0] = AbstractFNNXSupervisedModel.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[3];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;

        parameters = new Object[3];
    }

    public FNNXR2Tester(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[3];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXSupervisedModel)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                monitor.appendErrorMessage("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        res = new float[network.getAmountOfOutputs()];
        results = new InterThreadBuffer(res.length);

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        float[] SSerr = new float[res.length];
        float[] SStot = dataset.getDataset(1).getMSStot();
        float[] e;
        int i;

        this.initSaver();

        network.setDataset(dataset);
        for (network.initDatasetSequential(); network.hasNextExampleSequential() && monitor.isRunning(); network.nextExampleSequential()) {
            while (monitor.isPaused()) {
            }
            e = network.calcModelErrorByOutput();
            for (i = 0; i < network.getAmountOfOutputs(); i++) {
                SSerr[i] += e[i];
            }
        }

        for (i = 0; i < res.length; i++) {
            res[i] = 1 - (SSerr[i] / SStot[i]);
            if (res[i] < -0.1) {
                res[i] = (float)-0.1;
            }
//            System.err.println("SSerr=" + String.valueOf(SSerr[i]));
//            System.err.println("SStot=" + String.valueOf(SStot[i]));
//            System.err.println("R2=" + String.valueOf(res[i]));
        }

        if (hasSaver) {
            results.newData(res, 1);
            saver.getMonitor().newResults(results);
        }

        this.endSaver();

        return res;
    }

    private AbstractFNNXSupervisedModel network;
    private DataBlock dataset;

}
