/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXROCTester;

import visualFNNXROCTester.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import networks.AbstractFNNXFeedForwardNetwork;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;

/**
 *
 * @author hsatizab
 */
public class FNNXROCTester extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXROCTester(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[3];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";

        parameterTypes = new Class[3];
        parameterTypes[0] = AbstractFNNXFeedForwardNetwork.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[3];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;

        parameters = new Object[3];
    }

    public FNNXROCTester(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[3];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        try {
            network = (AbstractFNNXFeedForwardNetwork)parameters[0];
            dataset = (DataBlock)parameters[1];
            saver = (AbstractFNNXParallelObserver)parameters[2];

            if (network == null) {
                monitor.appendErrorMessage("Error. No network to test output.");
                return false;
            }

            if (network.getAmountOfOutputs() != 1) {
                monitor.appendErrorMessage("Error. The network must have one single binary output.");
                return false;
            }

            if (dataset == null) {
                monitor.appendErrorMessage("Error. No dataset to test output.");
                return false;
            }

            res = new float[7];     //[acc, fpr, tpr, tp, fp, tn, fn]
            results = new InterThreadBuffer(res.length);

        }
        catch(NullPointerException ne) {
            monitor.appendErrorMessage("Internal error. Null pointer.");
            return false;
        }
        catch(IndexOutOfBoundsException ie) {
            monitor.appendErrorMessage("Internal error. Index out of bounds.");
            return false;
        }
        catch(ClassCastException ce) {
            monitor.appendErrorMessage("Internal error. Class cast.");
            return false;
        }

        configured = true;
        return true;
    }

    public int getNumberOfReturningTraces() {
        return 0;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }
        
        if (!network.setDataset(dataset)) {
            monitor.appendErrorMessage("Error. Dimensions mismatch.");
            return null;
        }

        this.initPlotter(dataset.getNRows(), 1);
        this.initSaver();

        int k;
        float interval = (dataset.getDataset(1).getMax()[0] - dataset.getDataset(1).getMin()[0]) / 100;
        float min = dataset.getDataset(1).getMin()[0];
        float targetThreshold = (dataset.getDataset(1).getMax()[0] + dataset.getDataset(1).getMin()[0]) / 2;

        for (k = 0; k < 100 && monitor.isRunning(); k++) {
            while (monitor.isPaused()) {
            }
            res[3] = 0;
            res[4] = 0;
            res[5] = 0;
            res[6] = 0;
            for (network.initDatasetSequential(); network.hasNextExampleSequential() && monitor.isRunning(); network.nextExampleSequential()) {
                while (monitor.isPaused()) {
                }
                network.feedForward();
                network.getLastLayer().firstUnit();
                if (network.getLastLayer().getTarget() > targetThreshold) {     //positive
                    if ((float)network.getLastLayer().getOutput() > min+(k*interval)) {
                        res[3]++;       //tp
                    }
                    else {
                        res[6]++;       //fn
                    }
                }
                else {                                                          //negative
                    if ((float)network.getLastLayer().getOutput() > min+(k*interval)) {
                        res[4]++;       //fp
                    }
                    else {
                        res[5]++;       //tn
                    }
                }
            }
            //ACC = (TP + TN) / (P + N)
            res[0] = (res[3] + res[5]) / dataset.getNRows();

            //FPR = FP / N = FP / (FP + TN)
            res[1] = res[4] / (res[4] + res[5]);

            //TPR = TP / P = TP / (TP + FN)
            res[2] = res[3] / (res[3] + res[6]);

            //--------------------------------------------
            if (hasSaver) {
                results.newData(res, k);
                saver.getMonitor().newResults(results);
            }
            //--------------------------------------------
        }

        this.endPlotter();
        this.endSaver();
        
        return null;
    }

    private AbstractFNNXFeedForwardNetwork network;
    private DataBlock dataset;

}
