/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package toolFNNXSSETester;

import visualFNNXSSETester.ConfigurationWindow;
import plugin.*;
import data.DataBlock;
import java.util.Hashtable;
import javax.swing.JFrame;
import model.AbstractFNNXSupervisedModel;
import scripting.InterThreadBuffer;
import scripting.InterThreadMonitor;
import scripting.Names;

/**
 *
 * @author hsatizab
 */
public class FNNXSSETester extends AbstractFNNXFunction implements FNNXObject, FNNXFunction {

    public FNNXSSETester(Hashtable<String, Object> vars, InterThreadMonitor m) {
        this.setName();
        configurationWindow = null;
        variables = vars;
        monitor = m;
        plotMode = Names.plotModes.none;

        parameterNames = new String[3];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";

        parameterTypes = new Class[3];
        parameterTypes[0] = AbstractFNNXSupervisedModel.class;
        parameterTypes[1] = DataBlock.class;
        parameterTypes[2] = AbstractFNNXParallelObserver.class;

        isParameterMandatory = new boolean[3];
        isParameterMandatory[0] = true;
        isParameterMandatory[1] = true;
        isParameterMandatory[2] = false;

        parameters = new Object[3];
    }

    public FNNXSSETester(JFrame parent) {
        this.setName();
        configurationWindow = new ConfigurationWindow(parent, true);
        variables = null;

        parameterNames = new String[3];
        parameterNames[0] = "network";
        parameterNames[1] = "dataset";
        parameterNames[2] = "saver";
    }

    public boolean prepare(InterThreadMonitor m) {
        if (configured) {
            return true;
        }
        monitor = m;

        if (!this.testCompleteness()) {
            return false;
        }

        for (int i = 0; i < parameterNames.length; i++) {
            try {
                switch (i) {
                    case 0:
                        network = (AbstractFNNXSupervisedModel)parameters[i];
                        break;
                    case 1:
                        dataset = (DataBlock)parameters[i];
                        break;
                    case 2:
                        saver = (AbstractFNNXParallelObserver)parameters[i];
                        break;
                }
            }
            catch(NullPointerException ne) {
                monitor.appendErrorMessage("Warning. Variable \"" + parameterNames[i] + "\" not found when configuring " + this.name + ". Using default value");
            }
            catch(IndexOutOfBoundsException ie) {
                monitor.appendErrorMessage("Internal error. Index out of bounds \"" + String.valueOf(i) + "\" when configuring " + this.name);
                return false;
            }
            catch(ClassCastException ce) {
                monitor.appendErrorMessage("Error. Wrong type in variable \"" + parameterNames[i] + "\" when configuring " + this.name);
                return false;
            }
        }

        res = new float[1];
        results = new InterThreadBuffer(res.length);

        configured = true;
        return true;
    }

    public Object eval(InterThreadMonitor m) {
        if (!this.prepare(m)) {
            return null;
        }

        float SSE = 0;
        int up;

        this.initSaver();

        network.setDataset(dataset);
        for (network.initDatasetSequential(); network.hasNextExampleSequential() && monitor.isRunning(); network.nextExampleSequential()) {
            while (monitor.isPaused()) {
            }
            network.calcOutput();
            SSE += network.calcModelError();
        }

        up = dataset.getNRows();
        if (up == 0) {
            up = 1;
        }

        res[0] = (float)SSE / up;

        if (hasPlotter) {
            results.lastData(res, 1);
            plotter.getMonitor().newResults(results);
        }
        if (hasSaver) {
            results.newData(res, 1);
            saver.getMonitor().newResults(results);
        }

        this.endSaver();

        return res;
    }

    private AbstractFNNXSupervisedModel network;
    private DataBlock dataset;

}
